%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         CROSSSUN386-LAP.SL
% Description:  Prefix for crosscompiling for Intel 80386 PSL Assembler
%               on a MC 68020
% Author:       H. Melenk
% Created:      1-Sept 1989
% Modified:
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum Berlin all rights reserved.
%
 
(compiletime (load fasl-decls sys-consts sys-macros))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   CROSS-FASL-SECTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This must be here when building the cross fasler.
% because  the 80386 needs words and halfwords the other way stored

(compiletime
% (setprop 'putbyte nil)       
  (setprop 'puthalfword nil)  
  (setprop 'putword nil)     
  (setprop 'putmem nil)
% (setprop 'mkitem nil)
  (setprop 'deflist nil)
(put 'getmem 'assign-op 'putmem)
 
)  
                          
(on testlap)   %%%%%% reichlich output

(de puthalfword (adr d w)
     (setq adr (wplus2 adr d))
     (putbyte adr 3 (wand w 255))
     (putbyte adr 2 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 1 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 0 (wand (setq w (wshift w -8)) 255)))
 
(de putword (adr d w)
     (setq adr (wplus2 adr d))
     (putbyte adr 7 (wand w 255))
     (putbyte adr 6 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 5 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 4 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 3 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 2 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 1 (wand (setq w (wshift w -8)) 255))
     (putbyte adr 0 (wand (setq w (wshift w -8)) 255)))
                         
(de putmem (adr w) (putword adr 0 w))

(de deposit-relocated-word (offset)
  % Given an OFFSET from CODEBASE*, deposit a word containing the
  % absolute address of that offset.
  (putword codebase* currentoffset*
	  (wplus2 offset (if *writingfaslfile 0 codebase*)))
  (updatebittable 4 reloc-word)
  (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem)))
  
(de depositwordexpression (x)  % Only limited expressions now handled                                   
  (let (y)
  (cond ((fixp x) (depositword (int2sys x)))
	((stringp x) (deposit-relocated-word 
                     (if *writingfaslfile (labeloffset x)
				(wshift (labeloffset x) 2))))
	((memq  (first x) '($fluid $global fluid global))
		        (depositvaluecelllocation (second x)))
	((equal (first x) 'mkitem) (deposititem (second x) (third x)))
	((equal (first x) 'foreignentry)
		                   (depositfunctioncelllocation (second x)))
	((equal (first x) 'immediate)
	 (cond ((setq y (assoc (second x) label-alist*))
		        (deposit-relocated-word (cdr y)))
	       ((and (pairp (second x))) (equal (first (second x)) 'idloc)
		        (depositwordidnumber (second (second x))))
	       (t (stderror (bldmsg "Undefined label: %w" (second x))))))

	((setq y (wconstevaluable x))    (depositword (int2sys y)))
	(t  (stderror (bldmsg "Expression too complicated %r" x))))))

(de depositword (x)
  (putmem (wplus2 codebase* currentoffset*) x)
  (updatebittable 8 0)
  (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem)))

(de internallycallablep(u) nil)

(de depositwordidnumber (x) 
  (cond
    ((or (not *writingfaslfile) (wleq (idinf x) 255)) (depositword (idinf X)))
    (t (putword (wplus2 codebase* currentoffset*) 0 
	       (makerelocword reloc-id-number (findidnumber x))) 
       (updatebittable 8 reloc-word))))

(de deposititem (tagpart infpart)
  (if (not *writingfaslfile)
    (depositword 
       (mkitem tagpart 
	  (cond ((atom infpart) (wplus2 codebase* (labeloffset infpart)))
		((equal (first infpart) 'idloc)    (idinf (second infpart)))
		(t (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))))
    (progn 
	(cond ((atom infpart) (putword codebase* currentoffset* 
		                                        % RELOC_CODE_OFFSET = 0
		                      (mkitem tagpart (labeloffset infpart))))
	      ((equal (first infpart) 'idloc) 
		 (putword codebase* currentoffset*
		         (mkitem tagpart 
		          (makerelocinf reloc-id-number
		           (findidnumber (second infpart))))))
	      (t  (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))
        (updatebittable 8 reloc-inf))
        (setq currentoffset* (iplus2 addressingunitsperitem currentoffset*))))

(commentoutcode
(de systemfaslfixup ()
  (prog (x type label place fwdir )
	(while forwardinternalreferences*
	  (setq fwdir (car forwardinternalreferences*))
	  (setq type (car fwdir) label (cadr fwdir) place (caddr fwdir))
	  (if (null (get label 'internalentryoffset))
	    (errorprintf "*** %r not defined in this module, call is incorrect"
		         (cadr fwdir))
	    (cond ((eq type 'CALL)
		   (putword codebase* place
		     (wor (wshift (getmem (wplus2 codebase* place)) 24)
		       (get label 'internalentryoffset) )))
		  (t  (error 0 (list "Strange forward internal reference: "
		                  fwdir)))))
	  (setq forwardinternalreferences* (cdr forwardinternalreferences*)))
	(mapobl (function (lambda (y) (remprop y 'internalentryoffset))))
))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(define-constant reloc-movex 1)

(de deposit-relocatable-movex (code)
		                 % code is (add r,0,r) where r is the dst.
  (setf (byte (wplus2 codebase* currentoffset*) 3) (wshift code -24))
  (setf (byte (wplus2 codebase* currentoffset*) 2) 
	(wand 16#ff (wshift code -16)))
  (setf (byte (wplus2 codebase* currentoffset*) 1)
	(wand 16#ff (wshift code -8)))
  (setf (byte (wplus2 codebase* currentoffset*) 0)
	(wand 16#ff code))
  (updatebittable 1 reloc-word)
  (updatebittable 3 reloc-movex) )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

)
(de store-word (word)  (puthalfword codebase*
		 (lxor 4 currentoffset*) word))

(de updatebittable (numberofentries firstentry)
  (when *writingfaslfile
   (if (eq numberofentries 4) % "an instruction"
    (progn
    (putbittable bittablebase* (lxor bittableoffset* 4) firstentry)
    (setq bittableoffset* (iadd1 bittableoffset*))
    (for (from i 2 numberofentries 1)
         (do (progn (putbittable bittablebase* (lxor 4 bittableoffset*) 0)
                    (setq bittableoffset* (iadd1 bittableoffset*))))))
                  % "NO instruction"
    (putbittable bittablebase* bittableoffset* firstentry)
    (setq bittableoffset* (iadd1 bittableoffset*))
    (for (from i 2 numberofentries 1)
         (do (progn (putbittable bittablebase* bittableoffset* 0)
                    (setq bittableoffset* (iadd1 bittableoffset*))))))
    (when (igreaterp bittableoffset* maxfasloffset*)
      (fatalerror "BPS exhausted during FaslOut; output file too large"))))


(copyd 'originalbinarywrite 'binarywrite)
(de binarywrite (f w)
   % big->little endian conversion
    (prog(v)
           % byte 4 -> 1
       (setq v(wshift w 56))
           % byte 3 -> 2
       (setq v (wor v (wshift (wand (wshift w -48) 255) 8)))
       (setq v (wor v (wshift (wand (wshift w -40) 255) 16)))
       (setq v (wor v (wshift (wand (wshift w -32) 255) 24)))
       (setq v (wor v (wshift (wand (wshift w -24) 255) 32)))
       (setq v (wor v (wshift (wand (wshift w -16) 255) 40)))
       (setq v (wor v (wshift (wand (wshift w -8) 255)  48)))
       (setq v (wor v (wshift (wand w  255) 56)))
       (originalbinarywrite f v)))

(commentoutcode
       
(loadtime (mapobl 
    (function (lambda(x) (remprop x 'INSTRUCTIONDEPOSITFUNCTION)))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    END-OF-CROSS-FASL-SECTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

)
