%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:COMP-DECLS.SL
% Description:  Not quite Machine-independent compiler declarations
% Author:       Eric Benson, Sam Sands
% Created:      16-October-1981
% Modified:     21-Jun-84 09:00:35
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 27-Aug-88 (Julian Padget)
%  Removed various bits that had been moved to p1-decls but not removed
%  from here (see Ambras 13-Jun-84)
% 05-Apr-88 (Julian Padget)
%  Restored various bits for the extra types that CPSL has for
%  fliptst, opentst and openfn properties
% 25-Jul-86 (Leigh Stoller)
%  Added a new matchfn called openfn to test if a pattern argument is an
%  openfn. The memory cmacro must be resolved with the more complicated arg
%  first so that the temp reg is available. This handles nested memory's
% 21-Jul-86 (Leigh Stoller)
%  Changed &getdtemp and &getatemp to use new predicate functions. See the
%  note below for an explanation.
% 13-Jun-84 15:26:27 (Jim Ambras/CRC)
%  Removed machine-independent declarations and put into PASS-1-DECLS.SL.
% 25-Apr-84 08:15:38 (Brian Beach)
%  Pass 1 functions removed from this file.  They now reside in PASS-1.SL.
% 27-Mar-84 08:23:48 (Brian Beach)
%  Added COMPFNs for IDAPPLY and CODEAPPLY.
%  Added argument count properties for COMPFNs.
% 19-Apr-84 07:45:34 (Brian Beach)
%  Removed constructors from the list of side-effect-free functions, since
%  any subsequent modification of those constants will cause problems.
% 16-Mar-84 13:20:27 (Brian Beach)
%  Added declaration of variables known to be constant.
%  Added declaration of side-effect-free functions.  Removed &DOOP as DOFN
%  (on WSHIFT and WNOT) since constants are automatically handled now.
% 15-Dec-83 10:59:59 (Tim Tillson)
%  Added Brian Beach's pass1 functions for WARRAYS, WVARS and WSTRINGS to
%  Sam's D-register model version
%
%  <PSL.COMP>COMP-DECLS.RED.16,  3-Sep-82 09:46:43, Edit by BENSON
%  Added PA1REFORMFN for WNOT
%  <PSL.COMP>COMP-DECLS.RED.5,   3-Dec-82 18:20:08, Edit by PERDUE
%  Removed PA1REFORMFN for NE
%  <PSL.COMP>COMP-DECLS.RED.6,  24-Jan-83 16:04:00, Edit by MLGriss
%  Changed W to !%!%!%W in the EQCAR to avoid subst W into EQCAR form

(put '!*JUMPEQ         'NEGJMP '!*JUMPNOTEQ)
(put '!*JUMPNOTEQ      'NEGJMP '!*JUMPEQ)
(put '!*JUMPWLESSP     'NEGJMP '!*JUMPWGEQ)
(put '!*JUMPWGEQ       'NEGJMP '!*JUMPWLESSP)
(put '!*JUMPWLEQ       'NEGJMP '!*JUMPWGREATERP)
(put '!*JUMPWGREATERP  'NEGJMP '!*JUMPWLEQ)

(put '!*JUMPTYPE       'NEGJMP '!*JUMPNOTTYPE)
(put '!*JUMPNOTTYPE    'NEGJMP '!*JUMPTYPE)
(put '!*JUMPINTYPE     'NEGJMP '!*JUMPNOTINTYPE)
(put '!*JUMPNOTINTYPE  'NEGJMP '!*JUMPINTYPE)

(put '!*SJUMPEQ         'NEGJMP '!*SJUMPNOTEQ)
(put '!*SJUMPNOTEQ      'NEGJMP '!*SJUMPEQ)
(put '!*SJUMPWLESSP     'NEGJMP '!*SJUMPWGEQ)
(put '!*SJUMPWGEQ       'NEGJMP '!*SJUMPWLESSP)
(put '!*SJUMPWLEQ       'NEGJMP '!*SJUMPWGREATERP)
(put '!*SJUMPWGREATERP  'NEGJMP '!*SJUMPWLEQ)

(put '!*SJUMPTYPE       'NEGJMP '!*SJUMPNOTTYPE)
(put '!*SJUMPNOTTYPE    'NEGJMP '!*SJUMPTYPE)
(put '!*SJUMPINTYPE     'NEGJMP '!*SJUMPNOTINTYPE)
(put '!*SJUMPNOTINTYPE  'NEGJMP '!*SJUMPINTYPE)

(put '!*SJUMPEQ         'LONGBRANCH '!*JUMPEQ)
(put '!*SJUMPNOTEQ      'LONGBRANCH '!*JUMPNOTEQ)
(put '!*SJUMPWLESSP     'LONGBRANCH '!*JUMPWLESSP)
(put '!*SJUMPWGEQ       'LONGBRANCH '!*JUMPWGEQ)
(put '!*SJUMPWLEQ       'LONGBRANCH '!*JUMPWLEQ)
(put '!*SJUMPWGREATERP  'LONGBRANCH '!*JUMPWGREATERP)

(put '!*SJUMPTYPE       'LONGBRANCH '!*JUMPTYPE)
(put '!*SJUMPNOTTYPE    'LONGBRANCH '!*JUMPNOTTYPE)
(put '!*SJUMPINTYPE     'LONGBRANCH '!*JUMPINTYPE)
(put '!*SJUMPNOTINTYPE  'LONGBRANCH '!*JUMPNOTINTYPE)

(put 'AND 	'OPENTST 	'&TSTANDOR)
(put 'OR 	'OPENTST 	'&TSTANDOR)

(put 'EQ 	'OPENTST 	'(TSTPAT  !*JUMPEQ       ))
(put 'NE 	'OPENTST 	'(TSTPAT  !*JUMPNOTEQ    ))
(put 'WGREATERP 'OPENTST 	'(TSTPAT  !*JUMPWGREATERP))
(put 'WLEQ 	'OPENTST 	'(TSTPAT  !*JUMPWLEQ     ))
(put 'WGEQ 	'OPENTST 	'(TSTPAT  !*JUMPWGEQ     ))
(put 'WLESSP 	'OPENTST 	'(TSTPAT  !*JUMPWLESSP   ))

(PUT 'PAIRP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE      PAIR-TAG))
(PUT 'ATOM 	'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE   PAIR-TAG))
(PUT 'STRINGP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE      STRING-TAG))
(PUT 'NOTSTRINGP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE   STRING-TAG))
(PUT 'VECTORP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE      VECTOR-TAG))
(PUT 'NOTVECTORP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE   VECTOR-TAG))
(PUT 'CODEP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE      CODE-TAG))
(PUT 'NOTCODEP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE   CODE-TAG))
(PUT 'FLOATP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE      FLOATNUM-TAG))
(PUT 'NOTFLOATP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE   FLOATNUM-TAG))
(PUT 'EVECTORP  'OPENTST        '(TSTPAT2 !*JUMPTYPE      EVECTOR-TAG))
(PUT 'NOTEVECTORP 'OPENTST      '(TSTPAT2 !*JUMPNOTTYPE   EVECTOR-TAG))
(PUT 'INTP 	'OPENTST 	'(TSTPAT2 !*JUMPINTYPE    POSINT-TAG))
(PUT 'NOTINTP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTINTYPE POSINT-TAG))
(PUT 'FIXP 	'OPENTST 	'(TSTPAT2 !*JUMPINTYPE    BIGNUM-TAG))
(PUT 'NOTFIXP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTINTYPE BIGNUM-TAG))
(PUT 'NUMBERP 	'OPENTST 	'(TSTPAT2 !*JUMPINTYPE    FLOATNUM-TAG))
(PUT 'NOTNUMBERP 'OPENTST 	'(TSTPAT2 !*JUMPNOTINTYPE FLOATNUM-TAG))

(PUT 'FIXNP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    FIXNUM-TAG))
(PUT 'NOTFIXNP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE FIXNUM-TAG))
(PUT 'BIGP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    BIGNUM-TAG))
(PUT 'NOTBIGP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE BIGNUM-TAG))
(PUT 'POSINTP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    POSINT-TAG))
(PUT 'NOTPOSINTP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE POSINT-TAG))
(PUT 'NEGINTP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    NEGINT-TAG))
(PUT 'NOTNEGINTP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE NEGINT-TAG))
(PUT 'IDP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    ID-TAG))
(PUT 'NOTIDP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE ID-TAG))
(PUT 'BYTESP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    BYTES-TAG))
(PUT 'NOTBYTESP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE BYTES-TAG))
(PUT 'WRDSP 	'OPENTST 	'(TSTPAT2 !*JUMPTYPE    WORDS-TAG))
(PUT 'NOTWRDSP 	'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE WORDS-TAG))
(PUT 'HALFWORDSP 'OPENTST 	'(TSTPAT2 !*JUMPTYPE    HALFWORDS-TAG))
(PUT 'NOTHALFWORDSP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE HALFWORDS-TAG))

%JAP: So the compiler can understand the new types

(put 'CONTEXTP 'OPENTST 	'(TSTPAT2 !*JUMPTYPE    CONTEXT-TAG))
(put 'NOTCONTEXTP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE CONTEXT-TAG))
(put 'BVECTORP 'OPENTST 	'(TSTPAT2 !*JUMPTYPE    BVECTOR-TAG))
(put 'NOTBVECTORP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE BVECTOR-TAG))
(put 'BSTRUCTP 'OPENTST 	'(TSTPAT2 !*JUMPTYPE    BSTRUCT-TAG))
(put 'NOTBSTRUCTP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE BSTRUCT-TAG))
(put 'FUNARGP 'OPENTST	 	'(TSTPAT2 !*JUMPTYPE    FUNARG-TAG))
(put 'NOTFUNARGP 'OPENTST 	'(TSTPAT2 !*JUMPNOTTYPE FUNARG-TAG))

(put 'SGDP 'OPENTST	 	'(TSTPAT2 !*JUMPTYPE    SGD-TAG))
(put 'NOTSGDP 'OPENTST	 	'(TSTPAT2 !*JUMPNOTTYPE SGD-TAG))

(PUT 'EQ 	'OPENFN 	'(TVPAT !*SJUMPEQ       ))
(PUT 'NE 	'OPENFN 	'(TVPAT !*SJUMPNOTEQ    ))
(PUT 'WGREATERP 'OPENFN 	'(TVPAT !*SJUMPWGREATERP))
(PUT 'WLEQ 	'OPENFN 	'(TVPAT !*SJUMPWLEQ     ))
(PUT 'WGEQ 	'OPENFN 	'(TVPAT !*SJUMPWGEQ     ))
(PUT 'WLESSP 	'OPENFN 	'(TVPAT !*SJUMPWLESSP   ))

(PUT 'PAIRP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    PAIR-TAG))
(PUT 'ATOM 	'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE PAIR-TAG))
(PUT 'STRINGP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    STRING-TAG))
(PUT 'NOTSTRINGP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE STRING-TAG))
(PUT 'VECTORP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    VECTOR-TAG))
(PUT 'NOTVECTORP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE VECTOR-TAG))
(PUT 'CODEP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    CODE-TAG))
(PUT 'NOTCODEP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE CODE-TAG))
(PUT 'FLOATP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    FLOATNUM-TAG))
(PUT 'NOTFLOATP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE FLOATNUM-TAG))
(PUT 'EVECTORP  'OPENFN         '(TVPAT1 !*SJUMPTYPE    EVECTOR-TAG))
(PUT 'NOTEVECTORP 'OPENFN       '(TVPAT1 !*SJUMPNOTTYPE EVECTOR-TAG))

(PUT 'INTP 	'OPENFN 	'(TVPAT1 !*SJUMPINTYPE  POSINT-TAG))
(PUT 'NOTINTP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTINTYPE POSINT-TAG))
(PUT 'FIXP 	'OPENFN 	'(TVPAT1 !*SJUMPINTYPE  BIGNUM-TAG))
(PUT 'NOTFIXP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTINTYPE BIGNUM-TAG))
(PUT 'NUMBERP 	'OPENFN 	'(TVPAT1 !*SJUMPINTYPE  FLOATNUM-TAG))
(PUT 'NOTNUMBERP 'OPENFN 	'(TVPAT1 !*SJUMPNOTINTYPE FLOATNUM-TAG))

(PUT 'FIXNP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    FIXNUM-TAG))
(PUT 'NOTFIXNP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE FIXNUM-TAG))
(PUT 'BIGP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    BIGNUM-TAG))
(PUT 'NOTBIGP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE BIGNUM-TAG))
(PUT 'POSINTP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    POSINT-TAG))
(PUT 'NOTPOSINTP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE POSINT-TAG))
(PUT 'NEGINTP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    NEGINT-TAG))
(PUT 'NOTNEGINTP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE NEGINT-TAG))
(PUT 'IDP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    ID-TAG))
(PUT 'NOTIDP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE ID-TAG))
(PUT 'BYTESP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    BYTES-TAG))
(PUT 'NOTBYTESP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE BYTES-TAG))
(PUT 'WRDSP 	'OPENFN 	'(TVPAT1 !*SJUMPTYPE    WORDS-TAG))
(PUT 'NOTWRDSP 	'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE WORDS-TAG))
(PUT 'HALFWORDSP 'OPENFN 	'(TVPAT1 !*SJUMPTYPE    HALFWORDS-TAG))
(PUT 'NOTHALFWORDSP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE HALFWORDS-TAG))

%JAP: So the compiler can understand the new types

(put 'CONTEXTP 'OPENFN 		'(TVPAT1 !*SJUMPTYPE    CONTEXT-TAG))
(put 'NOTCONTEXTP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE CONTEXT-TAG))
(put 'BVECTORP 'OPENFN 		'(TVPAT1 !*SJUMPTYPE    BVECTOR-TAG))
(put 'NOTBVECTORP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE BVECTOR-TAG))
(put 'BSTRUCTP 'OPENFN 		'(TVPAT1 !*SJUMPTYPE    BSTRUCT-TAG))
(put 'NOTBSTRUCTP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE BSTRUCT-TAG))
(put 'FUNARGP 'OPENFN	 	'(TVPAT1 !*SJUMPTYPE    FUNARG-TAG))
(put 'NOTFUNARGP 'OPENFN 	'(TVPAT1 !*SJUMPNOTTYPE FUNARG-TAG))

(put 'SGDP 'OPENFN	 	'(TVPAT1 !*SJUMPTYPE    SGD-TAG))
(put 'NOTSGDP 'OPENFN 		'(TVPAT1 !*SJUMPNOTTYPE SGD-TAG))

(put 'WPLUS2 	'OPENFN 	'(ADDPAT       !*WPLUS2))       %      scs
(put 'WDIFFERENCE 'OPENFN 	'(SUBPAT       !*WDIFFERENCE))      
% (put 'WTIMES2 'OPENFN 	'(ASSOCPAT     !*WTIMES2))        NOT USED scs
(put 'WMINUS 	'OPENFN 	'(UNARYPAT     !*WMINUS))
(put 'WNOT 	'OPENFN 	'(UNARYPAT     !*WNOT))
(put 'WAND 	'OPENFN 	'(ANDORXORPAT  !*WAND))         %      scs
(put 'WOR 	'OPENFN 	'(ANDORXORPAT  !*WOR))          %      scs
(put 'WXOR 	'OPENFN 	'(ANDORXORPAT  !*WXOR))         %      scs
(put 'WSHIFT 	'OPENFN 	'(SHIFTPAT     !*WSHIFT))       %      scs
(put 'ASHIFT 	'OPENFN 	'(SHIFTPAT     !*ASHIFT))       %      scs
(put 'GETUNSBYTE 'OPENFN         '(GETUNSBYTE   !*MOVEUNSBYTE))  %      scs
(put 'PUTANYBYTE 'OPENFN         '(PUTANYBYTE   !*MOVEUNSBYTE))  %      scs
(put 'GETSGNHWORD 'OPENFN        '(GETSGNHWORD  !*MOVESGNHWORD)) %      scs
(put 'PUTANYHWORD 'OPENFN        '(PUTANYHWORD  !*MOVESGNHWORD)) %      scs
(put 'MKITEMREV 'OPENFN 	'(MKITEMPAT    !*MKITEM))       %      scs
(put 'FIELD 	'OPENFN 	'(FIELDPAT     !*FIELD))
(put 'SIGNEDFIELD 'OPENFN 	'(FIELDPAT     !*SIGNEDFIELD))
(put 'PUTFIELDREV 'OPENFN 	'(PUTFIELDPAT  !*PUTFIELD))

(put 'LOC 	'OPENFN 	'(LOCPAT       !*LOC))          %      scs

(put '$NAME      'OPENFN         '$NAME)             % DUMMY     %   scs
(put 'CAR        'OPENFN         '(CARCDRPAT CAR))               %   scs
(put 'CDR        'OPENFN         '(CARCDRPAT CDR))               %   scs
(put 'MEMORY     'OPENFN         '(MEMORYPAT MEMORY))            %   scs

(put 'SETQ 	'OPENFN 	'(SETQPAT NIL))
(put 'RPLACA 	'OPENFN 	'(RPLACPAT CAR))
(put 'RPLACD 	'OPENFN 	'(RPLACPAT CDR))

% In memory operations

(put '!*ADDMEM 	'OPENFN 	'(ADDMEMPAT !*ADDMEM))           %   scs
(put '!*SUBMEM 	'OPENFN 	'(SUBMEMPAT !*SUBMEM))           %   scs
%(put '!*MPYMEM 'OPENFN 	'(MODMEMPAT !*MPYMEM))     NOT USED

(flag '(!*ADDMEM !*SUBMEM)       'MEMMOD)       %  !*MPYMEM NOT USED  scs

(put 'WPLUS2       'MEMMODFN  '!*ADDMEM)
% (put 'WDIFFERENCE  'MEMMODFN  '!*SUBMEM)    sadly, there's a bug  %   scs
% (put 'WTIMES2      'MEMMODFN  '!*MPYMEM)                 NOT USED   scs

(put '!*ADDMEM  'UNMEMMOD  '!*WPLUS2) 
(put '!*SUBMEM  'UNMEMMOD  '!*WDIFFERENCE)                        %   scs
% (put '!*MPYMEM  'UNMEMMOD  '!*WTIMES2)                   NOT USED   scs


% Match functions

(put 'ANY       'MATCHFN '&ANY)
(put 'VAR       'MATCHFN '&VAR)
% (put 'REG       'MATCHFN '&REGFP)   REDEFINED BELOW             scs
(put 'DEST      'MATCHFN '&DEST)
(put 'USESDEST  'MATCHFN '&USESDEST)
(put 'REGN      'MATCHFN '&REGN)
(put 'NOTDEST   'MATCHFN '&NOTDEST)
(put 'NOTANYREG 'MATCHFN '&NOTANYREG)
(put 'MEM       'MATCHFN '&MEM)
(put 'ANYREGFN  'MATCHFN '&ANYREGFNP)

% Need this extra pattern predicate to test for openfn's as an argument to
%  anyreg. /LBS
%
(put 'OPENFN    'MATCHFN '&OPENFNP)

(put 'REG       'MATCHFN  'REGP)                              %     scs 
(put 'AREG      'MATCHFN  'AREGP)                             %     scs 
(put 'DREG      'MATCHFN  'DREGP)                             %     scs 
(put 'REGNP     'MATCHFN  '&REGNP)                           %     scs 
(put 'CONST     'MATCHFN  '&CONSTP)                          %     scs

(put 'DDEST     'MATCHFN  '&DDESTP)                          %     scs 
(put 'DREGNP    'MATCHFN  '&DREGNP)                          %     scs 
(put 'AREGNP    'MATCHFN  '&AREGNP)                          %     scs 
(put 'DREGDEST  'MATCHFN  '&DREGDESTP)                       %     scs 
(put 'AREGDEST  'MATCHFN  '&AREGDESTP)                       %     scs 
(put 'AFRAME    'MATCHFN  '&AFRAMEP)                         %     scs
(put 'NOTIMM    'MATCHFN  '&NOTIMMP)                         %     scs

(put 'ZERO      'MATCHFN  '&ZEROPP)                          %     scs
(put 'LEQ2      'MATCHFN  '&LEQ2P)                           %     scs
(put 'SMALL     'MATCHFN  '&SMALLP)                          %     scs
(put 'MOVEQ     'MATCHFN  '&MOVEQP)                          %     scs
(put 'DISP      'MATCHFN  '&DISPP)                           %     scs
(put 'DISPL     'MATCHFN  '&DISPLP)                          %     scs

(de &DDESTP    (X)  (DREGP DEST&))                           %     scs 
(de &DREGNP    (R)  (&TYPEREGNP R 'DREGP))                   %     scs 
(de &AREGNP    (R)  (&TYPEREGNP R 'AREGP))                   %     scs 
(de &DREGDESTP (R)  (AND (DREGP R) (&DEST R)))               %     scs 
(de &AREGDESTP (R)  (AND (AREGP R) (&DEST R)))               %     scs 
(de &AFRAMEP   (R)  (OR (AREGP R)  (EQCAR R '$LOCAL)))       %     scs 
(de &NOTIMMP   (R)  (AND (PAIRP R) (MEMBER (CAR R) 
		                 '($LOCAL MEMORY CAR CDR)))) %     scs 

(de SAFECADR (R) (COND ((PAIRP (CDR R)) (CADR R))))		%     scs
(de &ZEROPP (R) (AND (&CONSTP R) (ZEROP       (SAFECADR R))))   %     scs
(de &LEQ2P  (R) (AND (&CONSTP R) (OR (ONEP    (SAFECADR R)))    %    gls
                                     (TWOP    (SAFECADR R))))   %     scs
(de &SMALLP (R) (AND (&CONSTP R) (SMALLINUMP  (SAFECADR R))))   %     scs
(de &MOVEQP (R) (AND (&SMALLP R) (NOT (QINUMP (CADR R)))))	%     scs
(de &DISPP  (R) (AND (&CONSTP R) ( DISPINUMP  (SAFECADR R))))	%     scs
(de &DISPLP (R) (AND (&DISPP  R) (NOT (SMALLINUMP (CADR R)))))	%     scs


% Macro functions

(put 'A1  'SUBSTFN '&ARG1)
(put 'A2  'SUBSTFN '&ARG2)
(put 'A3  'SUBSTFN '&ARG3)
(put 'A4  'SUBSTFN '&ARG4)
(put 'FN  'SUBSTFN '&PARAM1)
(put 'MAC 'SUBSTFN '&PARAM2)
(put 'P2  'SUBSTFN '&PARAM3)
(put 'P3  'SUBSTFN '&PARAM4)
(put 'T1  'SUBSTFN '&GETTEMP)
(put 'T2  'SUBSTFN '&GETTEMP)
(put 'T3  'SUBSTFN '&GETTEMP)
(put 'T4  'SUBSTFN '&GETTEMP)
(put 'L1  'SUBSTFN '&GETTEMPLBL)
(put 'L2  'SUBSTFN '&GETTEMPLBL)
(put 'L3  'SUBSTFN '&GETTEMPLBL)
(put 'L4  'SUBSTFN '&GETTEMPLBL)

(put 'LA1 'SUBSTFN '&LOCARG1)                                         % scs
%(put 'TA1 'SUBSTFN '&GETATEMP)                                        % scs
%(put 'TA2 'SUBSTFN '&GETATEMP)                                        % scs
%(put 'TD1 'SUBSTFN '&GETDTEMP)                                        % scs
%(put 'TD2 'SUBSTFN '&GETDTEMP)                                        % scs

%Temp fix to register allocation. Probably not correct.
%
(put 'TA1 'SUBSTFN '&GETTEMP)                                       
(put 'TA2 'SUBSTFN '&GETTEMP)                                       
(put 'TD1 'SUBSTFN '&GETTEMP)                                       
(put 'TD2 'SUBSTFN '&GETTEMP)                                       

(put 'NEGMAC  'SUBSTFN  '&NEGPARAM2)                                  % scs

                                                                       % scs
(de &LOCARG1  (TNAME ARGS PARAMS) (&LOCATE (&ARG1 TNAME ARGS PARAMS)))
%(de &GETDTEMP (TNAME ARGS PARAMS) (&GETTYPETEMP   TNAME ARGS PARAMS 'DREGP))
%(de &GETATEMP (TNAME ARGS PARAMS) (&GETTYPETEMP   TNAME ARGS PARAMS 'AREGP))
%
% These two new definitions reflect the fact that there are now only 6 real
%  regs. The compiler, when trying to find temp regs can no longer use the
%  the orignals predicates, since they will return false for most of the regs.
%  When real regs = temp regs, this was not a problem. Now we must use a
%  seperate function that allows all regs to be temps. It is up to the compiler
%  to choose one that is not protected. If gettypetemp was not 14 real reg
%  dependent this would not be needed.
%
(de &GETDTEMP (TNAME ARGS PARAMS) 
	(&GETTYPETEMP   TNAME ARGS PARAMS 'TEMPDREGP))

(de &GETATEMP (TNAME ARGS PARAMS) 
	(&GETTYPETEMP   TNAME ARGS PARAMS 'TEMPAREGP))

% Emit functions

(put '!*LOAD    'EMITFN '&EMITLOAD)
(put '!*STORE   'EMITFN '&EMITSTORE)
(put '!*JUMP    'EMITFN '&EMITJUMP)
(put '!*LBL     'EMITFN '&EMITLBL)
(put '!*ADDMEM  'EMITFN '&EMITMEMMOD)
(put '!*SUBMEM  'EMITFN '&EMITMEMMOD)                           %       scs
% (put '!*MPYMEM  'EMITFN '&EMITMEMMOD)      NOT USED                   scs

(put '!*USES    'EMITFN '&USES)                                 %       scs
(put '!*RETURN  'EMITFN '&RETURN)                               %       scs


% Optimizing functions

(put '!*LBL 	'OPTFN 	'&LBLOPT)
(put '!*MOVE 	'OPTFN 	'&STOPT)
(put '!*JUMP 	'OPTFN 	'&JUMPOPT)		

% Assorted other flags

(flag '(!*JUMP !*LINKE !*EXIT)                'TRANSFER)
(flag '(!*LINK !*LINKE)                       'UNKNOWNUSE)

(put '!*LINK  'EXITING  '!*LINKE)

% Initialize variables

(setq !*MSG             T)				% Do print messages
(setq !*INSTALLDESTROY  NIL)
(setq !*USINGDESTROY    T)
(setq !*SHOWDEST        NIL)
(setq !*NOFRAMEFLUID    T)
(setq !*USEREGFLUID     NIL)
(setq !*NOLINKE         NIL)       %. Permit LINKE
(setq !*ORD             NIL)		%. Dont force ORDER
(setq !*R2I             T)		%. Do convert Rec to Iter
(setq GLOBALGENSYM&     (LIST (GENSYM)))	 % initialize symbol list

(COMMENTOUTCODE (    
(setq MAXNARGS&       15)                               %  A REGISTER MODEL  scs
(setq LASTACTUALREG&   5)
))

(setq MAXNARGS&       14)                              %  D REGISTER MODEL  scs
%(setq LASTACTUALREG&  14)
%
% This also needs to be set to 6, or the value of maxrealregs. See the comment
%  in sys-consts for an explanation. /LBS
%
(setq LASTACTUALREG& 6)




