%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:XXX-LAP.SL
% Description:  Motorola MC68000 PSL Assembler
% Author:       E. Benson
% Created:      4-July-1982
% Modified:
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 3-June-92 (Winfried Neun)
%  Added conditional cache flushing for NEXT and Bobcat
%  when compiling into memory
% 07-Apr-88 (Julian Padget)
%  Merged code from lap-cpsl.  Redefined DEPOSITWORDEXPRESSION and added
%  definition of DEPOSITWORDIDNUMBER.
% 22-Apr-87 (Harold Carr)
%  Added DEPOSITFOREIGNENTRY function and made FOREIGNENTRY operands use it.
%  This allows us to compile and assemble references to foreign functions.
% 12-Sep-84 08:56:03 (Brian Beach)
%  Removed load of SYSLISP.
% 6-Jul-84 (Vicki O'Day)
%  Took out foreign link function.
% 28-Jun-84 09:53:17 (Brian Beach)
%  Fixed LABELOFFSET so that when compiling in core, a code pointer is
%  translated into an offset from CODEBASE*.
% 28-mar-84 13:30:00 (Steve Lowder)
%  Added the foreign link function to allow non-cross-compiler use of
%  flagging a function foreign.  Fixed MOVEM instruction.
%  Also rewrote the reglist functions to allow a greater variety of syntax.
% 27-Mar-84 14:06:33 (Brian Beach)
%  Added ENTRY to the things DEPOSITWORDEXPRESSION will handle.
%  For use with: (*MOVE (ENTRY (ID2INT 0)) (REG 1))
% 23-Mar-84 11:21:32 (Brian Beach)
%  Added ability to compile (FULLWORD (INTERNALENTRY xxx)), but not for
%  forward references.
% 16-Jan-84 17:55:31 (Nancy Kendzierski)
%  Added contents of .BUILD file.
% 15-Dec-83 13:23:48 (Tim Tillson)
%  Added header to Sam's D-register model version
%  Nov-83  Sam Sands
%   Fixed bug with BSR.S and BRA.S to get instructionlength on property list
%  13-Oct-83  Sam Sands
%   Moved procedure 'RegisterNumber (and DEFLIST) to  "PHPC:HP-CMAC.SL"
%  22-Jul-83  Nancy Kendzierski
%   Changed use of "deferred" (in comments) to "indirect".
%  1-Jul-83 Steve Lowder
%   The addressing modes for autoincrement and autodecrement were changed to
%   postincrement and predecrement.
%  17-May-83  Stephen Lowder
%   Corrected problem in procedure FindLongBranches. This assembler initially
%   assumes all branches are short.  Prior to this fix, it only converted to
%   long branches those branches greater than 127 bytes. A branch to the next
%   instruction is a branch of 0 length, but this must also be assembled as 
%   a long branch.
%  17-May-83  Stephen Lowder
%   Renamed the procedure Move to be LapMove to remove name conflict.
%  MLG, 14 fEB 1983
%   Corrected small bug in TwoRegisterAux
%   was clobbering INSTR, for ADDX!.L
%  Corrected and substantially revised and commented! by G. Q. Maguire Jr.
%  <BENSON>m68k-lAP.RED.2,  4-Jul-82 15:06:45, Edit by BENSON
%   Stripped VAX-LAP for skeleton of m68k-lAP
%   Also examined in the process was <maguire.68000>simple68.red by Maguire
%   and Griss, an assembler based on the VAX work of Benson and the mc68000
%   work of Norman.  Places where more info is needed are indicated by ???.
%   There also are many other places where documentation could be added, but
%   the triple questions are some low points.
%   1. How come the knowledge of ".B", ".W", and ".L" are sprinkled throught
%      the code in MapLists?
%   3. Does not implement all of the m68k addressing modes, such as Absolute
%      Word, and PC REL with index.
%   4. Ragged coding with accessing parts of objects of know form
%      (see DisplacementOperand)
%   5. "6" is hardwired in as the ZeroAreg!!! [Clearly this is a VERY
%      specific assembler, one could get very badly bitten by these kind of
%      things]
%  25-Jul-82  Maguire
%   Added STOP opcode with zero operands
%  <PSL.68000-INTERP>m68k-lAP.RED.10, 25-Jul-82 17:53:14, Edit by BENSON
%   Changed STOP to one operand (status register value)
%  <PSL.68000-INTERP>m68k-lAP.RED.11, 26-Jul-82 03:44:30, Edit by MAGUIRE
%   Found some errors:
%   In QuickArithmetic the test for PairP constant should check to make sure
%   that the lap tag is Immediate rather than assuming it.
%   (This is only one example of where the code ASSUMES that it is given a
%   correct form.  If it is not then it does something random which will
%   bite you later.)
%   The fields of fixed size, such as the 3 bit field in QuickArithmetic
%   should be checking the size or atleast masking it to 3 bits.
%   RegisterOpModeAddress does not do the correct thing for EOR, as EOR must
%   have a Dreg destination and cannot have the switch take place.
%   In the code for memory shifts within ShiftAndRotate it is unclear how the
%   type of the instruction gets moved from bits 3 and 4 to bits 9 and 10.
%   In addition there are some other problems with these operations.
%   Note that BTST is already used, thus it needs to exist.
%   Note that TRAP is already used (in the HP version), thus it needs to exist.
%  9-Sep-82  Maguire
%   Extensive revision and commenting
%   1. All the InternalFunction stuff canned - if the name space is that sparse
%      then we should have packages!
%   2. Similarly made it so the code can be run interpretively.
%   3. Note that since the assumption is built althrough the code that
%      addresses are positive integers (i.e. SYS tag = 0) there is no need to
%      do Iarith as regular arithmetic can be done on integers.
%   4. Similarly Instructions are 16 bits which is certainly less than 24
%      bits thus all the construction of instructions can be done without
%      using Iarith routines.
%   5. Reworked the code so that it is much more readable, with lots of
%      rearrangement, better use of functions, some macros, ... .
%   6. Fixed the Immediate mode instructions to emit two 16 bit words when
%      the instruction is LONG mode. Eric's code only emitted 16 bits in
%      either case.
%  10-Sep-82  Maguire
%   1. Replaced the explict fix in (6. above) to use DepositImediate
%  11-Sep-82  Maguire
%   1. Fixed code generation for EXG instruction, in the case of mixed mode
%      the contents of the regY must be an Areg, there for the registers must be
%      switched; this is another example of Eric not reading the complete
%      description of the instruction.
%   2. There are still quite a few places where BYTE mode versions of an
%      instruciton are illegal but the code does not currently check for this
%      case.
%   3. The memory mode of Shift and Rotate operations is incorrect, the
%      instruction is missing the opcode which needs to be in a different
%      field but is lost!
%   4. Change the output radix for the base address of code to be HEX, since
%      all the 68000 systems have debuggers which preferto work in hex. The
%      message also includes the radix prefix for both the base address (16#)
%      and the length (10#).
%   5. The code that I have added and revised is based on:
%      my perception of the instruction classes via
%      reading the MC68000 User's Manual, in particular appendix C and the BCPL
%      based assembler that was written by Ian Wilson at Cambridge Univ. in
%      the U.K.
%   6. Added the bit operations (BCHG, BCLR, BSET, BTST); CMPM; MOVEM; MOVEQ;
%      MOVEP; OR; TRAP (I think that is all of the instructions)
%   7. Changed the size field of BTST and related instructions. There appears
%      to be a logic error regarding the fluid instructionsize in the procedure
%      oprandlength. Thsi procedure assumes that long instructions will require
%      long immediate words for immediate values. This is not necessarily the
%      case.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This must be here when building the cross fasler.
%(compiletime
%  (setprop 'putbyte nil)       % First three to allow us to define our own
%  (setprop 'puthalfword nil)   %  functions that reverse the byte ordering
%  (setprop 'putword nil)       %  for 68K machines.
%  (setprop 'putmem nil)
%  (setprop 'getmem nil)
%  (setprop 'mkitem nil)
%  (setprop 'deflist nil))      % We do not want deflists evaluated, or else
                               % the vax instruction lengths will be over-
			       % written, causing branches to calculate wrong.

% ------------------------------------------------------------
% Fluid declarations:
% ------------------------------------------------------------

(fluid '(LabelOffsets*                  % just the label entries from
                                        % BranchAndLabelAList!*
                                        % Has the form ( (Label.Offset) ... )
        BranchCodeList*                 % Used in Branch optimization
        BranchAndLabelAList*            % Used in Branch optimization
        CurrentOffset*                  % the global value of the current
                                        % byte displacement from the starting
                                        % point of the code
        CodeSize*                       % Current number of bytes generated
        CodeBase*                       % Starting address of the module
        Entries*                        % list of procedure entries of the
                                        % form
                                        % ((ProcedureName
                                        %   ProcedureType
                                        %   NumberOfArguments) .
                                        %   CurrentOffset!* )
        InstructionChanged*             % Boolean - indicates if any
                                        % instructions have changed due to
                                        % branch optimization
        InstructionSize*                % Contains the size constant Byte
                                        % Word or Long during length compute
                                        % and assembly of individual instr.
        ForwardInternalReferences*      % a-list of offsets of references to
                                        % internal functions, to be patched
                                        % by SystemFaslFixup
        LapReturnValue*                 % set by SaveEntry to the pointer
                                        % to be returned by LAP
        OperandRegisterNumber*          % see EffectiveAddress
        OperandExtension*               % ditto
        *WritingFaslFile                % FLAG: if true, then we are writing
                                        % the resulting code to a file,
                                        % otherwise we are depositing it into
                                        % memory directly
        InitOffset*                     % The offset from the module to the
                                        % Initialiization code which is to be
                                        % run when the module is loaded
        *PGWD                           % FLAG: if true, then mnemonics and
                                        % assembled instructions are printed
        *PWrds))                        % FLAG: if true, then base address
                                        % and size of each compiled
                                        % procedure are printed as they are
                                        % deposited into memory
        


(ds LabelP (X) 
    (atom X))

(setq *PWrds t)                         % By default show where the code is
                                        % put in memory

% ------------------------------------------------------------
% Constant declarations:
% ------------------------------------------------------------

(DefConst byte 0                        % instruction size constants on 68000
         Word 1 
         Long 2 
         RELOC_ID_NUMBER 1 
         RELOC_HALFWORD 2 
         RELOC_WORD 1 
         RELOC_INF 3)

(DefConst MaximumShortBranch 127)

% ------------------------------------------------------------
% Start of actual code
% ------------------------------------------------------------

% LAP68(Code:list)
% Purpose: Assembles the LAP and returns a Code-pointer to the generated
%          code or writes the generated code to a file depending on
%          the flag !*WritingFaslFile.
% The argument to the function LAP is a list of lists and atoms, where the
% lists are instructions, pseudo-ops and cmacros, and the atoms are labels
% which are used to refer to positions in the code.  Note that these need not
% be IDs, but can also be strings, which can save on ID space.  Instructions
% should be of the form (opcode . operands), where opcode is a
% mnemonic for an opcode, and operands is a list of operands.  Each
% operand should be either an integer, which represents an immediate integer
% operand, or a label, or a list of the form (mode . suboperands).  A
% mode is an addressing mode, such as INDEXED or INDIRECT on the PDP-10,
% and DISPLACEMENT, INDIRECT, POSTINCREMENT, etc. for the MC68000's.  REG must
% exist on all machines, others will be chosen as appropriate for the system.
% Remember that these will mainly be used for cmacro expansions, not so much
% for writing code, so choose names for mnemonic value rather than brevity.
% Suboperands may also be operands, or they may be specific to the mode,
% e.g. register names.

(de Lap (U) 
(prog (LabelOffsets* LapReturnValue* Entries* temp) 
    (cond ((not *WritingFaslFile) (setq CurrentOffset* 0))) 
    (setq u (&fillframeholes u))
    (setq U (Pass1Lap U))               % Pass1lap
                                        % expand all the LAP macros
                                        % Note that this is defined in
                                        % PC:PASS-1-LAP.SL
%    if CheckForInitCode U then U := ProcessInitCode  U
%      else
 (setq U (OptimizeBranches U)) 
                                        % ProcessInitCode makes all
                                        % generic branches long to save time
                                        % when compiling.
                                        % OptimizeBranches:
                                        % compute the positions of labels
                                        % and optimize branches

% Print the machine specific assembly code
% if the object is an atom then it is a LABEL
% otherwise it is an instruction

    (cond (*PGWD (foreach X in U do 
        (cond ((LabelP X) (Prin2 X)) (t (PrintF "          %p%n" X)))))) 

    (cond ((not *WritingFaslFile)       % round off to fullword address
        (setq CodeBase* (GTBPS (Quotient (Plus2 CodeSize* 3) 4))))) 

    (foreach Instruction_or_Label in U do 
        (cond ((not (LabelP Instruction_or_Label)) 
            (cond ((equal (first Instruction_or_Label) '*entry) 
                (SaveEntry Instruction_or_Label)) (t 
            (DepositInstruction Instruction_or_Label)))))) 

    (DefineEntries)                     % define entries to whom?

% If you are depositing it into memory the tell the user how much space the
% code took and where it was loaded.
% ??? Why is this using the error channel ???

    (cond ((and (not *WritingFaslFile) (getd 'trap_2)) (trap_2)))
% flush I cache on a NeXT
    (cond ((and (not *WritingFaslFile) (getd 'flushcache))
			(flushcache CodeBase* CodeSize*)))
% same for bobcat (with 68040)
    (cond ((and (not *WritingFaslFile) *PWrds) 
        (ErrorPrintF "*** %p: base 16#%x, length 10#%d bytes" 
                (foreach X in Entries* collect (first (car X))) 
                                CodeBase* CodeSize*))) 

    (return (MkCODE LapReturnValue*)))) % How does this point at the code?
                                        % It is a fluid variable that got
                                        % set up when the code was generated.


% CheckForInitCode will scan the Codelist for the first !*Entry
% testing for !*!*FASL!*!*Initcode!*!*.

(de CheckForInitCode (CodeList) 
     (foreach Instruction in CodeList do 
       (progn (cond ((PairP Instruction) 
           (cond ((equal (car Instruction) '*entry) 
             (cond ((equal (second Instruction) '**Fasl**InitCode**) 
                (return t))))))))))

% SaveEntry( '(!*entry ProcedureName ProcedureType NumberOfArguments) )
% Purpose: To associate with a procedure its location (so other routines can
%          access it

(de saveentry (x)
  (cond  
   % if X = ( _____ !*!*!*Code!*!*Pointer!*!*!* ... )
   ((equal (second x) '***code**pointer***) 
    (setq lapreturnvalue* 
      (if *writingfaslfile
	currentoffset*
	(plus codebase* currentoffset*))))

   % If depositing into memory
   ((not *writingfaslfile) 
    (setq entries* (cons (cons (rest x) currentoffset*) entries*)) 
    (unless lapreturnvalue* (setq lapreturnvalue* (plus codebase* currentoffset*))))

   % if X = ( _____ !*!*Fasl!*!*InitCode!*!* ... )
   ((equal (second x) '**fasl**initcode**) 
    (setq initoffset* currentoffset*))

   % if X is an InternalFunction
   ((flagp (second x) 'internalfunction) 
    (put (second x) 'internalentryoffset currentoffset*))

   (t (progn
       (findidnumber (second x))
       (dfprintfasl (list 'putentry (mkquote (second x)) 
			  (mkquote (third x)) currentoffset*))))))
     

% DefineEntries()
% Purpose: Defines each of the procedures named in the list Entries!*
%          by putting the code pointer into the function cells

(de DefineEntries nil 
    (foreach X in Entries* do 
        (PutD (first (car X)) (second (car X)) (MkCODE (plus CodeBase* (cdr X))))))

(de DepositInstruction (X) 
% This actually dispatches to the procedures to assemble the instrucitons
(prog (Y) 
    (setq InstructionSize* (get (first X) 'Size)) 

    (cond ((setq Y (get (first X) 'InstructionDepositFunction)) 
      (Apply Y (list X))) (t            % actually assemble the instructions
      (StdError (BldMsg "Unknown instruction %p" X))))))



(CompileTime (progn 

(dm DefOpcode (U) 
%
% (DefOpcode name number assemble-fn)
% (DefOpcode name number assemble-fn instr-size)
% (DefOpcode name number assemble-fn instr-size length)
%
(prog (OpName OpValue OpFn OpSize OpLen) 
    (setq U (rest U)) 
    (setq OpName (MkQuote (first U)))   % (quote name)
    (setq OpValue (second U)) 
    (setq OpFn (list 'function (third U))) 	% (function name)

    (setq U (cdr (cdr (cdr U))))        % Skip first 3 args

    (setq OpSize (const Word))          % size is WORD by default
    (setq OpLen nil)                    % length computation is "standard"
                                        % by default, see function
                                        % StandardInstructionLength.

    (cond (U 
    (progn (setq OpSize (first U)) 
        (cond ((IDP OpSize) 
            (setq OpSize (list 'const OpSize)))) 	% (const size)
        (setq U (rest U))))) 

    (cond (U 
    (progn (setq OpLen (first U))       % number or function
        (cond ((IDP OpLen) (setq OpLen (list 'function OpLen))))))) 

    % tear apart form, now put together result
    (return (EvDefOpcode OpName OpValue OpFn OpSize OpLen))))


(de EvDefOpcode (OpName OpValue OpFn OpSize OpLen) 
% Build a procedure to put the values on the opcode's property list
% The properties are: OpcodeValue
%                     InstructionDepositFunction
%                     Size
%                     InstructionLength
(prog (U) 
    (setq U (list (list 'put OpName ''OpCodeValue OpValue))) 
    (setq U (cons (list 'put OpName ''InstructionDepositFunction OpFn) U)) 
    (setq U (cons (list 'put OpName ''Size OpSize) U)) 
    (cond (OpLen 
        (setq U (cons (list 'put OpName ''InstructionLength OpLen) U)))) 
    (return (cons 'progn U)))) 


(dm DefOpcodes (U) 
%
% (DefOpcodes name number fn)
% (DefOpcodes name number fn length)
%
% Similar in form to DefOpcode, except defines all size variants, i.e.
% DefOpcodes(foo, ...), defines foo.b foo.w and foo.l.  Generic foo is
% defined to be the same as foo.w.  The exception here is those opcodes
% which are defined to use the Branch function.  These will only be defined
% in the foo.b and foo.w variants, and the generic version will be defined
% as equivalent to the byte version (this is for the benefit of the
% branch optimization phase).  Also, the WordBranch property is added for
% the generic instruction, as (put 'foo 'WordBranch 'foo.w).
%
(prog (OpName OpValue OpFn OpSize OpLen MapAList Result IsBranch) 
  (setq U (rest U)) 
  (setq OpName (first U))             % name, no surrounding quote
  (setq OpValue (second U)) 
  (setq OpFn (third U)) 
  
  (setq U (cdr (cdr U)))              % Skip first 2 args
  
  (cond ((setq IsBranch (eq OpFn 'Branch))                        % scs 
	 (progn (setq OpSize (const byte)) 
		(setq Result 
		  (cons (list 'put (MkQuote OpName) ''WordBranch (MkQuote (Intern (Concat (ID2String OpName) ".W")))) Result)) 
		(setq MapAList '((".B"  byte . 2) (".S" byte . 2) (".W"  Word . 4))))) 
	(t 
	 (progn (setq OpSize (const Word)) 
		(setq MapAList '((".B" . byte) (".W" . Word) (".L" . Long)))))) 
  (setq OpFn (list 'function OpFn))   % (function name)
  (setq U (rest U)) 

    (setq OpLen nil)                    % length computation is "standard"
                                        % by default, see function
                                        % StandardInstructionLength.

    (cond (U 
    (progn (setq OpLen (first U))       % number or function
        (cond ((IDP OpLen) (setq OpLen (list 'function OpLen))))))) 
    % tear apart form, now put together result
    (setq Result (cons (EvDefOpcode (MkQuote OpName) OpValue OpFn OpSize OpLen) 
                Result)) 
    (setq OpName (ID2String OpName)) 
    (cond (IsBranch                                           %  scs
	   (foreach X in MapAList do 
		    (setq Result (cons (EvDefOpcode 
					(MkQuote (Intern (Concat OpName (car X)))) 
					OpValue 
					OpFn 
					(list 'const (cadr X)) 
					(cddr X)) Result))))
	  (t (foreach X in MapAList do 
		      (setq Result (cons (EvDefOpcode 
					  (MkQuote (Intern (Concat OpName (car X)))) 
					  OpValue 
					  OpFn 
					  (list 'const (cdr X)) 
					  OpLen) Result)))))
    (return (cons 'progn Result)))) 


(dm DefCCOpcodes (U) 
%
% (DefCCOpcodes name number assemble-fn op-size length)
%
% This one defines all variants of an instruction using the condition codes.
% If the instruction uses the Branch function, word and byte versions are
% generated, and the WordBranch property is added.
%
(prog (OpName OpValue OpFn OpSize OpLen MapAList 
             ThisOpName ThisOpValue Result) 
    (setq U (rest U)) 
    (setq OpName (first U))             % name, no surrounding quote
    (setq OpValue (second U)) 
    (setq OpFn (third U)) 

    (setq U (cdr (cdr U)))              % skip first 2 args

    (cond ((equal OpFn 'Branch) 
    (progn (setq OpSize (const byte)) 
        (setq MapAList '((".B" byte . 2) (".S" byte . 2) (".W" Word . 4))))) (t 
    
    (progn (setq OpSize (const Word)) 
        (setq MapAList nil))))          % no variants are created, except branch
    (setq OpFn (list 'function OpFn))   % (function name)
    (setq U (rest U)) 

    (cond (U 
    (progn (setq OpSize (first U)) 
        (cond ((IDP OpSize) 
            (setq OpSize (list 'const OpSize)))) 	% (const size)
        (setq U (rest U))))) 
    (setq OpLen nil)                    % length computation is "standard"
                                        % by default, see function
                                        % StandardInstructionLength.

    (cond (U 
    (progn (setq OpLen (first U))       % number or function
        (cond ((IDP OpLen) (setq OpLen (list 'function OpLen))))))) 
    % tear apart form, now put together result
    (setq OpName (ID2String OpName)) 
                                        % Names from Table A-2 of M68K manual
    (foreach Y in '(("CC" . 4)          % Carry Clear
                    ("CS" . 5)          % Carry Set
                    ("EQ" . 7)          % EQual
                    ("F" . 1)           % False
                    ("GE" . 12)         % Greater than or Equal
                    ("GT" . 14)         % Greater Than
                    ("HI" . 2)          % HIgh
                    ("LE" . 15)         % Less than or Equal
                    ("LS" . 3)          % Low or Same
                    ("LT" . 13)         % Less Than
                    ("MI" . 11)         % MInus
                    ("NE" . 6)          % Not Equal
                    ("PL" . 10)         % PLus
                    ("T" . 0)           % True
                    ("VC" . 8)          % overflow Clear
                    ("VS" . 9)) do      % overflow Set
    (progn (setq ThisOpName (Concat OpName (car Y))) 
        (setq ThisOpValue (LOR OpValue (LSH (cdr Y) 8))) 
        (cond ((equal OpName "B")       % branch instruction has the form Bcc
        (setq Result (cons (list 'put (MkQuote (Intern ThisOpName)) 
                             ''WordBranch 
                             (MkQuote (Intern (Concat ThisOpName ".W")))) 
                   Result)))) 
        (setq Result (cons (EvDefOpcode (MkQuote (Intern ThisOpName)) 
                         ThisOpValue 
                         OpFn 
                         OpSize 
                         OpLen) Result)) 
        (foreach X in MapAList do 
            (setq Result (cons (EvDefOpcode 
				(MkQuote (Intern (Concat ThisOpName (car X)))) 
				ThisOpValue 
				OpFn 
				(list 'const (cadr X)) 
				(cddr X)) Result))))) 
    (return (cons 'progn Result)))) nil))




(de EffectiveAddress (Operand) 
%
% Purpose: convert an operand from symbolic to numeric form.
% Returns: Addressing mode in the range 0..7
% --------------------------------------------------
% M68K addressing modes (from appendix B of the M68K User's Manual)
% Addressing Mode         Mode  Reg        Valid Modes*         Assembler
%                                       Data MEM Cont Alter      Syntax
% Data Register Direct    000   reg no.   X   -   -    X           Dn
% Address Register Direct 001   reg no.   -   -   -    X           An
% Addr Reg Indirect       010   reg no.   X   X   X    X          (An)
%  with PostIncrement     011   reg no.   X   X   -    X          (An)+
%  with PreDecrement      100   reg no.   X   X   -    X         -(An)
%  with Displacement      101   reg no.   X   X   X    X         d(An)
%  with Index             110   reg no.   X   X   X    X         d(An,Ri)
% Absolute Short          111   000       X   X   X    X          xxxx
% Absolute Long           111   001       X   X   X    X        xxxxxxxx
% PC with Displacement    111   010       X   X   X    -         d(PC)
% PC with Index           111   011       X   X   X    -         d(PC,Ri)
% Immediate               111   100       X   X   -    -        #xxxxxxxx
%
% * = Valid Addressing modes for each type of Addressing Category
% Data              - used to refer to data operands
% Mem   = Memory    - used to refer to memory operands
% Cont  = Control   - used to refer to memory operands without an associated
%                     size
% Alter = Alterable - used to refer to alterable (writeable) operands
% --------------------------------------------------
% Operand is of the form:
%
% case 1:  numeric                 immediate data
%       or (immediate x)
% case 2: non-numeric atom         a local label, which uses PC with
%                                  displacement
% case 3: (reg x)                  x is a number or symbolic register name
% case 4: (indirect (reg x))       address register indirect in Motorola jargon
% case 5: (postincrement (reg x))  address register indirect with postincrement
% case 6: (predecrement (reg x))   address register indirect with predecrement
% case 7: (displacement (reg x) n) if (reg x) is an A reg
%                                    then if n is 0
%                                           then (indirect (reg x))
%                                           else address register indirect
%                                                 with displacement
%                                     else if (reg x) is a D reg
%                                            then address register indirect
%                                                   with index, using A6 (zero)
% case 8: (indexed (reg x) (displacement (reg y) n))
%                       address register indirect with index
%
% case 9+: various Lisp addressing modes, all of which are absolute long
%                                         addresses
%
% The value returned by this function is the mode field of the instruction
% for the operand.
% In addition, the fluid variables OperandRegisterNumber!*
%                              and OperandExtension!*
% will be set.
% If there are no words to follow, OperandExtension!* will be set to NIL.
% Otherwise, possible values of    OperandExtension!* are:
%
%       number or (immediate exp)  immediate data
%       (number)                   16-bit signed displacement
%       non-numeric atom           pc relative label
%       (displacement reg disp)    index extension word
%       other                      absolute long, i.e. LISP addressing mode
%
(prog (OpFn) 
    (setq OperandExtension* nil)        % off unless extra words are needed

    (cond                               % immediate data
      ((or (NumberP Operand) (EqCar Operand 'Immediate)) 
        (progn (setq OperandRegisterNumber* 4) 
                 (setq OperandExtension* Operand) 	% save it for extension
                 (return 7)))) 

    (cond ((atom Operand)               % non-numeric atom is a label
      (progn (setq OperandRegisterNumber* 2) 	% PC with displacement
               (setq OperandExtension* Operand) 	% save the label
               (return 7)))) 

    (cond ((setq OpFn (get (first Operand) 'OperandResolution)) 
      (return (Apply OpFn (list Operand)))) (t 	% non-atomic, dispatch on
                                                % the name
      (progn (setq OperandRegisterNumber* 1) 	% default case is absolute long
               (setq OperandExtension* Operand) 	% save it
               (return 7))))))


% Procedures to compute specific OperandRegisterNumber!*
% Each of the cases returns the Addrssing MODE
% and sets OperandRegisterNumber!* as a side effect

(de RegOperand (Operand)                % (reg x)
(progn (setq OperandRegisterNumber* (RegisterNumber (second Operand))) 
    (cond ((DregP Operand) 0) (t 1))))  % DRegP is in M68K-CMAC.SL
      % DReg mode
      % AReg mode

(de CompoundRegOperand (Operand)        % (mode (reg x))
(progn (setq OperandRegisterNumber* (RegisterNumber (second (second Operand)))) 
    (cdr (Assoc (first Operand) '((predecrement . 4) 
                               (postincrement . 3) 
                               (indirect . 2))))))

(de DisplacementOperand (Operand)       % (displacement (reg x) n)
(prog (reg Disp) 
    (setq reg (second Operand)) 
    (setq Disp (third Operand)) 

    (setq OperandRegisterNumber* (RegisterNumber (second reg))) 
    (return 
     (cond ((ARegP reg)                    % ARegP is in M68K-CMAC.SL
	    (cond ((equal Disp 0) 2) 
		  (t 
		   % indirect mode
		   (progn (setq OperandExtension* (cddr Operand)) 5)))) 
	   (t 	% save (Disp) for later
                % displacement mode
		% funny case, D reg displacement, use indexing with zero A reg
		(progn (setq OperandRegisterNumber* 6) 	% A6 is zero register
		       (setq OperandExtension* Operand) 6))))))
                                   % save whole thing for extendindex mode
    



(de IndexedOperand (Operand)            % (indexed (reg x) (disp.. (reg y) n))
%
% This one has to CONS, but it's so rare that there's no point in
% worrying about it.  It's only used in some funny OpenCodes
%
(progn 
  (setq OperandRegisterNumber* 
	(RegisterNumber (second (second (third Operand))))) 
  (setq OperandExtension*             % (displacement ireg disp)
        (cons 'Displacement (cons (second Operand) 
				  (rest (rest (third Operand)))))) 6))
    

% ------------------------------------------------------------
% Define all of the OperandResolution functions
% ------------------------------------------------------------

(put 'reg 'OperandResolution (function RegOperand))
(put 'Displacement 'OperandResolution (function DisplacementOperand))
(put 'indirect 'OperandResolution (function CompoundRegOperand))
(put 'postincrement 'OperandResolution (function CompoundRegOperand))
(put 'predecrement 'OperandResolution (function CompoundRegOperand))
(put 'Indexed 'OperandResolution (function IndexedOperand))

% ------------------------------------------------------------
% Procedures to compute OperandExtension!*
% ------------------------------------------------------------

(de DepositExtension (Exp) 
%
% See comments in EffectiveAddress concerning OperandExtension!* for
% description of Exp.
%
      (cond ((null Exp) nil)            % NIL means no extension
 ((NumberP Exp) 
        (cond ((eq InstructionSize* (const Long)) 
               (DepositWord (Int2Sys Exp))) (t 
               (DepositHalfWord Exp)))) 
 ((atom Exp)                            % PC Relative Label
        (DepositHalfWord (MakeExpressionRelative Exp 0))) 
 ((NumberP (first Exp))                 % 16-bit disp
        (DepositHalfWord (first Exp))) 

 (t 
  (Apply 
   (or (get (first Exp) 'OperandDepositFunction) 
       (StdError (BldMsg "Unknown operand %r" Exp))) (list Exp))))) 
 %  what is this??

(de DepositIndexExtensionWord (Exp)     % (displacement (REG x)  disp)
(prog (reg Displacement) 
    (setq reg (second Exp)) 
    (setq Displacement (third Exp))     % displacement

    (DepositHalfWord 
                        
                        (plus (cond ((ARegP reg) 34816) (t 2048)) 	% D/A, always long

                        (LSH (RegisterNumber (second reg)) 12) 	% Register
                        (LAND Displacement 255)))))	% 8-bit dsp



(de DepositImmediate (Exp)              % (immediate n) or n
(progn (cond ((not (NumberP Exp)) (setq Exp (second Exp)))) 	% strip the 'immediate tag
    (cond ((eq InstructionSize* (const Long)) 
      (DepositWordExpression Exp)) (t 
      (DepositHalfWordExpression Exp)))))


(de DepositFluid (X) 
    (DepositValueCellLocation (second X)))	% Defined in System-Faslin.Red

(de DepositExtraReg (X) 
    (DepositExtraRegLocation (second X)))	% Defined in System-Faslin.Red

(de DepositEntry (X) 
    (DepositFunctionCellLocation (second X)))	% Defined in System-Faslin.Red

(de depositforeignentry (x)
  (depositfunctioncelllocation (second x)))

(put 'Immediate 'OperandDepositFunction (function DepositImmediate))
(put 'Displacement 'OperandDepositFunction (function DepositIndexExtensionWord))
(put 'fluid 'OperandDepositFunction (function DepositFluid))
(put '$fluid 'OperandDepositFunction (function DepositFluid))
(put 'global 'OperandDepositFunction (function DepositFluid))
(put '$global 'OperandDepositFunction (function DepositFluid))
(put 'ExtraReg 'OperandDepositFunction (function DepositExtraReg))
(put 'entry 'OperandDepositFunction (function DepositEntry))
(put 'foreignentry 'operanddepositfunction (function depositforeignentry))

% Auxiliary functions used with the Instruction Depositing operations

(ds OpCodeValue (Instruction)           % (Opcode ..... )
 (get (first Instruction) 'OpCodeValue))

% --------------------------------------------------
% Instruction Deposit Functions
%
% The argument is the entire instruction
%
% --------------------------------------------------
% Notes:
%    the line LAP  ( ... ) shows what the incoming LAP instruction looks like
%    the line 68K instruction |xxxx| ... shows the words (16 bit sized
%     words) that are produced.
% --------------------------------------------------
% All of the forms for instructions
%
% --------------------------------------------------
% LAP    (opcode)
% LAP    (opcode immediate-data effective-address)
% LAP    (opcode effective-address)
% LAP    (opcode register)
% LAP    (opcode const register)
% LAP    (opcode reg label)
% LAP    (opcode 3-bit-immediate effective-address)
% LAP    (opcode label)
% LAP    (opcode address register)
% LAP    (opcode dreg address)
%    or (opcode address dreg)
% LAP    (opcode source (reg An))
% LAP    (opcode (reg x) (reg y))
%    or (opcode (predecrement (reg x)) (predecrement (reg y)))
% LAP    (op dreg dreg), or (op num dreg)
%    or (op ea)
% LAP    (move ea ea)
% LAP    (opcode Dn ea)
%    or (opcode (immediate n) ea)
% LAP    (opcode (postincrement (reg y)) (postincrement (reg x)))
% LAP    (opcode (RegList x y z ...) ea)
%    or (opcode (immediate n) ea)
%     (opcode ea (RegList x y z ...) )
%    or (opcode ea (immediate n) )
% LAP    (opcode (immediate xx) (reg Dn) )
% LAP    (opcode (reg Dn) ea)
%    or (opcode ea  (reg Dn) )
% LAP    (trap (Immediate x))
% LAP    (STOP datahalfword)
% --------------------------------------------------
% LAP             (opcode)
% 68K instruction |opcode|

(de ZeroOperandInstruction (Instr) 
    (DepositHalfWord (OpCodeValue Instr)))

(DefOpcode illegal 19196 ZeroOperandInstruction)	% 4AFC
(DefOpcode reset 20080 ZeroOperandInstruction)	% 4E70
(DefOpcode nop 20081 ZeroOperandInstruction)	% 4E71
(DefOpcode rte 20083 ZeroOperandInstruction)	% 4E73
(DefOpcode rts 20085 ZeroOperandInstruction)	% 4E75
(DefOpcode trapv 20086 ZeroOperandInstruction)	% 4E76
(DefOpcode rtr 20087 ZeroOperandInstruction)	% 4E77

% --------------------------------------------------
% LAP              (opcode immediate-data effective-address)
% 68K instruction  |opcode:mode:reg|
%                  |immediate-data|
%                  |address-extension|

(de ImmediateInstruction (Instr) 
(prog (ImmediateDataSource Destination Mode) 
    (setq ImmediateDataSource (second Instr)) 
    (setq Destination (third Instr)) 
    (setq Mode (EffectiveAddress Destination)) 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH InstructionSize* 6) 
                     (LSH Mode 3) 
                     OperandRegisterNumber*)) 
                   
                                        % immediate data goes first

    (DepositImmediate ImmediateDataSource) 	% 16 bits of data

    (DepositExtension OperandExtension*)))	% then destination extension


(DefOpcodes ori    0  ImmediateInstruction)     % 00xx
(DefOpcodes andi 512  ImmediateInstruction)	% 02xx
(DefOpcodes subi 1024 ImmediateInstruction)	% 04xx
(DefOpcodes addi 1536 ImmediateInstruction)	% 06xx
(DefOpcodes eori 2560 ImmediateInstruction)	% 0Axx
(DefOpcodes cmpi 3072 ImmediateInstruction)	% 0Cxx


% --------------------------------------------------
% LAP              (opcode effective-address)
% 68K instruction  |opcode:odd-size:mode:reg|
%                  |address-extension|

(de OneAddressWithSize (Instr) 
    (OneAddress Instr InstructionSize*))


% --------------------------------------------------
% LAP              (opcode effective-address)
% 68K instruction  |opcode:mode:reg|
%                  |address-extension|

(de OneAddressNoSize (Instr) 
% Note that the three instructions which use this mode NEVER have immediate
% data after them, but treat the "Immediate" numeric value as the address
% upon which to do their operation - i.e. it is either Absolute long or
% Absolute Word. Thus as a correction we will use the standard
% EffectiveAddress calculation routine but if the resulting MODE is 7 and
% the OperandRegisterNumber!* = 4, then we will use Absolute Word mode (0) if
% the operand OperandExtension!* will fit in 16 bits, otherwise we will use
% the Absolute Long mode (1). The information about which type of address
% (Word or Long) is to be inserted is communicated to DepositExtension by
% the fluid variable InstructionSize!*.
(prog (Mode) 
    (setq Mode (EffectiveAddress (second Instr))) 

    (cond ((equal Mode 7) 
      (cond ((equal OperandRegisterNumber* 4) 
             (progn                     % remove the 'Immediate tag if present
                     (cond ((EqCar OperandExtension* 'Immediate) 
                       (equal OperandExtension* (second OperandExtension*)))) 

                     (cond              % Less than -2^15
                      ((or (LessP OperandExtension* -32768) (GreaterP OperandExtension* 32767)) 	% Greater than 2^15
                       (progn           % Long mode must be used
                               (setq OperandRegisterNumber* 1) 
                               (setq InstructionSize* (const Long)) nil)) (t 
                            
                       (progn           % Word mode may be used
                               (setq OperandRegisterNumber* 0) 
                               (setq InstructionSize* (const Word)) nil))) nil))))) 
                            
                 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH Mode 3) 
                     OperandRegisterNumber*)) 
                   
    (DepositExtension OperandExtension*)))



(de OneAddress (Instr Size) 
(prog (Mode) 
    (setq Mode (EffectiveAddress (second Instr))) 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH Size 6) 
                     (LSH Mode 3) 
                     OperandRegisterNumber*)) 
                   
    (DepositExtension OperandExtension*)))


(DefOpcodes negx 16384 OneAddressWithSize)	% 40xx
(DefOpcodes clr 16896 OneAddressWithSize)	% 42xx
(DefOpcodes neg 17408 OneAddressWithSize)	% 44xx
(DefOpcodes not 17920 OneAddressWithSize)	% 46xx
(DefOpcodes tst 18944 OneAddressWithSize)	% 4Axx

(DefOpcode nbcd 18432 OneAddressNoSize byte)	% 48yx
(DefOpcode pea 18496 OneAddressNoSize Long)	% 484x
(DefOpcode tas 19136 OneAddressNoSize byte)	% 4ACx
(DefOpcode jsr 20096 OneAddressNoSize)  % 4E8x
(DefOpcode jmp 20160 OneAddressNoSize)  % 4ECx

% the Scc instructions
(DefCCOpcodes s 20672 OneAddressNoSize byte)	% 5xCx

% --------------------------------------------------
% LAP              (opcode register)
% 68K instruction  |opcode:reg|

(de OneRegister (Instr) 
    (OneRegisterAux Instr (second Instr)))


% --------------------------------------------------
% LAP              (opcode const register)
% 68K instruction  |opcode:reg|
%                  |constant|

(de ConstOneRegister (Instr)            % LINK is the only one here
(progn (OneRegisterAux Instr (third Instr)) 
    (DepositExtension (second Instr))))


% --------------------------------------------------
% LAP              (opcode reg label)
% 68K instruction  |opcode:reg|
%                  |pc-relative-word|

(de OneRegisterLabel (Instr)            % DBcc is the only instance
(progn (OneRegister Instr) 
    (DepositHalfWord (MakeExpressionRelative (third Instr) 0))))

(de OneRegisterAux (Instr reg) 
    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (RegisterNumber (second reg)))))
                   

(DefOpcode swap 18496 OneRegister Word 2)	% 484x
(DefOpcode ext!.w 18560 OneRegister Word 2)	% 488x
(DefOpcode ext!.l 18624 OneRegister Long 2)	% 48Cx
(DefOpcode unlk 20056 OneRegister Word 2)	% 4E58

(DefOpcode link 20048 ConstOneRegister Word 4)	% 4E5x

(DefCCOpcodes db 20680 OneRegisterLabel Word 4)	% 5xC8


% --------------------------------------------------
% LAP              (opcode 3-bit-immediate effective-address)
% 68K instruction  |opcode:data:opcode:size:mode:reg|
%                  |address-extension|

(de QuickArithmetic (Instr) 
(prog (Constant Mode) 
    (setq Constant (second Instr)) 
    (cond ((PairP Constant) 
      (setq Constant (second Constant)))) 	% (immediate n)
% needs to test the size of the constant to avoid destroying the opcode
    (cond ((equal Constant 8) (setq Constant 0))) 	% M68K convention for quick immediate

    (setq Mode (EffectiveAddress (third Instr))) 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH (LAND Constant 7) 9) 	% 3 bit immediate field
                     (LSH InstructionSize* 6) 
                     (LSH Mode 3) 
                     OperandRegisterNumber*)) 
                   
    (DepositExtension OperandExtension*)))


(de QuickLength (X) 
    (Plus2 (OperandLength (third X)) 2))

(DefOpcodes addq 20480 QuickArithmetic QuickLength)	% 5xxx
(DefOpcodes subq 20736 QuickArithmetic QuickLength)	% 51xx


% --------------------------------------------------
% LAP              (opcode label)
% 68K instruction  |opcode|
%                  |pc-relative-word|
% or
% 68K instruction  |opcode:pc-relative-byte|

(de Branch (Instr) 
  (cond ((eq InstructionSize* (const byte)) 
    (DepositHalfWord (plus (OpCodeValue Instr) 
                          (LAND (MakeExpressionRelative (second Instr) 2) 255)))) (t 
                               % 8 bits
    
    (progn (DepositHalfWord (OpCodeValue Instr)) 
        (DepositHalfWord (MakeExpressionRelative (second Instr) 0))))))


(DefCCOpcodes b 24576 Branch)           % 6xxx
(DefOpcodes bra 24576 Branch)           % 60xx
(DefOpcodes bsr 24832 Branch)           % 61xx


% --------------------------------------------------
% LAP              (opcode address register)
% 68K instruction  |opcode:reg:opcode:mode:reg|
%                  |address-extension|

(de RegisterAddress (Instr) 
    (RegisterAddressAux (OpCodeValue Instr) 
                       (third Instr)    % register
                       (second Instr) 0))	% address
                       % direction and size

% --------------------------------------------------
% LAP              (opcode dreg address)
%               or (opcode address dreg)
% 68K instruction  |opcode:reg:op-mode:mode:reg|
%                  |address-extension|
% You must be very careful of the allowed modes when generating code for
% this instruction.

(de RegisterSizeAddress (Instr)         % Only CMP goes here
    (RegisterAddressAux (OpCodeValue Instr) 
                       (third Instr)    % register
                       (second Instr)   % address
                       InstructionSize*))	% size


(de RegisterStrangeSizeAddress (Instr)  % ADDA, CMPA, SUBA
    (RegisterAddressAux (OpCodeValue Instr) 
                       (third Instr)    % register
                       (second Instr)   % address
                       (cond ((equal InstructionSize* (const Long)) 7) (t 3))))
                                                         % size
   % Note that a BYTE sized version is ILLEGAL and should be tested for here.

(de RegisterOpModeAddress (Instr) 
(prog (Source Destination) 
    (setq Source (second Instr)) 
    (setq Destination (third Instr)) 

    (return (cond ((DregP Destination) 
      % if Destination is a Dreg, then any source is OK
        (RegisterAddressAux (OpCodeValue Instr) 
                           Destination  % register
                           Source       % address
                           InstructionSize*)) (t 	% direction(=0) and size
      % source is a Dreg, destination must be memory alterable
        (RegisterAddressAux (OpCodeValue Instr) 
                           Source       % register
                           Destination  % address
                           (LOR 4 InstructionSize*)))))))	% direction and size



(de DRegisterEORAddress (Instr) 
(prog (Source Destination) 
    (setq Source (second Instr)) 
    (setq Destination (third Instr)) 

    (return (cond ((DregP Source) 
      % source is a Dreg, destination must be memory alterable
        (RegisterAddressAux (OpCodeValue Instr) 
                           Source       % register
                           Destination  % address
                           (LOR 4 InstructionSize*))) (t 	% direction and size
      (StdError " EOR must have a DREG in the source field "))))))



(de RegisterAddressAux (Result reg Addr OpMode) 
(progn                                  % with reg = (Reg  n)
    (DepositHalfWord (plus Result 
                     (LSH (RegisterNumber (second reg)) 9) 
                     (LSH OpMode 6) 
                     (LSH (EffectiveAddress Addr) 3) 
                     OperandRegisterNumber*)) 
                   
    (DepositExtension OperandExtension*)))

(DefOpcode chk 16768 RegisterAddress Word)	% 418x
(DefOpcode lea 16832 RegisterAddress Long)	% 41Cx
(DefOpcode divu 32960 RegisterAddress Word)	% 80Cx
(DefOpcode divs 33216 RegisterAddress Word)	% 81Cx
(DefOpcode mulu 49344 RegisterAddress Word)	% C0Cx
(DefOpcode muls 49600 RegisterAddress Word)	% C1Cx

(DefOpcodes or 32768 RegisterOpModeAddress)	% 8xxx
(DefOpcodes sub 36864 RegisterOpModeAddress)	% 9xxx
(DefOpcodes cmp 45056 RegisterSizeAddress)	% Bxxx
(DefOpcodes eor 45056 DRegisterEORAddress)	% Bxxx
(DefOpcodes and 49152 RegisterOpModeAddress)	% Cxxx
(DefOpcodes add 53248 RegisterOpModeAddress)	% Dxxx

(DefOpcodes suba 36864 RegisterStrangeSizeAddress)	% 9xxx
(DefOpcodes cmpa 45056 RegisterStrangeSizeAddress)	% Bxxx
(DefOpcodes adda 53248 RegisterStrangeSizeAddress)	% Dxxx


% --------------------------------------------------
% LAP              (opcode source (reg An))
% 68K instruction  |opcode:size:reg:2#001:mode:reg|
%                  |address-extension|

(de movea (Instr) 
(prog (Mode reg Size) 
    (setq Size (cond ((equal InstructionSize* (const Long)) 2) (t 3))) 
% Error if BYTE size

    (setq reg (third Instr)) 
    (setq Mode (EffectiveAddress (second Instr))) 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH Size 12) 
                     (LSH (RegisterNumber (second reg)) 9) 
                     (LSH Mode 3) 
                     OperandRegisterNumber*)) 
                   
    (DepositExtension OperandExtension*)))


(DefOpcodes movea 64 movea)             % 0040


% --------------------------------------------------
% LAP              (opcode (reg x) (reg y))
% 68K instruction  |opcode:regx:opmode:regy|

(de exg (Instr)                         % exchange two registers
(prog (OpMode Reg1 Reg2 RegT) 
    (setq Reg1 (second Instr)) 
    (setq Reg2 (third Instr)) 
    (setq OpMode (cond ((DregP Reg1) 
                (cond ((DregP Reg2) 8) (t 17))) 
                       % Dreg Dreg
                       % Dreg Areg
                ((DregP Reg2) 
                       (progn           % Areg Dreg - must switch the registers as
                               % Reg2 must be an Areg if mixed mode
                               (setq RegT Reg1) 
                               (setq Reg1 Reg2) 
                               (setq Reg2 RegT) 
                               (return 17))) (t 9))) 
                            
                       % Areg Areg
    (return (TwoRegisterAuxAux Instr 
                             Reg1 
                             Reg2 0 
                             
                             OpMode))))


% --------------------------------------------------
% LAP              (opcode (reg x) (reg y))
%               or (opcode (predecrement (reg x)) (predecrement (reg y)))
% 68K instruction  |opcode:regx:0:8+regy|

(de ExtendArithmetic (Instr) 
    (TwoRegisterAux Instr InstructionSize*))

(de BCDTwoRegister (Instr) 
    (TwoRegisterAux Instr 0))

(de TwoRegisterAux (Instr Size) 
(prog (RM Reg1 Reg2 regs) 
    (setq regs (rest Instr)) 
    (setq Reg1 (first regs)) 
    (cond ((DregP Reg1)                 % (opcode (reg x) (reg y))
    (progn (setq RM 0) 
        (setq Reg2 (second regs)))) (t 
    % (opcode (predecrement (reg x))
                                        %         (predecrement (reg y)))
        (progn (setq RM 1) 
        (setq Reg1 (second Reg1)) 
        (setq Reg2 (second (second Reg2)))))) 
    (return (TwoRegisterAuxAux Instr 
                             Reg2 
                             Reg1 
                             Size 
                             RM))))


(de TwoRegisterAuxAux (Instr Reg1 Reg2 Size RM) 
    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH (RegisterNumber (second Reg1)) 9) 
                     (LSH Size 6) 
                     (LSH RM 3) 
                     (RegisterNumber (second Reg2)))))
                   

(DefOpcode abcd 49408 BCDTwoRegister byte)	% C10x
(DefOpcode sbcd 33024 BCDTwoRegister byte)	% 810x

(DefOpcodes addx 53504 ExtendArithmetic)% D1xx
(DefOpcodes subx 37120 ExtendArithmetic)% 91xx

(DefOpcode exg 49408 exg Long)          % C1xx

% --------------------------------------------------
% LAP              (op dreg dreg), or (op num dreg)
%               or (op ea)
% 68K instruction  |opcode:count/register:dr:size:i/r:opcode:register|
%               or |opcode:0:type:dr:11:EA|

(de ShiftAndRotate (Instr) 
(prog (Opcode ShiftCount DestReg IR Mode Result) 
    (setq Opcode (OpCodeValue Instr)) 
    (setq ShiftCount (second Instr)) 

    (cond ((null (cdr (cdr Instr)))     % memory shift
    (progn (setq Mode (EffectiveAddress ShiftCount)) 
% Needs to have the Opcode in bits 10 and 9, but it has lost them at this
% point. The opcode patterns are: AS = 00, LS = 01, ROX = 10, and RO = 11 .
        (DepositHalfWord (plus (LAND Opcode 65472) 
                         (LSH Mode 3)   % mode
                         OperandRegisterNumber*)) 
                       
        (return (DepositExtension OperandExtension*))))) 

    (setq DestReg (third Instr)) 
    (cond ((DregP ShiftCount) 
    (progn (setq ShiftCount (RegisterNumber (second ShiftCount))) 
        (setq IR 32))) (t               % register holds shift count
    
    (progn (cond ((PairP ShiftCount) (setq ShiftCount (second ShiftCount)))) 	% (immed. x)
        (cond ((equal ShiftCount 8) (setq ShiftCount 0))) 
        (setq IR 0))))                  % immediate shift count

    (return (DepositHalfWord (plus (LAND Opcode 61759) 
                            (LSH ShiftCount 9) 
                            (LSH InstructionSize* 6) 
                            IR 
                            (RegisterNumber (second DestReg)))))))
                          


(de ShiftLength (Instr) 
    (cond ((or (FixP (second Instr)) (EqCar (second Instr) 'Immediate)) 2) (t 
      
      (StandardInstructionLength Instr))))

(DefOpcodes asl 57792 ShiftAndRotate ShiftLength)
(DefOpcodes asr 57536 ShiftAndRotate ShiftLength)
(DefOpcodes lsl 57800 ShiftAndRotate ShiftLength)
(DefOpcodes lsr 57544 ShiftAndRotate ShiftLength)
(DefOpcodes roxl 57808 ShiftAndRotate ShiftLength)
(DefOpcodes roxr 57552 ShiftAndRotate ShiftLength)
(DefOpcodes rol 57816 ShiftAndRotate ShiftLength)
(DefOpcodes ror 57560 ShiftAndRotate ShiftLength)

% --------------------------------------------------
% LAP              (move ea ea)
% 68K instruction  |opcode:size:dest-reg:dest-mode:source-mode:source-reg|

(de LAPMOVE (Instr) 
(prog (SourceMode SourceExtension SourceRegisterNumber DestMode) 
    (setq SourceMode (EffectiveAddress (second Instr))) 
    (setq SourceExtension OperandExtension*) 	% save the extension
    (setq SourceRegisterNumber OperandRegisterNumber*) 	% save source register

    (setq DestMode (EffectiveAddress (third Instr))) 

    (setq InstructionSize* (cond ((equal InstructionSize* (const byte)) 1) 
                           ((equal InstructionSize* (const Word)) 3) 
                           ((equal InstructionSize* (const Long)) 2))) 
                           % MOVE doesn't use standard values in size field.

    (DepositHalfWord (plus (LSH InstructionSize* 12) 
                     (LSH OperandRegisterNumber* 9) 
                     (LSH DestMode 6) 
                     (LSH SourceMode 3) 
                     SourceRegisterNumber)) 
                   
    (DepositExtension SourceExtension) 
    (DepositExtension OperandExtension*)))	% dest extension


(DefOpcodes move 0 LAPMOVE)             % 0xxx

% LAP             (STOP datahalfword)
% M68K instruction |STOP|
%                  |datahalfword|
(de stop (Instr) 
(progn (DepositHalfWord (OpCodeValue Instr)) 
    (DepositHalfWord (cond ((FixP (second Instr)) (second Instr)) (t 
                                         (second (second Instr)))))))

(DefOpcode stop 20082 stop Word 4)      % 4E72
% The size should be 4 as it is 2 bytes for the opcode + 2 for the data word
% this is really an UNsized mode opcode

% --------------------------------------------------
% LAP              (opcode Dn ea)
%               or (opcode (immediate n) ea)
% 68K instruction  |opcode:reg:opcode:EA|
%               or |opcode:reg:opcode:EA|
%                  |bit number|

(de BitOperation (Instr) 
(prog (Mode Source) 

    (setq Source (second Instr)) 
    (setq Mode (EffectiveAddress (third Instr))) 

  (cond ((DregP Source) 
    % Take care of the Dynamic Case
             (DepositHalfWord (plus (OpCodeValue Instr) 
                              (LSH (RegisterNumber (second Source)) 9) 
                              (LSH 1 8) 
                              (LSH Mode 3) 
                              OperandRegisterNumber*))) (t 
                            
     (progn                             % Handle the static case.
             (DepositHalfWord (plus (OpCodeValue Instr) 
                              (LSH 8 8) 
                              (LSH Mode 3) 
                              OperandRegisterNumber*)) 
                            
             (DepositHalfWordExpression Source) nil))) 
          

   (DepositExtension OperandExtension*)))



(DefOpcode BTST!.B 0 BitOperation Word)
(DefOpcode BTST!.L 0 BitOperation Word)
(DefOpcode BTST 0 BitOperation Word)
(DefOpcode BCHG!.B 64 BitOperation Word)
(DefOpcode BCHG!.L 64 BitOperation Word)
(DefOpcode BCLR!.B 128 BitOperation Word)
(DefOpcode BCLR!.L 128 BitOperation Word)
(DefOpcode BSET!.B 192 BitOperation Word)
(DefOpcode BSET!.L 192 BitOperation Word)

% --------------------------------------------------
% LAP              (opcode (postincrement (reg y)) (postincrement (reg x)))
% 68K instruction  |opcode:regx:1:size:8+regy|
% ------------------------------
%   Special Instruction Case 14
%  CMPM
%  "A silly instruction if ever there was one!" -- Ian Wilson
%  The operand types are:
%
%    (Ay)+,(Ax)+

(de CMPM (Instr) 
(prog (Reg1 Reg2) 
    (setq Reg1 (second Instr)) 
    (setq Reg2 (third Instr)) 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH (RegisterNumber (second Reg1)) 9) 
                     (LSH InstructionSize* 6) 
                     (RegisterNumber (second Reg2))))))
                   


(DefOpcodes CMPM 45320 CMPM)            % B14x

% --------------------------------------------------
% LAP              (opcode (RegList x y z ...)  ea)
%               or (opcode (immediate n) ea)
%                  (opcode ea (RegList x y z ...) )
%               or (opcode ea (immediate n) )
% 68K instruction  |opcode:dr:001:Sz:EA|
%                  |RegisterListMask|

(de RegListP (Operand) (EqCar Operand 'RegList))


(de RegBit (register Mode)                   % Reg:id, mode: [0..7]
(prog (Regtype RegNumber)
  (if (pairp register)  (setq regtype (regoperand register))
                   (setq regtype (regoperand (list 'reg register))))
  (setq regnumber operandregisternumber*)

(return 
 (cond ((equal Mode 4)                  % predecrement mode
         (LSH 1 (cond ((equal RegType 0) (Difference 15 RegNumber)) 
                      (t                 (Difference 7 RegNumber)))))
       (t   (LSH 1 (cond ((equal RegType 0) RegNumber) 
                         (t                 (plus 8 RegNumber)))))))))


(de registerlistmask (operand mode)
  (prog (regmaskbits)
        (cond ((eqcar operand 'immediate) (return (second operand)))
              % (immediate n)
              (t (setq regmaskbits 0) % (RegList id id id ...)
                 (foreach reg in (cdr operand) do 
                  (setq regmaskbits (plus regmaskbits (regbit reg mode))))
                 (return regmaskbits)))))
       



(de movem (instr)
  (prog (mode source destination size)
        (setq source (second instr))
        (setq destination (third instr))
        (setq size (if (equal instructionsize* (const long))
                1
                0))
        % here BYTE sized mode is an error
        (if (or (eqcar source 'immediate) (reglistp source))
          (progn
            (setq mode (effectiveaddress destination))
            (deposithalfword 
             (plus (opcodevalue instr) (lsh size 6) (lsh mode 3) 
                   operandregisternumber*))
            % This is a quick attempt to trick the faulty logic in this code.
            % The use if the fluid InstructionSize* is WRONG.. Instuction size
            % relates to the mode of the opcode and not to the number of words the
            % assembled instruction will require for space.
            % The assumption was made that a movem!.l always deposits a 32 bit word..
            (if (neq mode 7)
              (setq instructionsize* 1))
            (depositimmediate (registerlistmask source mode))
            nil)
          (progn
            (setq mode (effectiveaddress source))
            (deposithalfword 
             (plus (opcodevalue instr) (lsh 1 10) % predecrement version
                   (lsh size 6) (lsh mode 3) operandregisternumber*))
            (if (neq mode 7)
              (setq instructionsize* 1))
            (depositimmediate (registerlistmask destination mode))
            nil))
        (depositextension operandextension*)))

(DefOpcodes movem 18560 movem)        % 488x

% --------------------------------------------------
% LAP              (opcode (immediate xx)  (reg Dn) )
% 68K instruction  |opcode:reg:0:xx|

(de moveq (Instr) 
(prog (Destination ImmediateSource) 

    (setq Destination (third Instr)) 
    (setq ImmediateSource (second Instr)) 
    (cond ((PairP ImmediateSource) (setq ImmediateSource (second ImmediateSource)))) 
    
    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LSH (RegisterNumber (second Destination)) 9) 
                     (LAND 255 ImmediateSource)))))
                   


(DefOpcode moveq 28672 moveq Word 2)    % 7xxx

% --------------------------------------------------
% LAP              (opcode (reg Dn) ea)
%               or (opcode  ea      (reg Dn) )
% 68K instruction  |opcode:Dreg:OpMode:001:Areg|
%                  |displacement|

(de MOVEP (Instr) 
(prog (Mode OpMode Source Destination) 

    (setq Source (second Instr)) 
    (setq Destination (third Instr)) 


  (cond ((DregP Source) 
    (progn (setq Mode (EffectiveAddress Destination)) 
             (cond ((equal InstructionSize* (const Long)) 
               (setq OpMode 7)) (t 6)) 
               % BYTE size is and ERROR
             (DepositHalfWord (plus (OpCodeValue Instr) 
                              (LSH (RegisterNumber (second Source)) 9) 
                              (LSH OpMode 6) 
                              OperandRegisterNumber*)))) (t 
                            
          
     (progn (setq Mode (EffectiveAddress Source)) 
             (cond ((equal InstructionSize* (const Long)) 
               (setq OpMode 5)) (t 4)) 
               % BYTE size is and ERROR
             (DepositHalfWord (plus (OpCodeValue Instr) 
                              (LSH (RegisterNumber (second Destination)) 9) 
                              (LSH OpMode 6) 
                              OperandRegisterNumber*)) 
                            
             (DepositImmediate Source) nil))) 
          

   (DepositExtension OperandExtension*)))



(DefOpcode movep!.w 8 MOVEP Word)       % 0xx8
(DefOpcode movep!.L 8 MOVEP Long)       % 0xx8

% --------------------------------------------------
% LAP              (trap (Immediate x))
% 68K instruction  |opcode:x|

(de trap (Instr) 
(prog (TrapVector) 

    (setq TrapVector (second Instr)) 

    (DepositHalfWord (plus (OpCodeValue Instr) 
                     (LAND 15 (second TrapVector))))))
                   


(DefOpcode trap 20032 trap)             % 4E4x

% ------------------------------------------------------------
% Deposit instructions for Pseudo ops
% ------------------------------------------------------------

(de DepositWordBlock (X)                % (FULLWORD xxx xxx ... xxx)
    (foreach Y in (cdr X) do (DepositWordExpression Y)))

(de DepositHalfWordBlock (X)            % (HALFWORD xxx xxx ... xxx)
    (foreach Y in (cdr X) do (DepositHalfWordExpression Y)))

(de DepositByteBlock (X)                % (BYTE     xxx xxx ... xxx)
(prog (Z) 
    (setq Z 0) 
    (foreach Y in (cdr X) do 
    (progn (DepositByte Y) 
        (setq Z (LXOR Z 1))))           % toggle Z
    (cond ((not (equal Z 0)) (DepositByte 0)))))	% go to halfword boundary


% Deposit a string with a trailing ZERO byte

(de DepositString (X)                   % (STRING "xxxxxx")
(prog nil 
    (setq X (second X)) 
    (for (from I 0 (Size X) 1) (do (DepositByte (Indx X I)))) 
    (DepositByte 0) 
    (cond ((not (equal (Remainder (Size X) 2) 0)) (DepositByte 0)))))
% align to word boundary

(de DepositFloat (X)                    % this will not work in cross-assembly
(progn (setq X (FltInf (second X))) 
    (DepositWord (FloatHighOrder X)) 
    (DepositWord (FloatLowOrder X))))

(put 'fullword 'InstructionDepositFunction 'DepositWordBlock)
(put 'halfword 'InstructionDepositFunction 'DepositHalfWordBlock)
(put 'byte 'InstructionDepositFunction 'DepositByteBlock)
(put 'string 'InstructionDepositFunction 'DepositString)
(put 'float 'InstructionDepositFunction 'DepositFloat)

% Auxiliary functions for computing instruction bit patterns

(de MakeExpressionRelative (Exp OffsetFromHere) 
(prog (X Y Z) 
    (cond ((EqCar Exp 'InternalEntry) 
      (return (MakeInternalEntryRelative (second Exp) OffsetFromHere)))) 

    (cond ((not (LabelP Exp)) (return (StdError "Only labels can be relative")))) 

    (setq X (plus CurrentOffset* OffsetFromHere)) 
    (setq Y (LabelOffset Exp)) 
    (return (Difference Y X))))


(de makeinternalentryrelative (nam offsetfromhere)
  (prog (offset)
        (setq offset (atsoc nam labeloffsets*))
        (setq offset (if offset
                (cdr offset)
                (get nam 'internalentryoffset)))
        (return (if offset
                  (progn
                    (setq offset 
                     (difference offset 
                      (plus2 currentoffset* offsetfromhere)))
                    (if (or (geq offset 32768) % only 15 bit displacements valid
                         (lessp offset -32768))
                      (stderror 
                       (bldmsg "Internal entry %r too distant" nam))
                      offset))
                  (progn
                    (setq forwardinternalreferences* 
                     (cons (cons currentoffset* nam) 
                      forwardinternalreferences*))
                    0)))))
        % will be fixed in SystemFasl...


(de labeloffset (l)
  (let (offset)
    (cond 
     ((codep l) (if *writingfaslfile
		  (inf l)
		  (difference (inf l) codebase*)))
     ((setq offset  (atsoc l labeloffsets*)) (cdr offset))
     (t (stderror (bldmsg "Unknown label %r" l)))
     )))



% ------------------------------------------------------------
% Branch analysis and optimization
% ------------------------------------------------------------

% ProcessInitCode CodeList
% Purpose: Take a code list which has already been expanded by Pass1Lap
%          and replaces all generic branches with word mode branches.
% Returns: a new code list

(de ProcessInitCode (CodeList) 
(prog (BranchAndLabelAList*)            % find branches, labels, and entries
    (setq CodeList (BuildInitCodeOffsetTable CodeList)) 
    (setq LabelOffsets* (DeleteAllButLabels BranchAndLabelAList*)) 
    (return CodeList)))


% OptimizeBranches BranchCodeList!*;
% Purpose: Take a code list which has already been expanded by Pass1Lap
%          and try to optimize the branches
% Returns: a new code list

(de OptimizeBranches (BranchCodeList*) 
(prog (BranchAndLabelAList* InstructionChanged*) 
    (setq InstructionChanged* nil)      % initially there are now changes
    (BuildOffsetTable)                  % find branches, labels, and entries
    (while InstructionChanged* 
     (progn (setq InstructionChanged* nil) 
        (FindLongBranches))) 
    (setq LabelOffsets* (DeleteAllButLabels BranchAndLabelAList*)) 
    (return BranchCodeList*)))


(de DeleteAllButLabels (X) 
(prog (Y) 
   (while (not (LabelP (car (first X)))) (setq X (cdr X))) 

   (cond ((null X) (return nil))) 
    (setq Y X) 
    (while (cdr Y) 
        (cond ((LabelP (car (second Y))) 
            (setq Y (cdr Y))) (t 
        
            (Rplacd Y (cddr Y))))) 
    (return X)))


% BuildInitCodeOffsetTable CodeList;
% Purpose: generate a association list of labels, procedure entries
% Input is:
     %   labels:             label
     %   instructions:     ( opcode [operands]* )
     %   entry points:     ( !*entry procedurename proceduretype numberofargs)
% The Alist has the form:
     %   labels:           ( label . CurrentOffset)
     %   procedures:       ( procedurename . CurrentOffset)
     %   branch instrs     ( (opcode label) . CurrentOffset)
% otherwise, the CurrentOffset is advanced by the length of the instruction

(de BuildInitCodeOffsetTable (CodeList) 
(prog (AList Instr) 
    (setq CodeSize* CurrentOffset*) 
    (foreach X in CodeList do 
   (progn (cond ((LabelP X) 
          (setq AList (cons (cons X CodeSize*) AList))) 
   ((equal (setq Instr (first X)) '*entry) 
          (setq AList (cons (cons (second X) CodeSize*) AList))) 
   ((GeneralBranchInstructionP Instr) 
          (progn (Rplaca X (get Instr 'WordBranch)) 
                    (setq CodeSize* (plus CodeSize* (InstructionLength X))) 
                    (setq AList (cons (cons X CodeSize*) AList)))) (t 
                
   (setq CodeSize* (plus CodeSize* (InstructionLength X))))))) 
    

    (setq BranchAndLabelAList* (ReversIP AList)) 
    (return CodeList)))


%  BuildOffsetTable();
% Purpose: generate a association list of labels, procedure entries
% Input is:
     %   labels:             label
     %   instructions:     ( opcode [operands]* )
     %   entry points:     ( !*entry procedurename proceduretype numberofargs)
% The Alist has the form:
     %   labels:           ( label . CurrentOffset)
     %   procedures:       ( procedurename . CurrentOffset)
     %   branch instrs     ( (opcode label) . CurrentOffset)
% otherwise, the CurrentOffset is advanced by the length of the instruction

(de BuildOffsetTable nil 
(prog (AList Instr) 
    (setq CodeSize* CurrentOffset*) 
    (foreach X in BranchCodeList* do 
   (progn (cond ((LabelP X) 
          (setq AList (cons (cons X CodeSize*) AList))) 
   ((equal (setq Instr (first X)) '*entry) 
          (setq AList (cons (cons (second X) CodeSize*) AList))) 
   ((GeneralBranchInstructionP Instr) 
          (progn (setq CodeSize* (plus CodeSize* (InstructionLength X))) 
                    (setq AList (cons (cons X CodeSize*) AList)))) (t 
                
   (setq CodeSize* (plus CodeSize* (InstructionLength X))))))) 
    

    (setq BranchAndLabelAList* (ReversIP AList)) 
    (setq InstructionChanged* BranchAndLabelAList*)))


(de FindLongBranches nil 
(prog (CurrentDisplacement) 
  (foreach entry on BranchAndLabelAList* do 
    (cond ((not (LabelP (car (first entry)))) 
      
      (progn 
        (setq CurrentDisplacement (FindDisplacement (first entry))) 
        (cond 
           ((or (GreaterP CurrentDisplacement (const MaximumShortBranch)) (ZeroP CurrentDisplacement)) 	% Must have long brahch.
          
          (progn (setq InstructionChanged* t) 
              (IncreaseAllOffsets entry (MakeLongBranch entry)))))))))))
      


(de FindDisplacement (InstructionOffsetPair) 
    (Abs (Difference (cdr InstructionOffsetPair) 
                    (FindLabelOffset (second (first InstructionOffsetPair))))))

%  FindLabelOffset(Label)
% Purpose: looks up the location of Label in BranchAndLabelAList!*
% Returns: the offset of said Label

(de FindLabelOffset (L) 
(prog (Offset) 
    (return (cond ((EqCar L 'InternalEntry) (FindEntryOffset (second L))) 
    ((setq Offset (Atsoc L BranchAndLabelAList*)) 
        (cdr Offset)) (t 
    (StdError (BldMsg "Unknown label %r" L)))))))


%  FindEntryOffset(ProcedureName);
% Purpose: looks up the location of ProcedureName in BranchAndLabelAList!*
% Returns: the offset of said ProcedureName
%          if an entry point is not known for this procedure it returns a
%          dummy value of -2000

(de FindEntryOffset (L) 
    (cond ((setq L (Atsoc L BranchAndLabelAList*)) 
      (cdr L)) (t -2000)))
      % just a random number guaranteed to
                                        % make a long branch

(de MakeLongBranch (AList) 
(prog (InstructionList Result OppositeBranch) 
    (setq InstructionList (car (first AList))) 
    (Rplaca InstructionList 
           (get (first InstructionList) 'WordBranch)) 
    (cond ((cdr AList) 
        (Rplacw AList (cdr AList))) (t 
    (Rplacw AList (list (cons '~DummyLabel~ 0))))) 
    (return 2)))                        % byte branch -> word branch increases length by 2


(de IncreaseAllOffsets (X N) 
(progn (foreach Y in X do 
        (Rplacd Y (plus (cdr Y) N))) 
    (setq CodeSize* (plus CodeSize* N))))

% ------------------------------------------------------------
% Procedures to compute instruction lengths
% ------------------------------------------------------------

(de InstructionLength (X) 
(prog (Y) 
       (setq InstructionSize* (get (first X) 'Size)) 
(return (cond ((setq Y (get (first X) 'InstructionLength)) 
         (cond ((PosIntP Y) 
                Y) (t 
                (Apply Y (list X))))) 
  ((GeneralBranchInstructionP (first X)) 2) (t 
         
         (StandardInstructionLength X))))))


(de GeneralBranchInstructionP (X) 
    (get X 'WordBranch))                % generic branch has a word version

(de StandardInstructionLength (Instr) 
    (OperandListLength (rest Instr) 2))

(de OperandListLength (Tail Total) 
    (cond ((null Tail) Total) (t 
    (OperandListLength (rest Tail) (plus Total (OperandLength (first Tail)))))))

(de OperandLength (Operand) 
%
% Purpose: Compute the length of an operand to the indicated opcode
%
%  Operand has the form:
%                        (REG ... )
%                        (predecrement ... )
%                        (postincrement ... )
%                        (Indirect ... )
%                        (Displacement ... )
%                        (Indexed ... )
%                        (Entry ... )
%                        (ForeignEntry ... )
%                        (Immediate ... )
%
(prog (X) 
  (return (cond ((or (NumberP Operand) (EqCar Operand 'Immediate)) 
           (cond ((equal InstructionSize* (const Long)) 4) (t 2))) 
                  
                  % byte or word is 2
    ((atom Operand) 2) (t               % assuming word label
    
    (progn (setq X (get (car Operand) 'OperandLength)) 
        (cond ((null X) 4)              % LISP type operands require longs
        ((PosIntP X) X) (t (Apply X (list Operand))))))))))


(de PosIntP (X)                         % not defined interpretively
    (and (FixP X) (GEQ X 0)))

(de DisplacementLength (X) 
%
% (displacement xxx 0) is actually (indirect xxx)
%
    (cond ((equal (third X) 0) 0) (t 2)))

(de InlineConstantLength (X) 
% Purpose: returns the Size_Of_Unit_In_Bytes * Number_Of_Such_Units
%   X has the form:
%          (Unit  value_1  value_2 value_3 .... )
    (Times2 (cond ((equal (first X) 'fullword) 4) (t 2)) 
             
             
           (length (rest X))))

(de ByteConstantLength (X) 
    (Times2 (Quotient (Plus2 (length (rest X)) 1) 2) 2))

(de LapStringLength (X)                 % must fall on word boundary
% Purpose: Calculate the number of bytes occupied by a given string
%  X has the form: (STRING "xxxxxx")
    (Times2 (Quotient (Plus2 (Size (second X)) 3) 2) 2))

(DefList '((reg 0) 
          (predecrement 0) 
          (predecrement 0) 
          (postincrement 0) 
          (postincrement 0) 
          (indirect 0) 
          (reglist 2) 
          (Displacement DisplacementLength) 
          (Indexed 2) 
          (AbsoluteShort 2)) 'OperandLength)

(DefList '((fullword InlineConstantLength) 
          (halfword InlineConstantLength) 
          (byte ByteConstantLength) 
          (float 8) 
          (string LapStringLength)) 'InstructionLength)


(put '*entry 'InstructionLength 0)

% ------------------------------------------------------------
% Depositing Operations
% ------------------------------------------------------------

(de DepositByte (X) 
(progn (putbyte (Plus2 CodeBase* CurrentOffset*) 0 X) 
    (UpdateBitTable 1 0) 
    (setq CurrentOffset* (plus CurrentOffset* 1))))

(de DepositHalfWord (X) 
(progn (puthalfword (Plus2 CodeBase* CurrentOffset*) 0 X) 
    (UpdateBitTable 2 0) 
    (setq CurrentOffset* (plus CurrentOffset* 2))))

(de depositword (x)
  (putword (plus2 codebase* currentoffset*) 0 x)
  (updatebittable 4 0)
  (setq currentoffset* (plus currentoffset* 4)))

(de deposit-relocated-word (offset)
  % Given an OFFSET from CODEBASE*, deposit a word containing the
  % absolute address of that offset.
  (putword (plus2 codebase* currentoffset*)
	   0 
           (iplus2 offset (if *writingfaslfile 0 codebase*)))
  (updatebittable 4 (const reloc_word))
  (setq currentoffset* (plus currentoffset* 4)))
  
(de depositwordexpression (x)
  % Only limited expressions now handled
  (let (y)
    (cond
      ((fixp x) (depositword (int2sys x)))
      ((labelp x) (deposit-relocated-word (labeloffset x)))
      ((equal (first x) 'internalentry) 
       (let ((offset (get (second x) 'internalentryoffset)))
	 (if offset
	     (deposit-relocated-word offset)
	     (stderror (bldmsg "Internal entry forward reference: %r" x)))))
      ((equal (first x) 'idloc) (depositwordidnumber (second x)))
      ((equal (first x) 'mkitem) (deposititem (second x) (third x)))
      ((equal (first x) 'entry) (depositentry x))
      ((setq y (wconstevaluable x)) (depositword (int2sys y)))
      (t (stderror (bldmsg "Expression too complicated %r" x))))))

(de depositwordidnumber (x) 
  (cond
    ((or (not *writingfaslfile) (leq (idinf x) 128)) 
     (depositword (idinf X)))
    (t
      (putword (plus2 codebase* currentoffset*) 0 
               (makerelocword (const reloc_id_number) (findidnumber x))) 
      (setq currentoffset* (plus currentoffset* 4)) 
      (updatebittable 4 (const reloc_word)))))

(de DepositHalfWordExpression (X) 
(prog (Y) 
    (return (cond ((FixP X) (DepositHalfWord X)) 
    ((LabelP X) 
    (progn (puthalfword (Plus2 CodeBase* CurrentOffset*) 0 
                    (IPlus2 (LabelOffset X) 
                           (cond (*WritingFaslFile 0) (t CodeBase*)))) 
        (UpdateBitTable 2 (const RELOC_HALFWORD)) 
        (setq CurrentOffset* (plus CurrentOffset* 2)))) 
    
            ((and (equal (first X) 'Difference) (LabelP (second X)) 
            (LabelP (third X))) 
           (DepositHalfWord (Difference (LabelOffset (second X)) 
                                           (LabelOffset (third X))))) 
    ((equal (first X) 'IDLoc) 
           (DepositHalfWordIDNumber (second X))) 
    ((setq Y (WConstEvaluable X)) 
           (DepositHalfWord Y)) (t 
    (StdError (BldMsg "Halfword expression too complicated %r" X)))))))


(de DepositItem (TagPart InfPart) 
    (cond ((not *WritingFaslFile) 
	   (DepositWord 
	       (MkItem TagPart 
		       (cond ((LabelP InfPart) 
			      (Plus2 CodeBase* (LabelOffset InfPart))) 
			     ((equal (first InfPart) 'IDLoc) 
			      (IDInf (second InfPart))) 
			     (t 
				(StdError 
				    (BldMsg "Unknown inf in MkItem %r"
					    InfPart)))))))
	  (t 
	     (progn (cond 
		     ((LabelP InfPart)      % RELOC_CODE_OFFSET = 0
		      (putword (Plus2 CodeBase* CurrentOffset*) 0 
		       (MkItem TagPart (LabelOffset InfPart)))) 
		     ((equal (first InfPart) 'IDLoc) 
		      (putword (Plus2 CodeBase* CurrentOffset*) 0 
		       (MkItem TagPart 
			(MakeRelocInf (const RELOC_ID_NUMBER) 
			  (FindIDNumber (second InfPart))))))
		     (t 
		      (StdError (BldMsg "Unknown inf in MkItem %r"
				 InfPart))))
	      (setq CurrentOffset* (plus CurrentOffset* 4))
	      (UpdateBitTable 4 (const RELOC_INF))))))

(de DepositHalfWordIDNumber (X) 
    (cond ((or (not *WritingFaslFile) (LEQ (IDInf X) 128)) 
        (DepositHalfWord (IDInf X))) (t 
    
    (progn (puthalfword (plus2 codebase* currentoffset*) 0 
                    (makerelochalfword (const reloc_id_number) (findidnumber x))) 
        (setq currentoffset* (plus currentoffset* 2)) 
        (updatebittable 2 (const reloc_halfword))))))

% ------------------------------------------------------------
% this procedure was "inserted" to eliminate the problem with the compiler
% not generating the code for:
% function lambda y;remprop(y, 'internalentryoffset);
% who knows why it didn't generate anything reasonable - it generated nil.
(de remove-ieo-property (y) 
  (remprop y 'internalentryoffset))
% ------------------------------------------------------------

(de systemfaslfixup ()
  (prog (x)
        % fill the remaining byte in a long word, to make the length divisible by 4
        (setq x (remainder currentoffset* 16))
        (while (greaterp x 0)
          (depositbyte 0)
          (setq x (sub1 x)))
        (while forwardinternalreferences*
          (setq x 
                (get (cdr (first forwardinternalreferences*)) 
                     'internalentryoffset))
          (cond ((null x) % if the id has no offset - then report an error
                 (errorprintf 
                  "***** %r not defined in this module, call is incorrect" 
                  (cdr (first forwardinternalreferences*))))
                (t % compute the fixup
                   (progn
                     (setq x 
                      (difference x 
                       (car (first forwardinternalreferences*))))
                     (if (or (geq x 32768) (lessp x -32768))
                       (errorprintf 
                        "***** internal entry %r too distant; call is incorrect"
                        (cdr (first forwardinternalreferences*)))
                       (puthalfword 
                        (plus2 codebase* 
                         (car (first forwardinternalreferences*)))
                        0 x)))))
          (setq forwardinternalreferences* 
                (cdr forwardinternalreferences*)))
        % Now remove the InternalEntry offsets from everyone
        (mapobl 'remove-ieo-property)))

