%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/i860-asm.sl
% Title:          Intel i860 specific information for LAP-TO-ASM
% Author:         Winfried Neun
% Created:        5 May 1992
%
% Copyright (c) 1992, Konrad-Zuse-Zentrum Berin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(compiletime
 (progn (load if!-system common)
        (off usermode)))

(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
         outputsymfile!* commentformat!* labelformat!* 
         externaldeclarationformat!* exporteddeclarationformat!* 
         fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
         reservedatablockformat!* definedfunctioncellformat!* 
         undefinedfunctioncellinstructions!* mainentrypointname!* 
         codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
         numericregisternames!* pathin!*))
%%
%%      Sun-4 assembler must declare global name before use.
%%
(fluid '(*declarebeforeuse))
(setq *declarebeforeuse t)

(setq codefilenameformat!* "%w.s")

(setq datafilenameformat!* "d%w.s")

(setq inputsymfile!* "i860.sym")

(setq outputsymfile!* "i860.sym")

(setq mainentrypointname!* (intern "_psl_main"))

(setq commentformat!* " # %p %n")

(setq labelformat!* "%w:")

(setq exporteddeclarationformat!* " .globl %w%n")
(setq externaldeclarationformat!* " .globl %w%n")

(setq fullwordformat!* "   .long %e%n")

% FullWord expects %e for parameter                                        

(setq doublefloatformat!* "  .double 0d%w%n")

(setq reservezeroblockformat!* "  .comm %w,%e,4%n")

(setq reservedatablockformat!* "   .zero %e%n")

(setq definedfunctioncellformat!* "  .long %w%n")

(put 'undefinedfunction 'entrypoint 'undefinedfunction)

(put 'undefinedfunction 'symbol 'undefinedfunction)

(setq undefinedfunctioncellinstructions!*
      '((!.long undefinedfunction)
        (!.long !%w)))
 
%% From lap-to-asm: 

(de initializesymget ()
   (dataprintgloballabel (findgloballabel 'symget))
   (for (from i 0 255 1) (do
                               (dataprintfullword nilnumber*)))
   (foreach x in (car orderedidlist*) do
                               (dataprintfullword nilnumber*)))

(de initsymfnc1 (x)
  (prog (ep)
        (setq ep (get x 'entrypoint))
        (if (null ep)
          (dataprintundefinedfunctioncell (id2int x))
          (dataprintdefinedfunctioncell ep x))))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
        (setq oldout (wrs dataout*))
%        (foreach x in undefinedfunctioncellinstructions* do
%                 (asmoutlap1 x))
        (asmoutlap1 (car undefinedfunctioncellinstructions*))
        (wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
  (datadeclareexternal x)
     % dont disturb undefined function and interpreter
  (if (memq x '(COMPILEDCALLINGINTERPRETED undefinedfunction))
   (dataprintf " .long %w%n" x)
   (dataprintf definedfunctioncellformat* x (id2int idn))))

(fluid '(&links-present& locallabels*))

(de asmoutlap (u)
  (prog (locallabels* oldout)
        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
                                  (assoc '*call u)))

        (setq u (pass1lap (&fillframeholes u)))
        (setq u (lapoptreformjumps u)) % lapopt pass
        % Expand cmacros, quoted expressions
        (codeblockheader)
        (setq oldout (wrs codeout*))
        (foreach x in u do (asmoutlap1 x))
        (wrs oldout)
        (codeblocktrailer)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq asmopenparen!* '!()

(setq asmcloseparen!* '!))

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
         'binaryasmop)

(put 'lnot 'unaryasmop '!!)

%(put 'mkitem 'asmexpressionformat "(%e<<27)+%e")
(fluid '(*mkitemformat))           %modified for sun 4 os 4
(setq *mkitemformat "%e+%e<<27")

(put 'mkitem 'asmexpressionfunction 'asm-mkitem)

(de asm-mkitem (mkitem-tag-inf)
  (printf
    *mkitemformat
    `(printexpression ,(caddr mkitem-tag-inf))
    `(printexpression ,(cadr mkitem-tag-inf))))

(global '(*writingasmfile))

(de codefileheader ()

  (setq *writingasmfile t)      %% Somebody must set this flag. but who?
  (codeprintf " .text%n"))

(de datafileheader ()
  (dataprintf " .data%n"))

(de codefiletrailer ()
  (codeprintf " .align 8%n"))

(de datafiletrailer ()
  (dataprintf " .align 8%n"))

(de codeblockheader ()
  (codeprintf " .align 4%n"))

(de codeblocktrailer ()
  nil)

(de dataalignfullword ()
  (dataprintf " .align 4%n"))

(de printstring (s)
  (prog (n)
        (setq n (size s))
        (for (from i 0 n 1) (do (printbyte (indx s i))))
        (printbyte 0)
        (printf "   .align 4%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf " .byte ")
  (printexpression x)
  (terpri))

(de printhalfwordlist (l)
  (foreach x in l do (printhalfword x)))

(de printhalfword (x)
  (printf "     .half ")
  (printexpression x)
  (terpri))

(de printopcode (x)
  (prog (!*lower)
        (setq !*lower t)
        (prin2 x)))

(de specialactionformainentrypoint ()
  (codeprintf " .globl _psl_main%n_psl_main:"))

(de asmsymbolp (x)
  (unixsymbolp (if (idp x)
                 (id2string x)
                 x)))

(de unixsymbolp (x)
  (prog (n c i)
        (setq n (size x))
        (setq c (indx x 0))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i 1)
   loop
        (when (greaterp i n)
          (return t))
        (setq c (indx x i))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !0)) (leq c (char !9))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i (plus i 1))
        (go loop)))

(de printnumericoperand (x)
  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (i860-register-p x)
    (progn (prin2 '0!() (printoperand x) (prin2 '!)))
    (error 0 (list "In i860, differed operand can't be used like this:" x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))
  (printoperand (cadr x))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!)))

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintindexed (x)
  (setq x (cdr x))
  (printoperand (cadr x))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!) ))

(put 'indexed 'operandprintfunction 'operandprintindexed)

(de operandprintimmediate (x)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)
%%
%%      high and low is special keyword for i860 assembler.
%%
(de operandprinthigh (x)
  (prin2 "[")
  (printoperand (cadr x))
  (prin2 "]@h"))

(put 'high 'operandprintfunction 'operandprinthigh)
(put 'hi 'operandprintfunction 'operandprinthigh)

(de operandprintlow (x)
  (prin2 "[")
  (printoperand (cadr x))
  (prin2 "]@l"))

(put 'low 'operandprintfunction 'operandprintlow)
(put 'lo 'operandprintfunction 'operandprintlow)

(de operandprintwarray (x)
  (prin2 (cadr x)))

(put 'WARRAY 'operandprintfunction 'operandprintwarray)

(de operandprintforeignentry (x)
  (prog (!*lower)
        (setq !*lower t)
        (prin2 (cadr x))))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf " # %p %n" x)
  )

(de instructionprint (x)
  (codeprintf "/*         %p */%n" x)
  )

(de !*cerror (x)
  (prog (i)
        (setq i (wrs nil))
        (printf "%n *** CERROR: %r %n " x)
        (wrs i)
        (return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)

%%
%%      (word label)
%%
(put 'word 'asmpseudoop 'print-word)

(de print-word (x)
  (prin2 " .word ")
  (printoperand (cadr x))
  (terpri))

(put '*entry 'asmpseudoop 'asmprintentry)
(fluid '(*mainfound))
(de asmprintentry (x)
  (prog (y)
        (printcomment x)
        (setq x (cadr x))
        (setq y (findentrypoint x))
        (unless (flagp x 'internalfunction)
          (findidnumber x))
        (if (eq x mainentrypointname*)
          (progn (setq *mainfound t)
                 (specialactionformainentrypoint))
          (codedeclareexporteduse y))))

(de codedeclareexporteduse (y)
    (codedeclareexported y)
    (printlabel y)
    (terpri))

(de findentrypoint (x)
  (cond ((get x 'entrypoint) (get x 'entrypoint))
        ((and (asmsymbolp x)
              (not (get x 'symbol))
              (not (flagp x 'foreignfunction)))
         (put x 'entrypoint x) 
         x)
        (t 
         (let ((name (stringgensym)))
           (put x 'entrypoint name)
           name))
        ))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = "))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = ")
  (prin2 (caddr x))
  (terpri))

(put 'define 'asmpseudoop 'printdefine)

(de print-register-or-immediate(name)
  (cond ((and (pairp name) (or (eq (car name) 'low) (eq (car name) 'lo)))
		         (prin2 "[") (printoperand (cadr name)) (prin2 "]@l"))
        ((i860-register-p name) (printoperand name))
        (t (printoperand name))))

(de print-immediate (val)
  (cond ((atom val) (printoperand val))
        ((eq (car val) 'WARRAY) (prin2 (cadr val)))
        (t (error 0 (list "Unknown constant expression:" val)))))

%%      This function is originally defined in lap-to-asm.sl.
%%
(de printoperand (x)
  (setq x (to-standard-name x))
  (cond ((stringp x) (prin2 x))
        ((numberp x) (printnumericoperand x))
        ((i860-register-p x) (prog (*lower) (setq *lower t) (prin2 "%")
					(prin2 x)))
        ((idp x) (prin2 (findlabel x)))
        (t (prog (hd fn)
                 (setq hd (car x))
                 (cond ((setq fn (get hd 'operandprintfunction))
                        (apply fn (list x)))
                       ((and (setq fn (getd hd)) (equal (car fn) 'macro))
                        (printoperand (apply (cdr fn) (list x))))
                       ((setq fn (wconstevaluable x)) (printoperand fn))
                       (t (printexpression x)))))))

