%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/i860-instrs.sl
% Title:          Intel i860 instructions
% Author:         Winfried Neun
% Created:        19 May 1992
% Mode:           Lisp
% Package:        Compiler
%
% Copyright (c) 1992, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

(compiletime (load fasl-decls))

(global '(*writingasmfile))

(fluid '(labeloffsets!* currentoffset* codesize* codebase* entries*
	 forwardinternalreferences* lapreturnvalue*  *writingfaslfile
	 initoffset* *pgwd *pwrds label-alist*))

%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%k
%%      Assembler code generator
%%
%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

(loadtime
(dm define-instruction (u)
 (prog (name length result)
  (setq u (cdr u)) 
  (setq name (first u))
  (setq length (second u))
  (push (list 'put (mkquote name) (mkquote '*op-length*) length) result)
  (setq u (cddr u))    
  (setq result
     (append (mapcar u (function 
	 (lambda (x) (list 'put (mkquote name)
		                (mkquote (car x)) (mkquote (cdr x))))))
	result))
  (return(cons 'progn result))))
)

%%      Pseudo-Instructions

(define-instruction  *ENTRY 0 (*apply* entry-gen))
%
%       (*ENTRY name type argn)
%
(de entry-gen (instruction name type argn)
  (saveentry (list instruction name type argn)))

%
%       (WORD value ......)
%
(define-instruction WORD      4 (*apply* expand-word))
(define-instruction FullWORD  4 (*apply* expand-word))
(define-instruction STRING    0 (*apply* expand-string))
		            %will be treated by onelaplength
(define-instruction Float     8 (*apply* expand-float))

(de expand-word (in x)        (store-word-expression x))

(de expand-string (in x)
	(for (from i 0 (size x))
	     (do (setf (byte (wplus2 codebase* currentoffset*) i) (indx x i))))
	(setf (byte (wplus2 codebase* currentoffset*) (wplus2 (size x) 1)) 0)
	(setq currentoffset* (wplus2 (wtimes2 addressingunitsperitem 
					(strpack (size x)) currentoffset*))))

(de expand-float (in x)         (store-double x ))

%%    The Instruction processor

(de depositinstruction (in)
   (prog (name format spec oldcurrent lng oldbto)
 
      (setq oldcurrent currentoffset*)
      (when *writingfaslfile (setq oldbto bittableoffset*))
      (when (atom in) (return nil)) % skip labels
      (setq name (first in))
      (setq format (get name '*format*))
      (setq lng (cond ((eq name 'string) 
		        (wtimes2 addressingunitsperitem
				 (strpack (size (second in))) ))
                      ((eq name 'oddalign)
                        (if (eq currentoffset* (wand currentoffset* (wnot 7)))
                                addressingunitsperitem 0))
		      (t (get name '*op-length*) )))
      (when *writingfaslfile (updatebittable lng 0)
% clear bittable 
		             (setq bittableoffset* oldbto))
      
      (cond ((get name '*apply*) (apply (car (get name '*apply*)) in))
		          % escape for special pseudo instructions
	    ((eqcar format '*GENERAL_FORMAT*)
		 (format-matcher in (cdr format) nil '*general_format*))
	    ((eqcar format '*CTRL_FORMAT*)
                 (format-matcher in (cdr format) nil '*CTRL_format*))
	    ((eqcar format '*core-escape*)
		 (format-matcher in (cdr format) nil '*core-escape*))
	    ((eqcar format '*format3*)
		 (format-matcher in (cdr format) nil '*format3*))
	    ((eqcar format '*format4*)
                 (format-matcher in (cdr format) nil '*format4*))
	    ((setq format (get name '*fr-format*))  %float instructions
		  (format-matcher in (cdr format)
		         (list (cons 'fmt (car format))
		               (cons 'funct (car (get name '*cop1*))))
		         '*fr-format*))
	    ((setq format (get name '*fr-format-2*))  %float instructions
		  (format-matcher in (cdr format)
		         (list (cons 'fmt (car format))
		               (cons 'funct (car (or (get name '*cop0*)
						     (get name '*cop1*)))))
		         '*fr-format-2*))
	    ((setq format (get name '*fi-format*))  %float instructions
		  (format-matcher in (cdr format)
		         (list (cons 'fmt (car format)))
		         '*fi-format*))
   (t (error 4711 (list "Wrong i860 instruction :" in))))
  % 
  % update currentoffset*
  %
    (setq currentoffset* (plus2 oldcurrent lng)) 
    (when *writingfaslfile (setq bittableoffset* (plus2 oldbto lng ))) 
) )

(de format-matcher (ins format initlist formfkt) 
   (prog (in name item fmtitem alist appl)
       (setq alist 
	     (append initlist
		`((src1 . 0) (src2 . 0) (dest . 0) (off_hi . 0)
		 (simm_16 . 0) (i . 0) (bit0 . 0) (immediate . 0)
		 (funct . 0) (off_lo . 0) (imm_16 . 0) (imm_5 . 0) )))
       (setq name (car ins))
       (when (memq name '(ld!.l st!.l fld!.l fst!.l))
		(setq alist (cons '(bit0 . 1) alist)))
       (setq in (cdr ins)) 
	(while (pairp in)
	 (setq item (car in))
	 (setq fmtitem (car format))
	 (cond  ((numberp fmtitem)  nil)
		((and (idp fmtitem)
		      (setq appl (get fmtitem '*matchfn*)))
		      (setq alist (apply appl (list (car in) alist))))
		(t (error 4711 (list "wrong instruction " ins))))
	 (when (and (eq fmtitem 'src1)(or (eqcar (car in) 'entryaddress)
			(eqcar (car in) 'valueaddress)
			(eqcar (car in) 'displacement) 
			(eqcar (car in) 'deferred)
			(eqcar (car in) 'indexed)))
		(setq format (cdr format)))
	 (setq in (cdr in))
	 (setq format (cdr format))
	 (when (and (not (pairp format)) in)
		(error 4711 (list "i860 wrong instruction :" ins))))
	(while format
	   (when (idp (car format))
	       (error 4711 (list "i860 instruction too short :" ins)))
	   (setq format (cdr format)))

   (apply formfkt (list name alist))
)  )
   
(put 'dest '*matchfn* 'dest-match)

(de dest-match (r alist)
    (if (regp r) (cons (cons 'dest r) alist)
     (matcherror "Register" r)))

(put 'src1 '*matchfn* 'src1-match)

(de src1-match (x alist)
  (if (or (eqcar x 'displacement) (eqcar x 'deferred) (eqcar x 'indexed)
           (eqcar x 'entryaddress)(eqcar x 'valueaddress))
			 (based-match x alist)
    (if (regp x) 
	(cons (cons 'src1 x) alist)
	(if (and (setq x (addressvalue x)) (intp x) (wleq x 16#ffff)
 			(greaterp x -16#10000))
	    (cons (cons 'imm_16 x) (cons (cons 'src1 x) alist))
     (matcherror "Register or Immediate src1" x)))))

(put 'src2 '*matchfn* 'src2-match)

(de src2-match (x alist)
    (if (regp x) (cons (cons 'src2 x) alist)
     (matcherror "Register" x)))

(put 'fd '*matchfn* 'fd-match)

(de fd-match (r alist)
    (if (fregp r) (cons (cons 'fd r) alist)
     (matcherror "Float Register" r)))

(put 'fs '*matchfn* 'fs-match)

(de fs-match (x alist)
    (if (fregp x) (cons (cons 'fs x) alist)
     (matcherror "Float Register" x)))

(put 'ft '*matchfn* 'ft-match)

(de ft-match (x alist)
    (if (fregp x) (cons (cons 'ft x) alist)
     (matcherror "Float Register" x)))

(put 'shamt '*matchfn* 'shamt-match)

(de shamt-match (x alist)
    (if (and (posintp x) (wlessp x 33))  (cons (cons 'shamt x) alist)
     (matcherror "Shift amount"x)))

(put 'based '*matchfn* 'based-match)

(de based-match (x alist)
   (if (atom x)   (cons (cons 'immediate (addressvalue x))
		         (cons (cons 'src2 '(reg null)) alist))
      (cond ((eq (car x) 'deferred)
		 (cons (cons 'immediate 0) (cons (cons 'src2 (cadr x)) alist)))
	    ((eq (car x) 'displacement)
		     (cons (cons 'immediate (addressvalue (caddr x)))
		           (cons (cons 'src2 (cadr x)) alist)))
	    ((eq (car x) 'indexed)
                     (cons (cons 'src1 (caddr x))
                           (cons (cons 'src2 (cadr x)) alist)))
	    ((eq (car x) 'entryaddress)
		(cons (cons 'imm_16 
		        (wand (addressvalue (list 'xidloc (cadr x))) 16#ffff))
		        (cons (cons 'src2 '(reg symfnccenter)) alist)))
	    ((eq (car x) 'valueaddress)
		(cons (cons 'imm_16 
		        (wand (addressvalue (list 'xidloc (cadr(cadr x))))
		                 16#ffff))
		        (cons (cons 'src2 '(reg symvalcenter)) alist)))
	    (t (matcherror "based" x)))))

(put 'imm '*matchfn* 'imm-match)

(de imm-match (x alist)
  (prog (add)
     (return  
	(if (setq add (addressvalue x))  
		(cons (cons 'immediate (wand add 16#ffff)) alist)
	     (matcherror "immediate" x)))))

(put 'simm '*matchfn* 'simm-match)

(de simm-match (x alist)
  (prog (add)
     (return  
	(if (setq add (addressvalue x))  
		        (cons (cons 'immediate (wand add 16#ffff)) alist)
     (matcherror "signed immediate" x))))) 

(put 'offset '*matchfn* 'offset-match)

(de offset-match (x alist) % small branch address
  (prog (add)
     (return  
      (if (setq add (wquotient (jumpaddressvalue x) 4))  
	(cons (cons 'immediate
		(wand (wplus2 -1 add) 16#ffff)) 
		alist) % word addresses
     (matcherror "offset" x)))))

(put 'broffset '*matchfn* 'broffset-match)

(de broffset-match (x alist) % jump address
  (prog (add)
     (return  
	(if (setq add (wquotient (jumpaddressvalue x) 4)) 
		(cons (cons 'broffset (wand (isub1 add) 16#3ffffff))  alist) 
     (matcherror "broffset" x))))) 

(de matcherror (expected got)
    (error 4711 (list "Mismatch of Arguments (" expected "expected) :" got))) 

%
% REG-Format Instructions page B-3
%

(compiletime (ds valueofreg (x)
                (if (pairp (setq cdrass (cdr (atsoc x alist))))
                     (get-register-code cdrass) cdrass)))
(compiletime (ds src1isreg()
                (pairp (setq cdrass (cdr (atsoc 'src1 alist))))))

(de *general_format* (name alist)
   (prog (cdrass)
     (return
       (if (src1isreg)
        (general_format (get name '*opcode*) 0
		(valueofreg 'src2) (valueofreg 'dest)
		(valueofreg 'src1) (cdr (atsoc 'bit0 alist)))
	(general_format_16 (get name '*opcode*) 
		(valueofreg 'src2) (valueofreg 'dest)
		(wor (cdr (atsoc 'bit0 alist))
		     (wor (cdr (atsoc 'imm_16 alist))
			  (cdr (atsoc 'immediate alist)))))))))

(de general_format (opcode i src2 dest src1 null_imm_offset)
       (setq opcode (wor (wshift opcode 26)
                         (wor (wshift i 26)
                              (wor (wshift src2 21)
                                 (wor (wshift dest 16)
                                   (wor (wshift src1 11)
                                      null_imm_offset ))))))
        (cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
			(print (sys2int opcode)))) 
        (store-word opcode))

(de general_format_16 (opcode src2 dest imm_offset)
       (setq opcode (wor (wshift (wor opcode 1) 26)
                         (wor (wshift src2 21)
                              (wor (wshift dest 16)
				 (wand imm_offset 16#ffff)))))
        (cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
                        (print (sys2int opcode))))
        (store-word    opcode))

(de *ctrl_format* (name alist)
   (prog (cdrass)
        (ctrl_format (get name '*opcode*) (cdr (atsoc 'broffset alist)))))

(de ctrl_format (opcode broffset)
	(setq opcode (wor (wshift opcode 26) broffset))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
                        (prin2 bittableoffset* ) (prin2 "  ")
                        (print (sys2int opcode))))
        (store-word    opcode))
			

(de *format3* (name alist)
   (prog (cdrass imm)
     (setq imm (wor (cdr (atsoc 'imm_16 alist)) (cdr (atsoc 'immediate alist))))
     (return
        (st_bla_bte_blue (get name '*opcode*) 1
                (valueofreg 'src2) (wand (wshift imm -11)  16#1f)
		(valueofreg 'src1)
		  (wor (cdr (atsoc 'bit0 alist)) (wand imm 16#7ff))))))

(de *format4* (name alist)
   (prog (cdrass imm)
    (setq imm (cdr (atsoc 'immediate alist)))
    (if (src1isreg)
     (return
        (st_bla_bte_blue (get name '*opcode*) 1
                (valueofreg 'src2) (wand (wshift imm -11)  16#1f)
                (valueofreg 'src1) (wand imm 16#7ff)))
     (return
        (st_bla_bte_blue (get name '*opcode*) 0
                (valueofreg 'src2) (wand (wshift imm -11)  16#1f)
                (valueofreg 'src1) (wand imm 16#7ff))))))

(de st_bla_bte_blue (opcode i src2 offset_high src1 offset_low)
       (setq opcode (wor (wshift opcode 26)
                         (wor (wshift i 26)
                              (wor (wshift src2 21)
                                   (wor (wshift offset_high 16)
                                        (wor (wshift src1 11)
                                              offset_low))))))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
                        (print (sys2int opcode))))
        (store-word    opcode))

(de bte_bne_with_5 (opcode src2 offset_high imm offset_low)
       (setq opcode (wor (wshift (wor opcode 1) 26)
                         (wor (wshift src2 21)
                              (wor (wshift offset_high 16)
                                   (wor (wshift imm 11)
                                        offset_low)))))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
                        (print (sys2int opcode))))
        (store-word    opcode))

%
% Core Escape Instructions
%

(de *core-escape* (name alist)
   (prog (cdrass)
     (return
	(core_escape (valueofreg 'src1) (get name '*core-escape*)))))

(de core_escape (src1 opcode)
      (let ((hilf 2#010011)) 
        (setq opcode (wor (wshift hilf 26)
                          (wor (wshift src1 11) opcode)))
	(cond (*pgwd (prin2 currentoffset* ) (prin2 "  ")
			(prin2 bittableoffset* ) (prin2 "  ")
                        (print (sys2int opcode))))
        (store-word    opcode)))

(ds core_escape_inst (name op)
    (progn
      (put 'name '*format* '(*core-escape* src1))
      (put 'name '*OP-LENGTH* 4)
      (put 'name '*core-escape* op)))
 
(core_escape_inst lock   2#00001)
(core_escape_inst calli  2#00010)
(core_escape_inst intovr 2#00100) 
(core_escape_inst unlock 2#00111)

(define-instruction oddalign  0 (*apply* expand-oddalign))

(de expand-oddalign (in)
     (when (eq (wand currentoffset* (wnot 7)) currentoffset*)
        (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem))))

(define-instruction nop 4  (*apply* nop-gen))

(de nop-gen ()  (depositinstruction '(shl  (reg null) (reg null) (reg null) )))

%%      move-relocatable-object pseudo instructions

(define-instruction move-relocatable-object 8 (*apply* movex-gen))

(de movex-gen (name src reg)
	  (cond (*writingfaslfile
	     (store-word-expression src) %% when writing to faslfile, generate
	     (relocatable-movex-gen reg)) %% special code for relocation.
	   (t (depositinstruction `(orh (hi ,src) (reg 0) ,reg)) 
		                         %% otherwise normal code.
	      (depositinstruction `(or (lo ,src) ,reg ,reg)))))

(define-instruction  move-relocatable-object-and-load 8 (*apply* movex-load))

(de movex-load (name reg src )
   (cond (*writingfaslfile
	   (store-word-expression  src)  %% when writing to faslfile, generate
	   (relocatable-movex-load  reg))   %% special code for relocation.
	 (t (depositinstruction `(orh (adjhi ,src) (reg 0) ,reg))
		                 %% otherwise normal code.
	    (depositinstruction
		 `(ld!.l (displacement ,reg (adjlo ,src)) ,reg)))))

(define-instruction move-relocatable-object-and-store 8 (*apply* movex-store))

(de movex-store (name reg src cont)
  (cond (*writingfaslfile
	   (store-word-expression src) %% when writing to faslfile, generate
	   (relocatable-movex-store reg cont)) %% code for relocation.
	(t (depositinstruction `(orh (adjhi ,src) (reg 0) ,reg))
		                          %% otherwise normal code.
	   (depositinstruction
		 `(st!.l ,cont (displacement ,reg (ajdlo ,src)))))))

(de relocatable-movex-gen (reg)
  (let ((regcode (get-register-code reg))
	(mov-instr 2#111001))
	(store-relocatable-movex            % or instr 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode 16) )))))

(de relocatable-movex-store (reg cont)
  (let ((regcode (get-register-code reg))
	(regcode1 (get-register-code cont))
	(mov-instr 8#000111))  %st!.l
       (store-relocatable-movex 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21)
			 (wor (wshift regcode1 16) 1))))))

(de relocatable-movex-load (reg)
  (let ((regcode (get-register-code reg))
	(mov-instr 2#000101))
	(store-relocatable-movex                      % ld.!l  immediate
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21)
			 (wor (wshift regcode 16) 1))))))


(loadtime
(dm define-instr (u)
   (prog (name opcode fmt iallowed)
	(setq name (cadr u))
	(setq opcode (caddr u))
	(setq fmt '(*general_format* src1 src2 dest))
	(when (cdddr u) (if (eq (cadddr u) 'i) (setq iallowed t)
			  (if (eq (cadddr u) 'format3)
				(setq fmt '(*format3* src1 based))
		 			(setq fmt (cadddr u)))))
	(return `(progn (put  ',name '*opcode* ,opcode)
		 (put ',name '*format* ',fmt)
		 ,(when iallowed `(flag '(,name) '*iallowed))
		 (put ',name '*op-length* 4)
		 (put ',name  '*opcode* ,opcode ))))))

(define-instr	ld!.b	2#000000 I)
(define-instr	ld!.s	2#000100 I)
(define-instr	ld!.l	2#000100 I)

(define-instr	st!.b	2#000011 format3)
(define-instr	st!.s	2#000111 format3)
(define-instr	st!.l	2#000111 format3)

(define-instr	ixfr	2#000010)
(define-instr	reserved1	2#000110)

(define-instr	fld!.x	2#001000 I)
(define-instr	fst!.x	2#001010 I)
(define-instr	flush	2#001101)
(define-instr	pst!.d	2#001111)
(define-instr	ld!.c	2#001100)
(define-instr	st!.c	2#001110)

(define-instr	bri 	2#010000 (*GENERAL_FORMAT*  SRC1))
(define-instr	trap	2#010001)

(define-instr	FP_escape	2#010010)
(define-instr	core_escape	2#010011 core)

(define-instr	bte	2#010110 (*format4* src1 src2 offset))
(define-instr	btne	2#010100 (*format4* src1 src2 offset))
(define-instr	pfld!.y	2#011100 I)

(define-instr	ctrl1	2#011000)
(define-instr	ctrl2	2#011001)
(define-instr	br   	2#011010 (*ctrl_format* broffset))
(define-instr	call 	2#011011 (*ctrl_format* broffset))
(define-instr	bc	2#011100 (*ctrl_format* broffset))
(define-instr	bc!.t   2#011101 (*ctrl_format* broffset))
(define-instr	bnc	2#011110 (*ctrl_format* broffset))
(define-instr	bnc!.t  2#011111 (*ctrl_format* broffset))

(define-instr	addu	2#100000 I)
(define-instr	adds	2#100100 I)
(define-instr	subu	2#100010 I)
(define-instr	subs	2#100110 I)
(define-instr	shl 	2#101000 I)
(define-instr	shr 	2#101010 I)
(define-instr	shrd	2#101100)
(define-instr	bla	2#101101 format3)
(define-instr	shra	2#101110 I)
(define-instr	and 	2#110000 I)
(define-instr	andh	2#110010 I)
(define-instr	andnot	2#110100 I)
(define-instr	andnoth	2#110110 I)
(define-instr	or	2#111000 I)
(define-instr	orh	2#111010 I)
(define-instr	xor	2#111100 I)
(define-instr	xorh	2#111110 I)

(define-instr	reserved2	2#110010)
(define-instr	reserved3	2#110110)
(define-instr	reserved4	2#111010)
(define-instr	reserved5	2#111110)

 %
 % Basic Predicates
 %

(de i860-immediate-p(r)
  (cond ((fixp r) t)
	((stringp r) t)
	((idp r)  (cond ((get r '*register-code) nil)
	                ((get r '*f-register-code) nil)
	                ((get r '*g-register-code) nil)
	                 (t     t)))  %% label
	(t nil)))

(de i860-relocatable-immediate-p (r)  (eqcar r 'IMMEDIATE))

(de i860-short-immediate-p (r)
	  (and (fixp r) (geq r (minus 16#8000)) (lessp r 16#8000)))

(de i860-huge-displacement-p (r)
	   (and (eqcar r 'displacement)
                (numberp (caddr r))
		(not (i860-short-immediate-p (caddr r)))))

(de i860-deferred-p (r)    (eqcar r 'deferred))

(de i860-indirect-p (r)
  (and (pairp r)
       (or (i860-register-p (car r))
	   (eq (car r) 'deferred)
	   (eq (car r) 'displacement))))

(de i860-double-indirect-p (r)
   (and (eqcar r 'deferred) (i860-indirect-p (cadr r))))

(de i860-fluid-p (r)
	  (and (pairp r) (member (car r)  '(!$fluid !$global fluid global))))

(de i860-absolute-p (r)
	  (or (eqcar r 'quote) (stringp r)))

(de i860-foreignentry-p (r)
	  (eqcar r 'foreignentry))
