%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 370-CMAC-FNS.SL - predicates and fns for IBM370 PSL cmacro expansion     %
%                                                                          %
% Author:      Jim McDonald                                                %
%              IMSSS                                                       %
%              Stanford University                                         %
% Date:        11 December 1982                                            %
% Based on HP-CMAC.SL of Mar 82 by Maguire, Griss, Lowder                  %
% Based on VAX-CMAC.SL of 14 Jan 1982 by Eric Benson                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(Dm !* (x) Nil)


% seems pretty strange to me (jim), so I commented it out:
% % This is needed to set up real 32 bit arithmetic. These functions
% % are opencoded below..
% (loadtime (progn
% (RemProp 'Wtimes2 'OpenFn)     % So need explicit code
% (RemProp 'WQuotient 'OpenCode)
% (RemProp 'WRemainder 'OpenCode)))

(fluid '(AddressingUnitsPerItem
         AddressingUnitsPerFunctionCell
         CharactersPerWord
         StackDirection
         !*ImmediateQuote
         TagStartingBit
         TagBitLength
         InfStartingBit
         InfBitLength
         GCStartingBit
         GCBitLength
         TagDuringGCStartingBit
         TagDuringGCBitLength
         BitsPerWord
         BitsPerByte
        ))

(fluid '( CurrentLocation!* ))  % program counter for generated code
                                % 370 => displacement into current routine
                                % DEC => absolute address
                                % for now, only used by !*JUMPON

(setq CurrentLocation!* 8)      % overly conservative safe value
                                % (says can't refer back more than 8 bytes)

(fluid '(FloatMode!*))          % normally set to 'Long, the default

(flag '(Immediate UnImmediate Deferred Indexed DoubleIndexed LengthBased
        DisplacementOf BaseRegOf FnCell Fluid Global !$Fluid !$Global
        RegList Lit ByteLoc ThreeByteLoc)
      'TerminalOperand)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               MACHINE DEPENDENT CONSTANTS FOR THE 370                     %
%                                                                           %
% Four BYTES per word (32 bits).                                            %
% Four CHARACTERs per word (32 bits).                                       %
% Stack grows DOWN, rather than UP, terminates on protected page            %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq TagStartingBit 0)             % Tag in high order byte
(setq TagBitLength 8)
(setq InfStartingBit 8)             % Inf in three low order bytes
(setq InfBitLength 24)
(setq GCStartingBit 0)              % GC field overlaps top 3 bits of Tag
(setq GCBitLength 3)
(setq TagDuringGCStartingBit 3)     % Pure tag in low 5 bits of normal Tag
(setq TagDuringGCBitLength 5)
(setq AddressingUnitsPerItem 4)
(setq CharactersPerWord 4)
(setq BitsPerWord 32)
(setq BitsPerByte  8)
(setq AddressingUnitsPerFunctionCell 4)
(setq StackDirection -1)                % Stack grows DOWN, not up.
(setq !*ImmediateQuote NIL)



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                   370 INSTRUCTION FORMATS                                 %
%                                                                           %
% 370 instructions are variable length, with one of seven formats:          %
%                                                                           %
%                                                                           %
% Format | Byte 1  | Byte 2  | Byte 3  | Byte 4  | Byte 5  | Byte 6  |      %
% -------+---------+----+----+----------------------------------------      %
%   RR   | Opcode  | R1 | R2 |                                              %
% -------+---------+----+----+----+--------------+                          %
%   RX   | Opcode  | R1 | X2 | B2 | D2 (12 bit)  |  (double indexing)       %
% -------+---------+----+----+----+--------------+                          %
%   RS   | Opcode  | R1 | R3 | B2 | D2           |                          %
% -------+---------+----+----+----+--------------+                          %
%   SI   | Opcode  | I2      | B1 | D1           |                          %
% -------+---------+---------+----+--------------+                          %
%   S    | Opcode            | B1 | D1           |                          %
% -------+---------+---------+----+--------------+----+-------------+       %
%   SS   | Opcode  | L       | B1 | D1           | B2 | D2 (12 bit) |       %
% -------+---------+----+----+----+--------------+----+-------------+       %
%   SS   | Opcode  | L1 | L2 | B1 | D1           | B2 | D2          |       %
% -------+---------+----+----+----+--------------+----+-------------+       %
%                                                                           %
% There are 16 general registers (R0-R15).                                  %
% All but R0 can be used as a base or index register.                       %
%                                                                           %
% There may also be 8 floating point registers, which are used only in      %
% special floating point instructions.                                      %
%                                                                           %
% Registers may be referenced individually or as implicit pairs, in which   %
% case only the first register (which must be even) is specified.           %
% (The choice is fixed for each instruction.)                               %
%                                                                           %
% Data from memory can be accessed as bytes, halfwords, fullwords,          %
% or doublewords.  There are also "long" accesses which iterate through     %
% memory while the count in a register is updated.                          %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               PREDICATES FOR TESTING 370 OPERAND TYPES                    %
%                                                                           %
% RegP          -- general purpose registers                                %
%                                                                           %
% EvenRegP      -- even general purpose registers                           %
%                                                                           %
% FloatRegP     -- floating point registers                                 %
%                                                                           %
% EvenFloatRegP -- even floating point registers                            %
%                                                                           %
% FakeRegP      -- registers for extrareg (i.e. greater than 5)             %
%                                                                           %
% IndexRegP     -- register being used as index (never used)                %
%                                                                           %
% BaseRegP      -- register being used as base  (never used)                %
%                                                                           %
% DispInumP     -- objects that can be used as a displacement,              %
%                  i.e., positive 12-bit integers.                          %
%                                                                           %
% NegDispInumP  -- objects whose negation satisfies DispInumP.              %
%                                                                           %
% SIImmediateP  -- objects that can be used as immediate data in SI format  %
%                  i.e. positive 8-bit integers.                            %
%                  Note this is distinct from ImmediateP below, which tests %
%                  for a logical notion.                                    %
%                                                                           %
% NegSIImmediateP -- objects whose negation satisfies SIImmediateP.         %
%                                                                           %
% PosInumP      -- constant positive integers                               %
% ByteInumP     -- constant  8 bit positive integers                        %
% DispInumP     -- constant 12 bit positive integers                        %
% HalfInumP     -- constant 16 bit positive integers                        %
% FullInumP     -- constant 32 bit          integers                        %
% DoubleInumP   -- constant 64 bit          integers                        %
% StartingBitP  -- inum between 0 and 31                                    %
% BitLengthP    -- inum between 1 and 32                                    %
% Leq31P        -- inum between 0 and 31                                    %
% NegLeq31P     -- inum between -31 and -1                                  %
%                                                                           %
% NegInumP      -- constant negative integers                               %
% NegByteInumP  -- constants whose negation satisfies ByteInumP             %
% NegDispInumP  -- constants whose negation satisfies DispInumP             %
% NegHalfInumP  -- constants whose negation satisfies HalfInumP             %
%                                                                           %
% ByteP         -- Objects that occupy one   byte  in memory.               %
% HalfwordP     -- Objects that occupy two   bytes in memory.               %
% FullwordP     -- Objects that occupy four  bytes in memory.               %
% DoublewordP   -- Objects that occupy eight bytes in memory.               %
%                                                                           %
%                  (BitSizeOf calculates the number of bits needed to       %
%                   express an object.                                      %
%                   ByteLengthOf calculates the number of bytes.)           %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de      RegP (RegName) (EqCar RegName 'Reg))

% just a thought--never used
% (de IndexRegP (RegName) (EqCar RegName 'IndexReg))
% (de  BaseRegP (RegName) (EqCar RegName 'BaseReg))

(de  EvenRegP (RegName)
        (and (EqCar RegName 'Reg) (Even (cadr Regname))))

(de FloatRegP (RegName)
        (and (EqCar RegName 'FloatReg) (Even (cadr Regname))))

(de EvenFloatRegP (RegName)
        (and (EqCar RegName 'FloatReg) (Even (cadr Regname))))

(de FakeRegP (Expression)
        (and (FixP Expression) (GreaterP Expression 5)))

(de RegT0P (RegName) (Equal RegName '(Reg T0)))
(de RegT1P (RegName) (Equal RegName '(Reg T1)))
(de RegT2P (RegName) (Equal RegName '(Reg T2)))
(de RegT3P (RegName) (Equal RegName '(Reg T3)))

% following used by compiler when looking for patterns
(de !&RealRegP (U)
 (or (EqCar U 'Reg)
     (and (PairP U)
          (FlagP (Car U) 'Var)
          (Get (cadr U) 'RegisterAlias))))


(de SIImmediateP     (x) (ByteInumP x))
(de NegSIImmediateP  (x) (NegByteInumP x))

(de ByteInumP      (x) (and (fixp x) (leq3 0 x 16#FF)))
(de DispInumP      (x) (and (fixp x) (leq3 0 x 16#FFF)))
% note that 16#8000 is the halfword representation of -16#8000
% 16#80000000 is the fullword representation of -16#80000000
(de PosHalfInumP   (x) (and (fixp x) (leq3 0 x 16#7FFF)))
(de PosFullInumP   (x) (and (fixp x) (leq3 0 x 16#7FFFFFFF)))

(de HalfInumP      (x) (and (fixp x) (leq3 -16#8000 x 16#7FFF)))
(de UnsignedHalfInumP
                   (x) (and (fixp x) (leq3 0 x 16#FFFF)))
(de FullInumP      (x) (and (fixp x) (leq3 (ilsh -16#40000000 1) x
                                           16#7FFFFFFF)))
% ilsh above allows us to obtain -16#80000000,
%  which the 370 version can use internally, but cannot read directly
% lsh below allows us to avoid using 16#FFFFFFFF,
%  which would be interpreted as a negative number on the 370
(de UnsignedFullInumP
                   (x) (and (fixp x)
                            (leq3 0 (lsh x -1) 16#7FFFFFFF)))
(de DoubleInumP    (x) (and (fixp x)
                            (leq3 16#7FFFFFFF x "16#7FFFFFFFFFFFFFFF")))

% Better names for the following might be NegationOfAPositiveByteInumP, etc.
(de NegByteInumP   (x) (and (fixp x) (leq3 -16#FF x -1)))
(de NegDispInumP   (x) (and (fixp x) (leq3 -16#FFF x -1)))
% The following could use -16#8000 and -16#80000000, but then if either was
%  true, a second check would be needed if the number was to be negated.
%  (-16#8000 and -16#80000000 can be represented as a halfword and fullword,
%   respectively, but 16#8000 and 16#80000000 cannot.)
% Thus -16#8000 will be handled as a fullword, and -16#80000000 will be
%  handled as a doubleword or as an error condition.
(de NegHalfInumP   (x) (and (fixp x) (leq3 -16#7FFF x -1)))
(de NegFullInumP   (x) (and (fixp x) (leq3 -16#7FFFFFFF x -1)))

(de StartingBitP   (X) (and (fixp x) (leq3 0 x 31)) )
(de BitLengthP     (X) (and (fixp x) (leq3 1 x 32)) )
(de Leq31P         (X) (and (fixp x) (leq3 0 x 31)) )
(de NegLeq31P      (X) (and (fixp x) (leq3 -31 x -1)) )

(de ByteP          (x) (leq (BitSizeOf x)  8))

(de HalfwordP      (x) (leq (BitSizeOf x) 16))
(de FullwordP      (x) (and (not (RegP x)) (leq3 17 (BitSizeOf x) 32)))
(de DoublewordP    (x) (leq3 33 (BitSizeOf X) 64))

(de leq3 (x y z) (and (leq x y) (leq y z)))

(de BitSizeOf (x)
 (prog (s sizefn)
       (return
        (cond ((fixp x)
               (cond ((minusp x) 32)
                     ((LessP x 16#F) 4)
                     ((LessP x 16#FF) 8)
                     ((LessP x 16#FFF) 12)
                     ((LessP x 16#FFFF) 16)
                     (T 32)))
              ((floatp x)
               (cond ((eq FloatMode!* 'Short) 32)
                     ((eq FloatMode!* 'Extended) 128)
                     (t 64)))  % long is default
              ((atom x)
               (progn (TTYPrintF "%f***  BitSizeOf: %p%n" x)
                      (JimStop)
                      32))
              ((setq s (get (car x) 'BitSize)) s)
              ((setq SizeFn (get (car x) 'BitSizeFn))
               (apply SizeFn (cdr x)))
              (t (progn (TTYPrintF "%f***  BitSizeOf: %p%n" x)
                        (JimStop)
                        32))))))

% note: 4 * 16#2000     = 16#8000     which LH would fetch as NEGATIVE number
%       4 * 16#20000000 = 16#80000000 which L  would fetch as NEGATIVE number

(de HalfWordIdOffsetP (x)
 (cond ((eqcar x 'IdOffset) (lessp (FindIdNumber (cadr x)) 16#2000))))

(de FullWordIdOffsetP (x)
 (cond ((eqcar x 'IdOffset) (lessp (FindIdNumber (cadr x)) 16#20000000))))

(de jimstop nil nil)  % remd to break BitSizeOf above upon errors

(de ByteLengthOf (x) (add1 (quotient (sub1 (BitSizeOf x)) 8)))


(deflist '((Fluid 32) (!$Fluid 32) (Global 32) (!$Global 32) (FnCell 32)
           (Reg 32) (ExtraReg 32) (Indexed 32) (DoubleIndexed 32)
           (WVar 32) (WArray 32) (Quote 32)
           (ByteLoc 8) (ThreeByteLoc 24) (Unimmediate 32)
           (InternalEntry 32)
           (InternalCodePtr 32))
         'BitSize)

(deflist '( (Immediate ImmediateBitSize)
            (Plus2 Plus2BitSize)
            (WPlus2 Plus2BitSize)
            (Lit LitBitSize) )
         'BitSizeFn)

(de ImmediateBitSize (x)
 (cond ((EqCar x 'InternalEntry) 12)
       ((WArrayP X) 32)                 % includes WString's
       (t (BitSizeOf x))))


(de Plus2BitSize (x y)
 (prog (xsize ysize)
       (setq xsize (BitSizeOf x))
       (setq ysize (BitSizeOf y))
       (return (cond ((equal xsize ysize) (min 32 (add1 xsize)))
                     (t (max xsize ysize))))))
% (add1 (max xsize ysize)) is safer, but may yield 13 bits for 4+FOO,
% where FOO has 12 bits.
% 13 would rarely be necessary--offset of FOO needs to be within 4 of 4096.
% would like to get actual offsets, add them, and then compute bit size...

(de LitBitSize (x)
 (cond ((eqcar x 'fullword) 32)
       ((eqcar x 'halfword) 16)
       ((eqcar x 'Byte) 8)
       ((eqcar x 'EBCDIC) (add1 (size (cadr x))))
       ((eqcar x 'ASCII)  (add1 (size (cadr x))))
       ((eqcar x 'string) (add1 (size (cadr x))))
       (t (progn (TTYPrintF "%f*** LitBitSize: %p%n" x) 32))))

(de log2 (x)
 (cond ((minusp x) (progn (TTYPrintF "%f?? Log2 of: %p%n" x) 32))
       ((lessp x 3) 1)
       (t (add1 (log2 (quotient (add1 x) 2))))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                           %
%  Every ID has a function cell, which is a word in memory used to hold the %
%  starting address for the function associated with that ID.               %
%                                                                           %
%  If a function cell is located within 4096 bytes of the address stored in %
%  SymFncR, then its contents can be accessed with a single instruction:    %
%                                                                           %
%                         L CallReg,FOO(SymFncR)                            %
%                                                                           %
%  If the cell is 4096 or more bytes beyond the address stored in SymFncR,  %
%  then two instructions will be needed, one to load the cell's address,    %
%  and a second to load its contents.                                       %
%                                                                           %
%  Every ID also has a value cell, which is a word in memory used to hold   %
%  the value of that ID.                                                    %
%                                                                           %
%  If the value cell is located within 4096 bytes of the address stored in  %
%  SymValR, then its contents can be accessed with a single instruction:    %
%                                                                           %
%                         L R1,FOO(SymValR)                                 %
%                                                                           %
%  If the cell is 4096 or more bytes beyond the address stored in SymValR,  %
%  then two instructions will be needed, one to load the cell's address,    %
%  and a second to load its contents.                                       %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de FluidCellP (x)
  (and (pairp x) (member (car x) '(Fluid Global !$Fluid !$Global))))

(de BigOffsetIdP (x)
 (and (IdP x) (greaterp (FindIdNumber x) 1023)))

(de BigOffsetP (x)
 (and (pairp x)
      (not (RegP x))
      (not (member (car x) '(Lit DisplacementOf Quote)))
      (greaterp (OffsetOf x) 16#FFF)))

% DisplacementOf is a macro to please wconstevaluable
(dm DisplacementOf (x) (OffsetOf (cadr x)))

(de OffsetOf (x)
 (prog (fn)
       (cond ((WVarP X) (return 1000))                  % assume < 4096
             ((WArrayP X) (return 6000))                % assume > 4096
             ((atom x)
              (progn (TTYPrintF "%f*** OffsetOf: %p%n" x)
                     (return 1200))))                   % assume < 4096
       (setq fn (get (car x) 'OffsetFn))
       (cond ((null fn)
              (progn (TTYPrintF "%f*** OffsetOf: %p%n" x)
                     (return 1300))))                   % assume < 4096
       (return (apply fn (cdr x)))))

(deflist '(
        (Indexed IndexedOffset)                 % straightforward
        (DoubleIndexed DoubleIndexedOffset)     % straightforward
        (FnCell FnCellOffset)                   % easy to calculate
        (Fluid ValueCellOffset)                 % easy to calculate
        (Global ValueCellOffset)                % easy to calculate
        (!$Fluid ValueCellOffset)               % easy to calculate
        (!$Global ValueCellOffset)              % easy to calculate
        (WConst OffsetOf)                       % no-op (?) X may be WString
        (Immediate ImmediateOffset)             % guesses
        (Unimmediate UnimmediateOffset)         % guesses
        (ExtraReg ExtraRegOffset)               % near start of DSECT
        (WVar WVarOffset)                       % near start of DSECT
        (WArrayP WArrayOffset)                  % near start of DSECT
        (InternalEntry InternalEntryOffset)     % guesses
        (InternalCodePtr InternalCodePtrOffset) % guesses
        (Plus2 Plus2Offset)                     % guesses
        (WPlus2 Plus2Offset)                    % guesses
        (Times2 Times2Offset)                   % guesses
        (WTimes2 Times2Offset)                  % guesses
        ) 'OffsetFn)

(de IndexedOffset       (x y)   y)                              % good
(de DoubleIndexedOffset (x y z) z)                              % good
(de FnCellOffset        (x) (times2 4 (FindIdNumber x)))        % good
(de ValueCellOffset     (x) (times2 4 (FindIdNumber x)))        % good

(de ImmediateOffset     (x) (OffsetOf X))                       % ???
(de UnImmediateOffset   (x) (OffsetOf X))                       % ???


% these should suffice until we know how to calculate them precisely:

(de ExtraRegOffset      (x) 1300)                         % assume < 4096
(de WVarOffset          (x) 1400)                         % assume < 4096
(de WArrayOffset        (x) 1500)                         % assume < 4096

(de InternalEntryOffset (x) 1600)                         % assume < 4096
(de InternalCodePtrOffset (X) 1700)                   % just to be < 4096

(de Plus2Offset (x y)
 (cond ((FixP x) (plus2 x (OffsetOf y)))
       ((FixP y) (plus2 (OffsetOf x) y))
       (t (progn (TTYPrintF "%f*** Plus2Offset: %p, %p%n" X Y)
                 2000))))         % assume < 4096

(de Times2Offset (x y)
 (cond ((FixP x) (times2 x (OffsetOf y)))
       ((FixP y) (times2 (OffsetOf x) y))
       (t (progn (TTYPrintF "%f*** Times2Offset: %p, %p%n" X Y)
                 2100))))  % assume < 4096




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               PREDICATES FOR TESTING LOGICAL PROPERTIES                   %
%                                                                           %
% ImmediateP -- objects that the compiler has tagged as IMMEDIATE.          %
%               Such an object is to be used as a value, not an address.    %
%                                                                           %
%               Note that immediate in this logical sense does NOT imply    %
%               that the object can appear as an immediate operand in an    %
%               instruction (see DispInumP and SIImmediateP above).         %
%                                                                           %
%               WConsts and WArrays are tagged immediate when they are not  %
%               inside a Memory construct.  For example, if (WArray Foo) is %
%               the base of a table, then (!*WPlus2 (Reg 1) (WArray Foo))   %
%               will have (WArray Foo) tagged as immediate, meaning that    %
%               the address of Foo, NOT the contents at that address,       %
%               should be added to (Reg 1).                                 %
%                                                                           %
%               Because immediate operands on the 370 must be rather small, %
%               the logically immediate values detected by this predicate   %
%               will usually be stored as a literal in memory (or will be   %
%               used to compute such a literal), and the address of the     %
%               literal will actually appear in the relevant instruction.   %
%                                                                           %
%  InternalCurrentEntryP                                                    %
%               True if the arguement is an entry point to the current      %
%               routine and we are converting recursions to iterations.     %
%                                                                           %
%  IndirectViaSymFncP                                                       %
%               True for objects of the form (Memory xxx SymFnc),           %
%               which become (DoubleIndexed xxx (reg SymFncR)).             %
%                                                                           %
%  NullP -- Nil or (Immediate Nil)                                          %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de NullP (x) (or (null x) (equal x '(Immediate nil))))

(de ImmediateP (X) (and (EqCar X 'Immediate) (Null (FixP (cadr X)))))

(de InternalCurrentEntryP (X)
 (and !*R2I (memq X EntryPoints!*)))    % R2I = recursion-to-iteration

(de WVarP (X) (or (EqCar X 'WVar) (and (IdP X) (get X 'WVar))))

(de IndirectViaSymFncP (x)
 (and (EqCar x 'DoubleIndexed) (equal (caddr x) '(Reg SymFncR))))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                 MISCELLANEOUS USEFUL PREDICATES                           %
%                                                                           %
%  EqualP, OneP, TwoP, FourP, EightP, SixteenP, etc.                        %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de EqualP (x y) (Equal x y))

% ZeroP, OneP already defined (true for floating pt as well, though)
(de TwoP        (x) (equal x  2))
(de FourP       (x) (equal x  4))
(de EightP      (x) (equal x  8))
(de SixteenP    (x) (equal x 16))
(de TwentyFourP (x) (equal x 24))
(de ThirtyTwoP  (x) (equal x 32))

(de ByteAllOnesP (x) (and (fixp x)
                          (equal (land x (sub1 (expt 2 8)))
                                 (sub1 (expt 2  8)))))
(de HwAllOnesP   (x) (and (fixp x)
                          (equal (land x (sub1 (expt 2 16)))
                                 (sub1 (expt 2 16)))))
(de FwAllOnesP   (x) (and (fixp x)
                          (equal (land x 16#FFFFFFFF) 16#FFFFFFFF)))

(de NeqLengthP (x y) (not (equal (ByteLengthOf x) (ByteLengthOf y))))

(de Mem16P   (x) (leq (ByteLengthOf x) 16))

(de Mem256P  (x) (leq (ByteLengthOf x) 256))

(de LongMemP (x y)
  (or (geq (ByteLengthOf x) 257) (geq (ByteLengthOf y) 257)))

% bit of a kludge -- want to see if sum of two args is less than 8
% second routine is only called after first routine
(fluid '(LeqEightNum!*))
(de LeqEightP (x) (and (lessp x 9) (setq LeqEightNum!* x)))
(de AndSumLeqEightP   (x) (lessp (plus LeqEightNum!* x) 9))
(de AndSumEqualEightP (x) (equal (plus LeqEightNum!* x) 8))

(de TwoDescendingRegisters (x y)
 (and (regp x) (regp y)
      (setq x (RegisterNumberOf (cadr x)))
      (setq y (RegisterNumberOf (cadr y)))
      (or (equal x (add1 y))
          (and (zerop x) (equal y 15)))))

(de RegisterNumberOf (x)
 (cond ((numberp x) x)
       ((get x 'RegisterNumber))
       ((get x 'RegisterName)   % don't recur on NIL
        (RegisterNumberOf (get x 'RegisterName)))))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       FUNCTIONS USED TO TEST FOR AND SELECT TEMP REGISTERS                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de UsesT0P (expression) (member 'T0 (RegsUsedBy Expression)))
(de UsesT1P (expression) (member 'T1 (RegsUsedBy Expression)))
(de UsesT2P (expression) (member 'T2 (RegsUsedBy Expression)))
(de UsesT3P (expression) (member 'T3 (RegsUsedBy Expression)))
(de AvoidsT0T1P (expression)
 (prog (regs)
       (setq regs (RegsUsedBy Expression))
       (return (and (not (member 'T0 regs))
                    (not (member 'T1 regs))))))
(de AvoidsT2T3P (expression)
 (prog (regs)
       (setq regs (RegsUsedBy Expression))
       (return (and (not (member 'T2 regs))
                    (not (member 'T3 regs))))))

(de RegsUsedBy (expression)
 (cond ((atom expression) nil)
       ((eqcar expression 'Reg) (list (cadr expression)))
       (t (nconc (RegsUsedBy (car expression))
                 (RegsUsedBy (cdr expression))))))

(put 'RegNotUsedBy 'AnyRegResolutionMacro 'RegNotUsedBy2)
(put 'RegNotUsedBy 'Pass1PseudoOp 'Eval)  % evals (RegNotUsedBy ...)

(!* "these definitions are results of running the macroexand utility over
    the definitions in 370macfn.sl which used backwuote, unquote,
    quoted-list and *list.")

(DM RegNotUsedBy (x)
 (CONS 'PROG
       (CONS '(REGS)
             (CONS (LIST 'SETQ 'REGS (LIST 'REGSUSEDBY (LIST 'QUOTE (CDR X))))
                   (CONS '(COND ((NOT (MEMBER 'T1 REGS)) (RETURN '(REG T1)))
                                ((NOT (MEMBER 'T2 REGS)) (RETURN '(REG T2)))
                                ((NOT (MEMBER 'T3 REGS)) (RETURN '(REG T3)))
                                ((NOT (MEMBER 'T0 REGS)) (RETURN '(REG T0))))
                         (CONS '(TTYPRINTF
        "%f%nError: Using (Reg T4) since can't find temp reg free throughout:")
                               (CONS (CONS 'PROGN
                                           (MAPCAR (CDR X)
                                                   (FUNCTION
                                                    (LAMBDA (EXP)
                                                     (LIST
                                                      'TTYPRINTF
                                                      "%f      %w"
                                                      (LIST 'QUOTE EXP))))))
                                     '((TTYPRINTF "%n")
                                       (RETURN '(REG T4))))))))))

(DM RegNotUsedBy2 (x)
 (CONS 'PROG
       (CONS '(REGS)
             (CONS (LIST 'SETQ
                         'REGS
                         (LIST 'REGSUSEDBY (LIST 'QUOTE (CDDR X))))
                   (CONS '(COND ((NOT (MEMBER 'T1 REGS)) (RETURN '(REG T1)))
                                ((NOT (MEMBER 'T2 REGS)) (RETURN '(REG T2)))
                                ((NOT (MEMBER 'T3 REGS)) (RETURN '(REG T3)))
                                ((NOT (MEMBER 'T0 REGS)) (RETURN '(REG T0))))
                         (CONS '(TTYPRINTF
        "%f%nError: Using (Reg T4) since can't find temp reg free throughout:")
                               (CONS (CONS 'PROGN
                                           (MAPCAR (CDDR X)
                                                   (FUNCTION
                                                    (LAMBDA (EXP)
                                                     (LIST
                                                      'TTYPRINTF
                                                      "%f      %w"
                                                      (LIST 'QUOTE EXP))))))
                                     '((TTYPRINTF "%n")
                                       (RETURN '(REG T4))))))))))



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% (dm RegNotUsedBy (x)
%  `(prog (regs)
%         (setq regs (RegsUsedBy (quote ,(cdr x))))
%         (cond ((not (member 'T1 regs)) (return '(Reg T1)))
%               ((not (member 'T2 regs)) (return '(Reg T2)))
%               ((not (member 'T3 regs)) (return '(Reg T3)))
%               ((not (member 'T0 regs)) (return '(Reg T0))))
%         (TTyPrintF
%      "%f%nError: Using (Reg T4) since can't find temp reg free throughout:")
%         (progn . ,(mapcar (cdr x)
%                           (function (lambda (exp)
%                                 `(TTyPrintF "%f      %w" ',exp)))))
%         (TTYPrintF "%n")
%         (return '(Reg T4))))
%
% (dm RegNotUsedBy2 (x)  % (RegNotUsedBy2 Register . Form)
%  `(prog (regs)
%         (setq regs (RegsUsedBy (quote ,(cddr x))))  % ignore first arg
%         (cond ((not (member 'T1 regs)) (return '(Reg T1)))
%               ((not (member 'T2 regs)) (return '(Reg T2)))
%               ((not (member 'T3 regs)) (return '(Reg T3)))
%               ((not (member 'T0 regs)) (return '(Reg T0))))
%         (TTyPrintF
%      "%f%nError: Using (Reg T4) since can't find temp reg free throughout:")
%         (progn . ,(mapcar (cddr x)
%                           (function (lambda (exp)
%                                 `(TTyPrintF "%f      %w" ',exp)))))
%         (TTYPrintF "%n")
%         (return '(Reg T4))))
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(put 'SecondRegNotUsedBy 'AnyRegResolutionMacro 'SecondRegNotUsedBy2)

(put 'SecondRegNotUsedBy 'Pass1PseudoOp 'Eval)
                % evals (SecondRegNotUsedBy ...)

(!* "these definitions are results of running the macroexand utility over
    the definitions in 370macfn.sl which used backwuote, unquote,
    quoted-list and *list.")

(Dm SecondRegNotUsedBy (x)
    (PROG (REGS)
          (SETQ REGS (EVAL (LIST 'REGNOTUSEDBY (CDR X))))
          (RETURN (LIST 'REGNOTUSEDBY REGS (CDR X)))))

(Dm SecondRegNotUsedBy2 (x)
    (PROG (REGS)
          (SETQ REGS (EVAL (LIST 'REGNOTUSEDBY (CDDR X))))
          (RETURN (LIST 'REGNOTUSEDBY REGS (CDDR X)))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% (dm SecondRegNotUsedBy (x)
%   (prog (regs)
%         (setq regs (eval `(RegNotUsedBy ,(cdr x))))
%         (return `(RegNotUsedBy ,regs ,(cdr x)))))
%
% (dm SecondRegNotUsedBy2 (x)
%   (prog (regs)
%         (setq regs (eval `(RegNotUsedBy ,(cddr x))))
%         (return `(RegNotUsedBy ,regs ,(cddr x)))))
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(put 'ThirdRegNotUsedBy 'AnyRegResolutionMacro 'ThirdRegNotUsedBy2)

(put 'ThirdRegNotUsedBy 'Pass1PseudoOp 'Eval)
                % evals (ThirdRegNotUsedBy ...)

(!* "these definitions are results of running the macroexand utility over
    the definitions in 370macfn.sl which used backwuote, unquote,
    quoted-list and *list.")

(Dm ThirdRegNotUsedBy (x)
    (PROG (NEWREG NEWREG2)
          (SETQ NEWREG (EVAL (LIST 'REGNOTUSEDBY (CDR X))))
          (SETQ NEWREG2 (EVAL (LIST 'REGNOTUSEDBY NEWREG (CDR X))))
          (RETURN (LIST 'REGNOTUSEDBY NEWREG NEWREG2 (CDR X)))))

(Dm ThirdRegNotUsedBy2 (x)
    (PROG (NEWREG NEWREG2)
          (SETQ NEWREG (EVAL (LIST 'REGNOTUSEDBY (CDDR X))))
          (SETQ NEWREG2 (EVAL (LIST 'REGNOTUSEDBY NEWREG (CDDR X))))
          (RETURN (LIST 'REGNOTUSEDBY NEWREG NEWREG2 (CDDR X)))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% (dm ThirdRegNotUsedBy (x)
%   (prog (NewReg NewReg2)
%         (setq NewReg (eval `(RegNotUsedBy ,(cdr x))))
%         (setq NewReg2 (eval `(RegNotUsedBy ,NewReg ,(cdr x))))
%         (return `(RegNotUsedBy ,NewReg ,NewReg2 ,(cdr x)))))
%
% (dm ThirdRegNotUsedBy2 (x)
%   (prog (NewReg NewReg2)
%         (setq NewReg (eval `(RegNotUsedBy ,(cddr x))))
%         (setq NewReg2 (eval `(RegNotUsedBy ,NewReg ,(cddr x))))
%         (return `(RegNotUsedBy ,NewReg ,NewReg2 ,(cddr x)))))
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               FUNCTIONS USED BY CMACROS                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de !*SS1 (OpCode ArgOne ArgTwo L1)
 (prog (L2)
       (cond ((or (EqCar ArgOne 'DoubleIndexed)
                  (EqCar ArgTwo 'DoubleIndexed))
              (return (list (list 'Problem 'SS1 OpCode ArgOne ArgTwo L1)))))
       (cond ((null L1)
              (progn (setq L1 (ByteLengthOf ArgOne))
                     (setq L2 (ByteLengthOf ArgTwo))
                     (cond ((not (equal L1 L2))
                            (TTYPrintF "%f !*SS1: %p L(%p)= %p, L(%p)= %p %n"
                                       OpCode ArgOne L1 ArgTwo L2))))))
       (return (list (list OpCode
                           (list 'LengthBased L1 ArgOne)
                           ArgTwo)))))

(de !*SS2 (OpCode ArgOne ArgTwo)
 (prog (L1 L2)
       (cond ((or (EqCar ArgOne 'DoubleIndexed)
                  (EqCar ArgTwo 'DoubleIndexed))
              (return (list (list 'Problem 'SS2 OpCode ArgOne ArgTwo L1)))))
       (setq L1 (ByteLengthOf ArgOne))
       (setq L2 (ByteLengthOf ArgTwo))
       (cond ((not (equal l1 l2))
              (TTYPrintF "%f !*SS2: %p L(%p)= %p, L(%p)= %p %n"
                         OpCode ArgOne L1 ArgTwo L2)))
       (return (list (list opcode
                           (list 'LengthBased L1 ArgOne)
                           (list 'LengthBased L2 ArgTwo))))))


(de !*Link (FunctionName FunctionType NumberOfArguments)
 (prog (ResultingCode!* OpenCodeSequence)
  (return
   (cond ((FlagP FunctionName 'ForeignFunction)
          (list (list '!*ForeignLink FunctionName FunctionType
                      NumberOfArguments)))
         ((setq OpenCodeSequence (get FunctionName 'OpenCode))
          OpenCodeSequence)
         (t (cons (list '!*Move NumberOfArguments '(Reg NargReg))
                  (CMacroPatternExpand (list FunctionName)
                    (get '!*Link 'CMacroPatternTable))))))))


(de !*LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
 (prog (ResultingCode!* OpenCodeSequence)
  (return
   (cons (list '!*DeAlloc DeAllocCount)
    (cond ((FlagP FunctionName 'ForeignFunction)
           (list (list '!*ForeignLink FunctionName FunctionType
                       NumberOfArguments)
                 '(!*Exit 0)))
          (t (cond ((setq OpenCodeSequence (get FunctionName 'ExitOpenCode))
                    OpenCodeSequence)
                   ((setq OpenCodeSequence (get FunctionName 'OpenCode))
                    (Append OpenCodeSequence (list '(!*Exit 0))))
                   (t (cons (list '!*Move NumberOfArguments '(Reg NargReg))
                            (CMacroPatternExpand (list FunctionName)
                              (get '!*LinkE 'CMacroPatternTable)))))))))))

(de !*JumpIf (Arg1 Arg2 Label Test ReverseTest)
 (prog (ResultingCode!*)
       (return
        (CmacroPatternExpand
         (list (ResolveOperand '(Reg T1) Arg1)
               (ResolveOperand '(Reg T2) Arg2)
               (ResolveOperand '(Reg Error) Label)
               Test
               ReverseTest)
         (get '!*JumpIf 'CmacroPatternTable)))))

% last argument is used if args are reversed in comparison

(de !*JumpWGeq      (Lbl Arg1 Arg2) (!*JumpIf Arg1 Arg2 Lbl 'BNL 'BNH))

(de !*JumpWGreaterP (Lbl Arg1 Arg2) (!*JumpIf Arg1 Arg2 Lbl 'BH  'BL))

(de !*JumpWLeq      (Lbl Arg1 Arg2) (!*JumpIf Arg1 Arg2 Lbl 'BNH 'BNL))

(de !*JumpWLessP    (Lbl Arg1 Arg2) (!*JumpIf Arg1 Arg2 Lbl 'BL  'BH))

(De !*Compare (Arg1 Arg2)
       (Expand2OperandCmacro Arg1 Arg2 '!*Compare))

(de !*CLCL (ArgOne ArgTwo)
 (cond ((member 'T0 (RegsUsedBy ArgTwo))
        (list (list 'PROBLEM 'Compare ArgOne ArgTwo)))
       (t
        (list (LoadAddress '(Reg T0) ArgOne)
              (LoadAddress '(Reg T2) ArgTwo)
              (list 'LA '(Reg T1) (ByteLengthOf ArgOne))
              (list 'LA '(Reg T3) (ByteLengthOf ArgTwo))
              (list 'CLCL '(Reg T0) '(Reg T2))))))

(De !*MaybeMVC (Source Dest Length)
 (cond ((or (EqCar Source 'DoubleIndexed)
            (EqCar Dest 'DoubleIndexed))
        (cond ((greaterp Length 4)
               (list (list 'Problem 'MVC Source Dest Length)))
              (t (list (list '!*Move Source (RegNotUsedBy Dest))
                       (list '!*Move (RegNotUsedBy Dest) Dest)))))
       (t (!*SS1 'MVC Dest Source Length))))


% (de !*IndexedLoad (arg1 arg2 arg3)
%  (Expand3OperandCmacro Arg1 Arg2 Arg3 '!*IndexedLoad ))
%
% (de !*IndexedStore (arg1 arg2 arg3)
%  (Expand3OperandCmacro Arg1 Arg2 Arg3 '!*IndexedLoad ))

(de !*MVCL (Dest Source)
 (cond ((member 'T0 (RegsUsedBy Source))
        (list (list 'Problem 'Move Source Dest)))
       (t
        (list (LoadAddress '(Reg T0) Dest)
              (LoadAddress '(Reg T2) Source)
              (list 'LA '(Reg T1) (ByteLengthOf Dest))
              (list 'LA '(Reg T3) (ByteLengthOf Source))
              (list 'MVCL '(Reg T0) '(Reg T2))))))

(de LoadAddress (Reg x) (list 'LA reg x))  % needs work


(de !*ForeignLink (FunctionName FunctionType NumberOfArguments)
 (prog (MoveSequence PushSequence TempArgCount)
       (CodeDeclareExternal FunctionName)
       (return
        (list '(PROBLEM "Foriegn Functions don't work yet")
              '(STM (Reg 1) (Reg 5) (WVar FFArg1))   % args 1-5
              '(MVC (LengthBased 44 (WVar FFArg6)) (WArray ArgumentBlock))
                               % 44 = 4 * (16-5)   % args 11-16
              (list 'L '(Reg LinkReg)
                    (list 'Lit (list 'ForeignFunction FunctionName)))
              '(BALR (Reg 14) (Reg 15))  % why 15 ??
              '(MVC (LengthBased 44 (WArray ArgumentBlock)) (WVar FFArg6))
              '(LM (Reg 1) (Reg 5) (WVar FFArg1)) % what about 5-16????
              '(PROBLEM "End of Foriegn function call")
                ))))

(de Expand3OperandCMacro (Arg1 Arg2 Arg3 CMacroName)
  (prog (ResultingCode!*)
    (return (CMacroPatternExpand (list (ResolveOperand '(Reg T1) Arg1)
                                       (ResolveOperand '(Reg T2) Arg2)
                                       (ResolveOperand '(Reg T3) Arg3))
                                 (get CMacroName 'CMacroPatternTable)))))

(de !*FetchReturnAddress (Arg1)
  (Expand1OperandCMacro Arg1 '!*FetchReturnAddress))

(de !*SetReturnAddress (Arg1)
  (Expand1OperandCMacro Arg1 '!*SetReturnAddress))

(de !*GetCallInfo (Arg1 Arg2)
  (Expand2OperandCMacro Arg1 Arg2 '!*GetCallInfo))

(de !*SetCallInfo (Arg1 Arg2)
  (Expand2OperandCMacro Arg1 Arg2 '!*SetCallInfo))

(de !*PopCallInfo (Arg1 Arg2)
  (cond ((and (PairP Arg2) (cdr Arg2))
         (progn (ErrorPrintF "%f***** Too many args to PopCallInfo: %w %w%n"
                             Arg1 Arg2)
                (list 'Problem "*PopCallInfo" Arg1 Arg2)))
        ((atom Arg2)
           (Expand2OperandCmacro Arg1 Arg2 '!*PopCallInfo))
        (t (Expand2OperandCmacro Arg1 (car Arg2) '!*PopCallInfo))))

(de !*PushCallInfo (Arg1 Arg2)
  (cond ((and (PairP Arg2) (cdr Arg2))
         (progn (ErrorPrintF "%f***** Too many args to PushCallInfo: %w %w%n"
                             Arg1 Arg2)
                (list 'Problem "*PushCallInfo" Arg1 Arg2)))
        ((null (car Arg2))
         (progn (ErrorPrintF "%f***** Too few args to PushCallInfo: %w %w%n"
                             Arg1 Arg2)
                (list 'Problem "*PushCallInfo" Arg1 Arg2)))
        ((atom Arg2)
           (Expand2OperandCmacro Arg1 Arg2 '!*PushCallInfo))
        (t (Expand2OperandCmacro Arg1 (car Arg2) '!*PushCallInfo))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                 SHIFT/MASK FUNCTIONS                                  %
%                                                                       %
% BitMask will provide a mask of ones beginning at bit (start) and      %
% continuing for (size) number of bits.                                 %
%                                                                       %
% NOTE:  This is NOT the convention for BitMask on the DEC 10/20.       %
%        That BitMask uses start and end.                               %
%                                                                       %
% The bit numbering convention is that the most significant bit is bit  %
%  0, the least is bit 31.                                              %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(ds BitMask (StartingBit Length)
 (land (lsh 16#FFFFFFFF (minus StartingBit))
       (lsh 16#FFFFFFFF
            (difference BitsPerWord (plus2 StartingBit Length)))))

(ds NegMask (Length) (lsh 16#FFFFFFFF Length))

(ds ShiftAmt (StartingBit Length)
 (difference BitsPerWord (plus2 StartingBit Length)))  %always positive

(ds OneByteBitMask (StartingBit Length)
 (land (lsh 16#FF (minus StartingBit))
       (lsh 16#FF (difference BitsPerByte (plus2 StartingBit Length)))))

(ds OneByteShiftAmt (StartingBit Length)
 (difference BitsPerByte (plus2 StartingBit Length)))  %always positive

(ds MakeTag (Tag) (lsh (land Tag 16#FF) 24))

% better def in <psl-compiler>data-machine.red
% (ds MkCode (x) (MkItem 15 x))

(de TaggedLabel (X) (EqCar X 'Label))

%------------------------------------------------------------------------
% Given an OBJECT tag (Not an assembler tag!) Look up on the property
% list to find out the actual numeric value included in the ASM output.
% The Wconst property name referances the specific number.
%------------------------------------------------------------------------

(de TagNumber (X)
  (cond ((IDP X)     (get 'ID   'WConst))
        ((PairP X)   (get 'PAIR 'WConst))
        ((StringP X) (get 'STR  'WConst))
        ((InumP X)   (cond ((MinusP X) 16#FF) (t 0)))
        ((FloatP X)  (get 'FltN 'WConst))
        ((VectorP X) (get 'VECT 'WConst))
        ((FixP X)    (get 'FixN 'WConst))
        ((CodeP X)   (get 'Code 'WConst))
        (t (StdError (BldMsg "Unknown type of %r" X)))
 ))


% these were translated from RLISP format versions...
% superceded by portable versions (??)
% (de !*LamBind (Regs FLst)
%  (prog (X Y)
%        (setq FLst (reverse (cdr FLst)))
%        (setq Regs (reverse (cdr Regs)))
%        (while FLst
%               (progn
%                (cond ((null Regs) (setq X 0))
%                      (t (progn (setq X (cadr (car Regs)))
%                                (setq Regs (cdr Regs)))))
%                (setq Y (cons (list 'HalfWord X (list 'IdLoc (cadar FLst)))
%                            y))
%                (setq FLst (cdr FLst))))
%        (return (cons '(!*Call FastBind) Y))))
%

% todo -- change B to !*Jump, fix !*Jump to not become !*JCall,
%         change Indexed to Memory, use offset of baselbl

(de !*JumpOn (Register LowerBound UpperBound LabelList)
 (prog (ExitLbl BaseLbl Result LowerOffset)
       (setq ExitLbl (gensym))
       (setq BaseLbl (gensym))
       (setq Result (cons nil nil))
       (tconc Result (list '!*JumpWLessP
                           (list 'Label ExitLbl) Register LowerBound))
       (tconc Result (list '!*JumpWGreaterP
                           (list 'Label ExitLbl) Register UpperBound))
       (tconc Result (list '!*Move Register '(Reg T1)))
       (tconc Result '(!*WTimes2 (Reg T1) (WConst AddressingUnitsPerItem)))
       (setq LowerBound (ResolveWConstExpression LowerBound))
       (cond ((not (Numberp LowerBound))
              (StdError
                (BldMsg "!*Jumpon: non-numeric Lowerbound %w" LowerBound))))
       (setq LowerOffset (times AddressingUnitsPerItem LowerBound))
       (cond ((greaterp CurrentLocation!* LowerOffset)
              (tconc Result
                     (list 'B % don't use !*JUMP -- turns this into jcall
                           (list 'DoubleIndexed
                                 '(Reg T1) '(ImplicitReg BaseR)
                                 (list 'Difference BaseLbl LowerOffset)
                                 ))))
             (t
              (progn    % in IBM terms, *-LowerOffset is not addressible
               (tconc Result (list '!*WDifference '(Reg T1) LowerOffset))
               (tconc Result
                      (list 'B
                            (list 'DoubleIndexed
                                  '(Reg T1) '(ImplicitReg BaseR)
                                  BaseLbl))) )))
                             % don't use !*JUMP -- turns this into jcall
       (tconc Result BaseLbl)
       (foreach x in LabelList do (tconc Result (list '!*Jump x)))
       (tconc Result ExitLbl)
       (tconc Result (list 'DS "0H"))  % avoid null statements
       (return (car Result))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(fluid '(!*WritingFaslFile))

(de InternallyCallableP (X)
 (progn (cond ((TheirInternallyCallableP X)
               (* TTYPrintF "%f*** Ignoring internal declaration for %w%n" X)))
        nil))

(de TheirInternallyCallableP (X)        % only when writing a file
  (and (or !*WritingFaslFile (not (FUnBoundP 'AsmOut)))
       (or !*FastLinks
           (and !*R2I (memq X EntryPoints!*))
           (FlagP X 'InternalFunction)
           (FlagP X 'FastLink))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de !&DispWConstP (X)
  (DispInumP (ResolveWConst X)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       end of file                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
