%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC/IBMA-SPEC.SL
% Title:          Some special things for IBM XA /370  PSL compiler
% Author:         Winfried Neun , ZIB Berlin
% Created:        9 Feb 1991
% Status:         Experimental
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Revisions:
% 22-August-1990 (Winfried Neun)
%   added code that attempts to compile setq to frame directly with the
%   frame element (which is a register in most cases) as &dest

% turn off lapopt in faslend

(fluid '(*lapopt))

(setq *lapopt t)

(when (funboundp 'old-faslend) (copyd 'old-faslend 'faslend))

(de faslend ()
    (prog (*lapopt)
       (return (old-faslend))))

(put 'bittable 'opencode
      '((*move (reg 2) (reg 3))
	(srl (reg 3) 2 ) % offset
	(sr (reg 5) (reg 5))
        (ic (reg 5) (doubleindexed (reg 1) (reg 3) 0))
	(*move (reg 2) (reg 4))
        (*Wand (reg 4) 3)
        (sll (reg 4) 1)
	(*move 6 (reg 3))
        (sr (reg 3) (reg 4))   % 6 - fraction
%       (LCR (reg 4)(reg 4))      %   fraction
        (srl (reg 5) (displacement (reg 3) 0))
        (*Wand (reg 5) 3 )
	(*move (reg 5) (reg 1))))


% This code is intended for filling the 'holes' in the frame.
% e.g. in the following function
%
%  (de hugo (x) (cond ((null x) (call-another))
%                     (t (prog (a b c) ....))
%
% a frame of length 4 will be allocated, but it will not be preset (by nil)
% completely at entering the function but the frame elements for a b and c
% will be set (to nil) if the prog is entered.
% So if the (null x) case is taken we have 3 unset frame elements on the stack
% which contain random information and this is dangerous for gc.
% to resolve that problem, the following function will work on the input
% to pass-1-lap and insert nil settings where needed.

(fluid '(&fillframeholes &frame-numbers))

(de &fillframeholes (code)
   (prog (allo cmactoextern externfound ll lll nn linearcode restcode)
      (when (not &fillframeholes) (return code))
      (setq cmactoextern
                  '(*link *linke *call *jcall *cons *xcons *fastmkfltn))
             % trivial tests whether we have to work at all
      (unless (setq allo (findcmacro code '*alloc)) (return code))
      (setq ll cmactoextern)
      (while (and (not externfound) ll)
            (setq externfound (findcmacro code (car ll)))
            (setq ll (cdr ll)))
      (when (not externfound) (return code))

            % so we have to work, starting at the *alloc

      (setq nn (cadar allo))  %the frmae size
      (setq linearcode t)
      (setq restcode (cdr allo))

      (setq &frame-numbers nil)

      (while (and linearcode restcode)
         (setq linearcode (lookatinstruction (car restcode)))
         (setq restcode (cdr restcode)))

      (when linearcode (return code))

      (setq lll (setq ll (cons nil nil)))

             % search frame that have not been set, prepare code,
             % put into place after *alloc

      (for (from i 1 nn 1) (do
            (unless (memq i &frame-numbers)
                  (rplaca ll `(*move (reg nil) (frame ,i)))
%                 (prin2 "++ in ") (prin2 name&)
%                 (prin2 " presetting frame ")
%                 (prin2t i)
                  (push 'nil ll))))

      (when (eq lll ll) (return code))
      (rplacd lll (cdr allo))

      (rplacd allo (cdr ll))
      (return code)))

(de lookatinstruction (instr)
      (cond ((atom instr) t)
            ((and (eq (car instr) '*move)
                  (eqcar (caddr instr) 'frame))
                       (push (cadr (caddr instr)) &frame-numbers) 17)
            ((memq (car instr) '(*link *call *linke *jcall *jumpintype
                 *jumpeq *jump  *JumpType *JumpNotType *JumpNotInType
                 *jumpnoteq *jumpwlessp *jumpwgreaterp *jumpwleq *jumpwgeq
                 *jumpon)) nil)
            (t t)))


(de findcmacro (list pat)
     (cond ((atom list ) nil)
           ((eqcar (car list) pat) list)
           (t        (findcmacro (cdr list) pat ))))

(setq &fillframeholes t)

(de faslout (fil)
  (setq modulename* fil)
  (unless *quiet_faslout
    (if (not (funboundp 'begin1))
      (progn (prin2t "FASLOUT: IN files; or type in expressions")
             (prin2t "When all done execute FASLEND;"))
      (progn (prin2t "FASLOUT: (DSKIN files) or type in expressions")
             (prin2t "When all done execute (FASLEND)"))))

  (when (not (eq (car (reverse (explode2 fil))) '!) ))
        (setq fil (bldmsg faslfilenameformat* fil)))

  (setq codeout* (binaryopenwrite fil))
  (codefileheader)
  (setq dfprint* 'dfprintfasl)
  (setq fasl-preeval* 'faslpreeval)
  (setq *writingfaslfile t)
  (setq *defn t)
  (setf fasl-idnumber-property* (intern (concat "FASL-IDNUMBER-" (stringgensym))
))
  t
  )
