%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:SYS-CONSTS.SL
% Description:  HPUX PSL constants needed in the compiler
% Author:       Cris Perdue
% Created:      24-Feb-84
% Modified:     27-Jun-84 09:31:48
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 04-Mar-90 (Winfried Neun)
%  Added unix to  system_list*
% 04-Apr-87 (Leigh Stoller)
%  Added 68020 to system_list*
% 17-Jul-86 (Leigh Stoller)
%  Changed number of maxrealregs to 6 for 68020. T regs were being clobbered.
% 21-Jun-84 (Vicki O'Day)
%  Changed system-list and magic numbers for HP-UX.
%  Took out import of tags.
% 13-Jun-84 12:40:52 (Jim Ambras/CRC)
%  Increased size of hash-table-size and maxsymbols.
% 11-May-84 07:32:09 (Brian Beach)
%  Added some constants for the kernel.
% 30 Mar 1984 1528-PST (Nancy Kendzierski)
%  Added (compiletime (load sys-macros)) for the wconst declarations
%  and (load wdeclare) since use of sys-macros at compiletime generates
%  initcode that needs wdeclare.
% 28 Mar 1984 1717-PST (Nancy Kendzierski)
%  Added declare-wconst of faslin-magic and faslout-magic.
%  HP9836 Pascal = 299 and 299
% 12 Mar 1984 1654-PST (Nancy Kendzierski)
%  Modified maxrealregs and maxargs to match new D-register model.
% 28 Feb 1984 1341-PST (Cris Perdue)
%  Added maxargs and maxrealregs to this file from global-data.
% 21 Feb 1984 1217-PST (Nancy Kendzierski)
%  Added remprop of 'wconst properties to ensure the new declarations will
%  take effect.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(system_list*))

(setq system_list* '(mc68000 mc68020 apollo bobcat sun unix IEEE))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MACHINE DEPENDENT CONSTANTS FOR THE MC68000                           %
% Four BYTES per LONGWORD (32 bits).                                    %
% Four CHARACTERs per LONGWORD (32 bits).                               %
% Stack grows down, rather than UP, used to control Stack Scan loops.   %
% Items are placed on stack with -(SP)xx addressing  (predecrement).    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(define-constant faslin-magic  399)
(define-constant faslout-magic 399)


(define-constant tagstartingbit                 0) 
(define-constant tagbitlength                   5)
(define-constant infstartingbit                 5)
(define-constant infbitlength                   27) 
(define-constant addressingunitsperitem         4)
(define-constant charactersperword              4) 
(define-constant bitsperword                    32)
(define-constant bitsperword2                   16)
(define-constant addressingunitsperfunctioncell 4) 
(define-constant stackdirection                 -1)



(define-constant maxargs     14)   % Dreg model  (Areg was 15)
%(define-constant maxrealregs 14)   % Dreg model  (Areg was 5)
%
% The 68020 needs to do tag stripping. This introduces the use of registers
% t1 and t3, which are (reg 7) and (reg 13), when doing car and cdr. No
% problem if there are less then 6 arguments. Above that, they get clobbered
% when t1 and t3 are used. This will ensure that the t regs are always safe
% to use.
%
(define-constant maxrealregs     6)

% At present this is the single machine-dependent character
% (for ASCII machines).  Defined here in case cross-compilation
% is being done.  Defined for the kernel in php:scan-table.sl. /csp

(put 'eof 'charconst (char (control d)))


(define-constant maxtokensize          5000)
(define-constant bndstksize            2000)
(define-constant catchstacksize        400)
(define-constant hash-table-size       18013)
(define-constant savesize              4)
(define-constant maxargblock           15)
(define-constant stacksize             20000)
(define-constant maxsymbols            15000)
(define-constant compressedbinaryradix 16)
