%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXK:SYS-SUPPORT.SL
% Description:  Sparc-specific kernel support functions
% Author:       Chris Burdorf
% Created:      10-May-88
% Modified:       
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
%               Changes to All functions by W. Neun, ZIB Berlin, Nov-15-88
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ADDRESSAPPLY0 is defined as open code.

% BITTABLE is defined as open code.

% changes in   Symfnc jumps:                     W. Neun
%     CALLER:                   SYMFNC Vect         CALLEE
%       Call  SYMFNC+4*id  -->  BA   target   ->    
%                               SETHI idnumber,%o0   in delay slot
%
%   Note : the callee gets the idnumber shifted by ten bits in i0/o0
%
%       This allows to use normal calling sequence (instead of two calls)
%       (it also limits the code space to 8Mbyte, hope that does not matter)

(lap '((*entry plantunbound expr 1)
       (*alloc 0)
       (*move (reg 1) (reg 5))                  %% the id number
       (sll (reg 1) 2 (reg 1))
       (*wplus2  (reg 1) (fluid symfnc))        %% get cell
       (*movex undefinedfunctioninstruction (reg 2))  %% get address of entry
       (*movex (deferred (reg 2)) (reg 2))      % get contents
       (st (reg 2) (deferred (reg 1)))          %% 
       (*exit 0)
       undefinedfunctioninstruction
       (word "undefinedfunction")))

(lap '((*entry plantlambdalink expr 1)
       (*alloc 0)
       (*move (reg 1) (reg 5))
       (sll (reg 1) 2 (reg 1)) % multiply ID number by SymFnc scale factor
       (*wplus2 (reg 1) ($fluid symfnc))
       (*movex lambdalinkinstruction (reg 2))
       (*movex (deferred (reg 2)) (reg 2))    % get contents of lamdalinkinst.
       (st (reg 2) (deferred (reg 1)))       %% 
       (*exit 0)
       lambdalinkinstruction
       (word "compiledcallinginterpreted")))

(compiletime (flag '(plantcodepointer-normal plantcodepointer-plus4
		     plantcodepointer-special) 'internalfunction))

(de plantcodepointer (x y)
   (if (or (eq x (inf 'compiledcallinginterpreted))
	   (eq x (inf 'undefinedfunction)))
	(plantcodepointer-special x y)
      (prog (startofcode transfer)
	(setq startofcode (getmem y))
	(setq transfer
	    (or (eq 2#0001000010 (wshift startofcode -22))   % BA
		(eq 2#0011000010 (wshift startofcode -22))   % BA,a
		(and (eq 2#10 (wshift startofcode -30))      % JMPL
		     (eq 2#111000 (wshift (wshift startofcode 7) -26)))))
       % (if transfer (plantcodepointer-normal x y)
       %        (plantcodepointer-plus4 x (wplus2 4 y) startofcode))
	(plantcodepointer-normal x y))))
		% flush the cache area of the implanted call

(lap '((*entry plantcodepointer-normal expr 2)
       (*alloc 0)
       (*move (reg 1) (reg 5))                    %% the id number
       (sll (reg 1) 2 (reg 1)) % multiply ID number by SymFnc scale factor
       (*wplus2 (reg 1) ($fluid symfnc))
       (st (reg 2) (deferred (reg 1)))
       (*exit 0)
)) 

% this one is for compilecallinginterpreted
(lap '((*entry plantcodepointer-special expr 2)
       (*alloc 0)
       (*move (reg 1) (reg 5))                    %% the id number
       (sll (reg 1) 2 (reg 1)) % multiply ID number by SymFnc scale factor
       (*wplus2 (reg 1) ($fluid symfnc))
       (st (reg 2) (deferred (reg 1)))
       (*exit 0)
))

(lap '((*entry plantforeigncodepointer expr 2)
       (sll (reg 1) 3 (reg 5)) % multiply ID number by SymFnc scale factor
       (*wplus2 (reg 5) (fluid symfnc))

       (sub (reg 2) (reg 5) (reg 2))            %% make offset
       (srl (reg 2) 2 (reg 2))                  %% make 22 bit offset
       (sll (reg 2) 10 (reg 2))
       (srl (reg 2) 10 (reg 2))
       (sethi (hi 16#10800000) (reg 3))         %% BA   here.
       (sethi (hi 16#01000000) (reg 4))         % its a NOP
       (add (reg 2) (reg 3) (reg 2))
       (st (reg 2) (deferred (reg 5)))
       (*movex (reg 4) (displacement (reg 5) 4)) % move nop into location after
	                                         % call.
       (*exit 0)
))

(lap '((*entry undefinedfunction expr 0)
	(*move  ($fluid symfnc) (reg t3))
	(*Move (sethiidloc undefinedfunctionaux) (reg t4))
	(srl (reg t4) 8 (reg t4))
	(*move (displacement (reg t3) (reg t4)) (reg t4))
	(jmpl (displacement (reg t4) 0) (reg 0))
	(nop)))

(lap '((*entry compiledcallinginterpreted expr 0)
	(*move  ($fluid symfnc) (reg t3))
	(*Move (sethiidloc compiledcallinginterpretedaux) (reg t4))
	(srl (reg t4) 8 (reg t4))
	(*move (displacement (reg t3) (reg t4)) (reg t4))
	(jmpl (displacement (reg t4) 0) (reg 0))
	(nop)))

(lap '((*entry undefinedfunctionaux expr 0)
       %. Error Handler for non code                                       
       %                                                                   
       % (reg t1) <- caller's address                                      
       %                                                                   
       (*alloc 1)
       (*wshift (reg i0) -8)      % to annul the sethi shift and byte address
       (*move (reg i0) (frame 1))
       (*move (quote "Undefined function called: ") (reg 1))
       (*call console-print-string)   %return address is r15+8
       (*wplus2 (frame 1) (fluid symnam)) %add offset to symnam
				     % to get real address of name.
       (*move (frame 1) (reg 1))
       (*move (deferred (reg 1)) (reg 1))  %move string object to r1
       (*call console-print-string)
       (*move (wconst -1) (reg 1))
       (*linke 0 exit-to-caller expr 1)))


(de kernel-fatal-error (message)
  (console-print-string "FATAL ERROR:") 
  (console-print-string message)
  (console-newline)
  (exit-to-caller -1)
  )

(de exit-to-caller (sta)
   (external_exit sta))

(lap '((*entry WQuotient expr 2)
       (*alloc 3)
       (mov (reg 1) (reg o0))
       (call (foreignentry div) 2)
       (mov (reg 2) (reg o1))

       (mov (reg o0) (reg 1))
       (*exit 3)))

 (lap '((*entry WRemainder expr 2)
	(*alloc 3)
	(mov (reg 1) (reg o0))
	(call (foreignentry rem) 2)
	(mov (reg 2) (reg o1))

	(mov (reg o0)  (reg 1))
	(*exit 3)))
 
(lap '((*entry BitTable expr 2)
       (*alloc 3)
       (mov (reg 1) (reg o0))
       (call (foreignentry bit_table) 2)
       (mov (reg 2) (reg o1))
       (mov (reg o0) (reg 1))
       (*exit 3)))

(lap '((*entry PutBitTable expr 3)
       (*alloc 3)
       (*movex (reg 1) (reg o0))
       (mov (reg 2) (reg o1))
       (call (foreignentry put_bit_table) 3)
       (mov (reg 3) (reg o2))

       (mov (reg o0) (reg 1))
       (*exit 3)))


(lap '((*entry WTimes2 expr 2)
       (*alloc 3)
       (mov (reg 2) (reg o0))
       (call (foreignentry mul) 2)  % C mulitply routine.
       (mov (reg 1) (reg o1))

       (mov (reg o0) (reg 1))   %store return value in r1
       (*exit 3)))
	   
(fluid '(*second-value*))

(lap '((*entry WxTimes2 expr 2)   % extended times
       (*alloc 3)     % returning in reg 1 and *second-value*(upper bits)
       (mov (reg 2) (reg o0))
       (call (foreignentry !.mul) 2)  % extended mulitply routine.
       (mov (reg 1) (reg o1))
 
       (mov (reg o0) (reg 1))   %store return value in r1
       (*move (reg o1) ($fluid *second-value*))
       (*exit 3)))

(lap '((*entry WuxTimes2 expr 2)   % extended unsigned times
       (*alloc 3)     % returning in reg 1 and *second-value*(upper bits)
       (mov (reg 2) (reg o0))
       (call (foreignentry !.umul) 2)  % extended mulitply routine.
       (mov (reg 1) (reg o1))

       (mov (reg o0) (reg 1))   %store return value in r1
       (*move (reg o1) ($fluid *second-value*))
       (*exit 3)))

(lap '((*entry wtimesdouble expr 2)
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The hig order bits are placed in fluid variable.
       (*alloc 2)
       (mov (reg 2) (reg o0))
       (call (foreignentry !.mul) 2)  % extended mulitply routine.
       (mov (reg 1) (reg o1))
      % now we have 32 low bits in REG o0, 32 high bits in REG o1
       (*move (reg o0) (reg 1))
       (*move (reg o1) (reg 3))
       (*WSHIFT (reg 3) 2)
       (*MOVE (reg 1) (reg 2))
       (*WSHIFT (reg 2) -30)
       (*WOR (reg 3)(reg 2))
       (*MOVE (reg 3)($FLUID *second-value*))
       (*WSHIFT (reg 1) 2)
       (*WSHIFT (reg 1) -2)
       (*exit 2)))


(de Wdivide (x y) (setq *second-value* (wremainder x y)) (wquotient x y))

(commentoutcode lap '((*entry Wdivide expr 2) % problems under Solaris 2.5
       (*alloc 3) 
       (*move (reg 1) (frame 1)) % remember(the sign)of dividend
       (mov (reg 2) (reg o1))
       (call (foreignentry !.div) 2)  % divide routine.
       (mov (reg 1) (reg o0)) 
 
       (*move (reg o0) (reg 1)) 
       (subcc (reg o3) (reg 0) (reg 0))
       (bge got_result)
       (nop)
       (add (reg o3) (reg o1) (reg o3))
     got_result
       (*move (frame 1) (reg 5))
       (subcc (reg 5) (reg 0) (reg 0))
       (bge hugo)
       (nop)
       (neg (reg o3))
      hugo
       (*move (reg o3)  ($fluid *second-value*))
	       % impure, from architecture manual
       (*exit 3)))

(lap '((*entry unix-profile expr 4)
       (*alloc 1)
       (*move (reg 1) (reg o0))
       (*move (reg 2) (reg o1))
       (*move (reg 3) (reg o2))
       (call (foreignentry profil) 4)
       (*move (reg 4) (reg o3))

       (*move 0 (reg 1))
       (*exit 1)))
