%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% File:         Arithmetic.sl - Generic arithmetic routines for PSL
%               This is a (very) CRAY specific version
% Author:       Herbert Melenk and Winfried Neun
%               Konrad Zuse-Zentrum Berlin
%               derived from NBARITH.RED
%               by Eric Benson and Martin Griss
% Date:         9 August 1986
% Copyright (c) 1987 Konrad-Zuse Zentrum Berlin for CRAY version
% Copyright (c) 1982 University of Utah
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions:
% 25-Aug-1987 W. Neun Corrected -betahi and -betalow for + and -.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% A UNARY operation (UN x) is done as:
%   Procedure UN x;
%     If mbetap x then <<x:=WUN x; if IntRangeP x then x else Sys2Int x>>
%       else UN-hard(x);
% A UNARY predicate   (UNP x) is done as:
%   Procedure UNP x;
%     If mbetap x then WUNP x
%       else UNP-hard(x);
% A BINARY operation (BIN x y) is done as:
%   Procedure BIN(x,y);
%     If mbetap x and mbetap y
%       then <<x:=WBIN(x,y);
%              if IntRangeP x then x else Sys2Int x>>
%      else BIN-hard(x,y);
% A BINARY predicate (BINP x y) is done as:
%   Procedure BINP(x,y);
%     If mbetap x and mbetap y then WBINP(x,y)
%      else BINP-hard(x,y);
% IN some "safe" cases, mbetap can become IntP (beware of *)
% In others, mbetap(y) may be too weak (eg, Lshift and Expt)
% Note: Loading BIG or Nbig-V is supposed to define (or redefine)
%        the functions:
%             Coerce1 and Coerce2
%             Sys2Big
%             FloatFrombignum
%             Sys2Int
%             FloatFix
 
(compiletime (fluid '(ARITHARG ArithArgs0 ArithArgs1 STATICFloat
                      STATICBIG*)) )
%ArithArg 0/1 reside in Tregs!!
 
(compiletime (flag '(wgcdn wgcdn-aux wdivide coerce2-hardcase
                                 )            'internalfunction))
 
(setq aritharg    (mkwrds (gtwarray 2)))
(setq staticFloat (mkwrds (gtwarray 3)))
(setq staticbig*  (gtwarray 5))
 
% betap and beta2p are smacros now.
% same code as in bignums but with dummy constants
% in this way, inline-code will be produced
 
(compiletime
 (progn
 
(ds STATICINTFloat (X)
      (progn
         (wputv staticFloat 1 (*wFloat x))
         (mkfltn staticFloat)))
 
(put '*Fgreaterp 'opencode
     '((*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (fsub (reg s0) (reg 2) (reg 1))
       (*move (QUOTE t) (reg 1))
       (jsm greater)
       (*move (QUOTE nil) (reg 1))
     greater
)     )
 
(put '*Flessp  'opencode
     '((*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (fsub (reg s0) (reg 1) (reg 2))
       (*move (QUOTE nil) (reg 1))
       (jsp greater1)
       (*move (QUOTE t) (reg 1))
     greater1
)     )
 
(put '*Fquotient  'opencode
     '((*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (hrecip (reg 3) (reg 2))
       (fmul  (reg 1) (reg 1) (reg 3))
       (imul  (reg 2) (reg 2) (reg 3))
       (fmul  (reg 1) (reg 1) (reg 2))
)     )
 
(put '*Fplus2 'opencode
     '((*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (fadd (reg 1) (reg 1) (reg 2))
)     )
 
(put '*Fdifference 'opencode
     '((*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (fsub (reg 1) (reg 1) (reg 2))
)     )
 
(put '*Ftimes2 'opencode
     '((*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (*move (reg 2) (reg a6))
       (*move (memory (reg a6) 1) (reg 2))
       (rmul (reg 1) (reg 1) (reg 2))
)     )
(put '*WFloat  'opencode
     '((sconst06 (reg 2))
       (sub (reg 1) (reg 2) (reg 1))
       (fsub (reg 1) (reg 2) (reg 1))
)     )
 
(put '*Wfix  'opencode
     '((*move (reg 1) (reg a6))
       (*move (memory (reg a6) 1) (reg 1))
       (sconst06 (reg 2))
       (fsub (reg 1) (reg 2) (reg 1))
       (sub (reg 1) (reg 2) (reg 1))
)     )
 
(ds mbetap (x)
       (and (wleq x betahi*) (wgeq x betalow*)))
 
(ds mbeta2p (x y) (and (mbetap x) (mbetap y)))
 
))%end compiletime
 
(fluid '(betahi* betalow*))
 
(setq betahi* (wor (wshift 16#3ffffff 32) 16#ffffffff))
(setq betalow* (wshift 31 59))
 
% Test tagged number is in Beta Range when bignum loaded
% Will redefine if NBIG loaded
 
(de SYS2BIG (W)
  (CONTINUABLEerror 99
                    "Sys2Big cant convert to bignum, no bignum's loaded"
                    (Sys2Int W)))
 
  % test inum range: fits inum in inf part?
(compiletime
(ds isinum(u) (weq (signedfield u 26 38) u))
                 %      (wplus2 (compiler-constant 'infstartingbit) -1)
                 %       (wplus2 infbitlength 1)
                 %  ) u ) )
)
 
(COMPILETIME
 (progn
(de NameGen (NAME PART)
          (INTERN (CONCAT (ID2STRING NAME) (ID2STRING PART))))
 
(ds nextarg () (progn (setq U (CDR U))
                              (CAR U)))
(ds Prologue NIL
         (progn (setq Generic (nextarg))
                (setq WGen (nextarg))
                (setq FGen (nextarg))
                (setq BGen (nextarg))
                (setq hardGen (NAMEGen Generic '-hardcase))
                (setq trivial (NAMEGen Generic '-trivialcase))
                (setq genbetap (NAMEGen Generic '-betap))
                (FLAG1 hardGen 'NOTYETINTERNALFUNCTION)
                NIL))
 
(dm defarith2entry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap
         bhi-function blow-function)
   (Prologue)
   (setq bhi-function (NAMEGen Generic '-betahi))
   (setq blow-function (NAMEGen Generic '-betalow))
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap
                  bhi-function blow-function)
             (list Generic WGen FGen BGen hardGen trivial genbetap
                  bhi-function blow-function))
  '(progn
    (de Generic (X Y)
      (prog (tr)
          (when (genbetap X Y)
             (setq X (WGen X Y))
             (if (isinum X) (return X) (return (Sys2Int x))))
          (setq tr (trivial x y))
          (when tr (return tr))
          (return (hardGen X Y))))
    (de hardGen (X Y)
      (prog (co bhi* blow*)
            (setq bhi* betahi*)
            (setq blow* betalow*)
            (setq betahi* (bhi-function))
            (setq betalow* (blow-function))
            (setq co (Coerce2 X Y 'Generic))
            (setq betahi* bhi*)
            (setq betalow* blow*)
          (return
           (case co
            ((posint-tag)   (Sys2Int (WGen ArithArgs0 ArithArgs1)))
            ((Floatnum-tag) (FGen ArithArgs0 ArithArgs1))
            ((bignum-tag)   (BGen ArithArgs0 ArithArgs1))))))
    NIL)))))
 
(dm defarith2divideentry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap
         bhi-function blow-function)
   (Prologue)
   (setq bhi-function (NAMEGen Generic '-betahi))
   (setq blow-function (NAMEGen Generic '-betalow))
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap
                  bhi-function blow-function)
             (list Generic WGen FGen BGen hardGen trivial genbetap
                  bhi-function blow-function))
  '(progn
    (de Generic (X Y)
      (prog (tr)
          (when (genbetap X Y) (return (wgen x y)))
          (setq tr (trivial x y))
          (when tr (return tr))
          (return (hardGen X Y))))
    (de hardGen (X Y)
      (prog (co bhi* blow*)
            (setq bhi* betahi*)
            (setq blow* betalow*)
            (setq betahi* (bhi-function))
            (setq betalow* (blow-function))
            (setq co (coerce2 X Y 'Generic))
            (setq betahi* bhi*)
            (setq betalow* blow*)
          (return
           (case co
            ((posint-tag)   (WGen ArithArgs0 ArithArgs1))
            ((Floatnum-tag) (FGen ArithArgs0 ArithArgs1))
            ((bignum-tag)   (BGen ArithArgs0 ArithArgs1))))))
    NIL)))))
 
 (dm defarithpred2entry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap)
   (Prologue)
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap)
             (list Generic WGen FGen BGen hardGen trivial genbetap))
 '(progn 
     (de Generic (X Y)
        (cond ((mbeta2p X Y) (WGen X Y))
              (T (hardGen X Y))))
     (de hardGen (X Y)
      (case (Coerce2 X Y 'Generic)
            ((posint-tag)   (WGen ArithArgs0 ArithArgs1))
            ((Floatnum-tag) (FGen ArithArgs0 ArithArgs1))
            ((bignum-tag)   (BGen ArithArgs0 ArithArgs1))))
     NIL)))))
 
 (dm defint1entry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap)
   (Prologue)
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap)
             (list Generic WGen FGen BGen hardGen trivial genbetap))
 '(progn 
      (de Generic (X)
         (cond ((mbetap X)
                  (progn (setq X (WGen X))
                      (cond ((isinum X) X)
                            (T  (Sys2Int X)))))
               (T (hardGen X))))
      (de hardGen (X)
        (case (Coerce1 X 'Generic)
            ((posint-tag)   (Sys2Int (WGen ArithArgs0)))
            ((Floatnum-tag) (NonIntegererror X 'Generic))
            ((bignum-tag)   (BGen ArithArgs0))
            (NIL   (nonnumber1error X 'Generic))))
  NIL)))))
 
 (dm defint2entry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap)
   (Prologue)
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap)
             (list Generic WGen FGen BGen hardGen trivial genbetap))
 '(progn 
   (de Generic (X Y)
       (cond ((mbeta2p X Y)
               (progn (setq X (WGen X Y))
                      (cond ((isinum X) X)
                            (T (Sys2Int X)))))
             (T (hardGen X Y))))
   (de hardGen (X Y)
      (case (Coerce2 X Y 'Generic)
            ((posint-tag) (Sys2Int (WGen ArithArgs0 ArithArgs1)))
            ((Floatnum-tag) (NonInteger2error X Y 'Generic))
            ((bignum-tag) (BGen ArithArgs0 ArithArgs1))))
 NIL)))))
 
 (dm defint2remainderentry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap
         bhi-function blow-function)
   (Prologue)
   (setq bhi-function (NAMEGen Generic '-betahi))
   (setq blow-function (NAMEGen Generic '-betalow))
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap
                  bhi-function blow-function)
             (list Generic WGen FGen BGen hardGen trivial genbetap
                  bhi-function blow-function))
 '(progn 
   (de Generic (X Y)
     (prog (tr)
       (cond ((genbetap X Y)
               (progn (setq X (WGen X Y))
                      (cond ((isinum X) (return X))
                            (T (return (Sys2Int X))))))
             ((setq tr (trivial x y)) (return tr))
             (T (return (hardGen X Y))))))
   (de hardGen (X Y)
      (prog (co bhi* blow*)
            (setq bhi* betahi*)
            (setq blow* betalow*)
            (setq betahi* (bhi-function))
            (setq betalow* (blow-function))
            (setq co (Coerce2 X Y 'Generic))
            (setq betahi* bhi*)
            (setq betalow* blow*)
          (return
           (case co
            ((posint-tag)   (Sys2Int (WGen ArithArgs0 ArithArgs1)))
            ((Floatnum-tag) (NonInteger2error X Y 'Generic))
            ((bignum-tag)   (BGen ArithArgs0 ArithArgs1))))))
 NIL)))))
 
(dm defarith1entry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap)
   (Prologue)
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap)
             (list Generic WGen FGen BGen hardGen trivial genbetap))
 '(progn 
   (de Generic (X)
      (cond ((mbetap X)
              (progn (setq X (WGen X))
                     (cond ((isinum X) X)
                           (T (Sys2Int X)))))
            (T (hardGen X))))
   (de hardGen (X)
      (case (Coerce1 X 'Generic)
            ((posint-tag) (Sys2Int (WGen ArithArgs0)))
            ((Floatnum-tag) (FGen ArithArgs0))
            ((bignum-tag)   (BGen ArithArgs0))
            (nil (nonnumber1error X 'Generic))))
         NIL)))))
 
 (dm defarithpred1entry (U)
  (prog (Generic WGen FGen BGen hardGen trivial genbetap)
   (Prologue)
   (return
    (subla (pair '(Generic WGen FGen BGen hardGen trivial genbetap)
             (list Generic WGen FGen BGen hardGen trivial genbetap))
 '(progn 
    (de Generic (X)
        (cond ((mbetap X) (WGen X))
              (T (hardGen X))))
    (de hardGen (X)
      (case (Coerce1 X 'Generic)
            ((posint-tag)   (WGen ArithArgs0))
            ((Floatnum-tag) (FGen ArithArgs0))
            ((bignum-tag)   (BGen ArithArgs0))
            (NIL NIL)))
   NIL)))))
 
 (ds defFloatentry (NAME PRIM)
     (de NAME (X Y)
         (fastmkfltn (PRIM x y))))
 NIL))
 
% The support procedures for coercing types
 
% returns type tag of Coerced X type and sets ArithArg0 to be Coerced X
% Beware of ADD1/SUB1 cases, maybe can optimize later
 
(de Coerce1 (X F)
  (prog (T1)
     (setq T1 (TAG X))
     (case T1 ((NegInt-tag) (setq T1 posint-tag))
              ((fixnum-tag) (progn (setq T1 posint-tag)
                                   (setq X (FIXVAL X)))))
     (setq ArithArgs0 X)
     (cond ((wgreaterp t1 floatnum-tag) (return (nonnumber1error X f))))
     (return T1)))
 
 
% returns type tag of strongest type and sets ArithArg0 to be Coerced X
% and ArithArg1 to Coerced Y.
(de Coerce2 (X Y F)
  (prog (T1 T2 PPP)
        (setq T1 (tag x) t2 (tag y))
        (case T1 ((NegInt-tag) (setq T1 posint-tag))
                 ((fixnum-tag) (progn (setq T1 posint-tag)
                                      (setq X (FIXVAL X)))))
        (case T2 ((NegInt-tag) (setq T2 posint-tag))
                 ((fixnum-tag) (progn (setq T2 posint-tag)
                                      (setq Y (FIXVAL Y)))))
        (setq ArithArgs0 X)
        (setq ArithArgs1 Y)
        (setq PPP 1)
        (cond ((WgreaterP T1 Floatnum-tag)
                                 (return (nonnumber2error X Y F)))
              ((eq T1 t2) (return T1))
              ((Wlessp T1 T2)
                  (progn (setq ppp T2)
                         (setq T2 T1)
                         (setq T1 ppp)
                         (setq PPP 0))))
        (cond ((WgreaterP T1 Floatnum-tag)
                                 (return (nonnumber2error X Y F))))
        (when (eq (wor (wxor T1 Floatnum-tag) T2) 0)
           % this means  t1 = float and t2 = posint
               (return (coerce2-hardcase T1 PPP)))
        (return T1)))
 
(de coerce2-hardcase (T1 PPP)
    (if (weq PPP 0)
        (setq ArithArgs0 (STATICINTFloat ArithArgs0))
        (setq ArithArgs1 (STATICINTFloat ArithArgs1)))
    T1)
 
(de NonInteger2error (X Y F)
  (CONTINUABLEerror 99 "Non-integer argument in arithmetic"
                    (list F (MKQUOTE X) (MKQUOTE Y))))
 
(de nonnumber1error (X F)
  (CONTINUABLEerror 99 "Non-numeric argument in arithmetic"
                    (list F (MKQUOTE X))))
 
(de nonnumber2error (X Y F)
  (CONTINUABLEerror 99 "Non-numeric argument in arithmetic"
                    (list F (MKQUOTE X) (MKQUOTE Y))))
 
% Now Generate the entries for each operator
 
(compiletime
(progn
 (ds plus2-trivialcase (x y)
     (cond ((eq x 0) y)
           ((eq y 0) x)
           (t nil)))
 (ds plus2-betap (x y)
    (and (plus2-betap-1 x)
         (plus2-betap-1 y)))
 (ds plus2-betap-1 (x)
    (or (eq 0 (wand (lmask5) x))
        (eq (lmask5) (wand (lmask5) x))))
 (put 'lmask5 'opencode '((Lmask (Reg 1) 5)))
 (put 'plus2-betahi  'opencode '((rmask (Reg 1) 41)))
 (put 'plus2-betalow 'opencode '((lmask (Reg 1) 23)))
%(put 'plus2-betahi  'opencode '((rmask (Reg 1) 62)))
%(put 'plus2-betalow 'opencode '((lmask (Reg 1) 2)))
))
 
(defarith2entry PLUS2 WPLUS2 FloatPLUS2 BIGPLUS2)
 
(defFloatentry FloatPLUS2 *FPLUS2)
 
(compiletime
(progn
 (ds difference-trivialcase (x y)
    (cond ((eq y 0) x)
          ((eq x y) 0)
          (t nil)))
 (ds difference-betap (x y)
    (and (difference-betap-1 x)
         (difference-betap-1 y)))
 (ds difference-betap-1 (x)
    (or (eq 0 (wand (lmask5) x))
        (eq (lmask5) (wand (lmask5) x))))
 (put 'difference-betahi  'opencode '((rmask (Reg 1) 41)))
 (put 'difference-betalow 'opencode '((lmask (Reg 1) 23)))
%(put 'difference-betahi  'opencode '((rmask (Reg 1) 62)))
%(put 'difference-betalow 'opencode '((lmask (Reg 1) 2)))
))
 
(defarith2entry DIFFERENCE WDIFFERENCE FloatDIFFERENCE BIGDIFFERENCE)
 
(defFloatentry FloatDIFFERENCE *FDIFFERENCE)
 
(compiletime
(progn
 (ds times2-trivialcase (x y)
    (cond ((eq x 0) 0)
          ((eq x 1) y)
          ((eq y 0) 0)
          ((eq y 1) x)
          (t nil)))
 (ds times2-betap (x y)
    (and (times2-betap-1 x)
         (times2-betap-1 y)))
 (ds times2-betap-1 (x)
    (or (eq 0 (wand (lmask43) x))
        (eq (lmask43) (wand (lmask43) x))))
 (put 'lmask43 'opencode '((Lmask (Reg 1) 43)))
 (put 'times2-betahi  'opencode '((move (Reg 1) (Reg Betahi*))))
 (put 'times2-betalow 'opencode '((move (Reg 1) (Reg Betalow*))))
))
 
(defarith2entry TIMES2 longTIMES2 FloatTIMES2 BIGTIMES2)
 
        % Beware of Overflow
(defFloatentry FloatTIMES2 *FTIMES2)
 
(compiletime
(progn
 (ds quotient-trivialcase (x y)
    (cond ((eq x 0) 0)
          ((eq y 1) x)
          ((eq x y) 1)
          (t nil)))
 (ds quotient-betap (x y)
    (and (quotient-betap-1 x)
         (quotient-betap-1 y)))
 (ds quotient-betap-1 (x)
    (or (eq 0 (wand (lmask20) x))
        (eq (lmask20) (wand (lmask20) x))))
 (put 'lmask20 'opencode '((Lmask (Reg 1) 20)))
 (put 'quotient-betahi  'opencode '((rmask (Reg 1) 41)))
 (put 'quotient-betalow 'opencode '((lmask (Reg 1) 23)))
))
 
(defarith2entry QUOTIENT longQUOTIENT FloatQUOTIENT BIGQUOTIENT)
 
(de wdivide (x y)
  (prog (qu rem)
  (if (eq y 0) (return (conterror 99 "Attempt to divide by 0 in DIVIDE"
                   (divide x y)))
      (setq qu (longdivide-quotient (longdivide x y)))
      (setq rem (longdivide-remainder))
      (return (cons (if (isinum qu) qu (Sys2Int qu))
                    (if (isinum rem) rem (Sys2Int rem)))))))
 
(compiletime(progn
 (ds Floatdivide  (x y) (NonInteger2error X Y 'Divide))
 
 (ds divide-trivialcase (x y)
    (cond ((eq x 0) (cons 0 0))
          ((eq y 1) (cons x 0))
          ((eq x y) (cons 1 0))
          (t nil)))
 (ds divide-betap (x y)
    (and (divide-betap-1 x)
         (divide-betap-1 y)))
 (ds divide-betap-1 (x)
    (or (eq 0 (wand (lmask20) x))
        (eq (lmask20) (wand (lmask20) x))))
 (put 'lmask20 'opencode '((Lmask (Reg 1) 20)))
 (put 'divide-betahi  'opencode '((Rmask (Reg 1) 41)))
 (put 'divide-betalow 'opencode '((Lmask (Reg 1) 23)))
))
 
(defarith2divideentry divide wdivide Floatdivide BIGdivide)
 
(defFloatentry FloatQUOTIENT *FQUOTIENT)
 
(defarithpred2entry GREATERP WGREATERP FloatGREATERP BIGGREATERP)
 
(de FloatGREATERP (X Y)
        (*FGREATERP X y))
 
(defarithpred2entry lessp Wlessp Floatlessp BIGlessp)
 
(de Floatlessp (X Y)
        (*Flessp x y))
 
(de FDUMMY (X Y)
  (STDerror "Fdummy should never be called"))
 
(compiletime
(progn
 (ds remainder-trivialcase (x y)
    (cond ((eq x y) 0)
          ((eq y 1) 0)
          (t nil)))
 (ds remainder-betap (x y)
    (and (remainder-betap-1 x)
         (remainder-betap-1 y)))
 (ds remainder-betap-1 (x)
    (or (eq 0 (wand (lmask20) x))
        (eq (lmask20) (wand (lmask20) x))))
 (put 'lmask20 'opencode '((Lmask (Reg 1) 20)))
 (put 'remainder-betahi  'opencode '((Rmask (Reg 1) 41)))
 (put 'remainder-betalow 'opencode '((Lmask (Reg 1) 23)))
))
 
(defint2remainderentry REMAINDER WREMAINDER FDUMMY BIGREMAINDER)
 
(compiletime
(progn
 (ds gcdn-trivialcase (x y)
    (cond ((eq x y) x)
          ((eq y 1) 1)
          (t nil)))
 (ds gcdn-betap (x y)
    (and (gcdn-betap-1 x)
         (gcdn-betap-1 y)))
 (ds gcdn-betap-1 (x)
    (or (eq 0 (wand (lmask20) x))
        (eq (lmask20) (wand (lmask20) x))))
 (put 'lmask20 'opencode '((Lmask (Reg 1) 20)))
 (put 'gcdn-betahi  'opencode '((Rmask (Reg 1) 41)))
 (put 'gcdn-betalow 'opencode '((Lmask (Reg 1) 23)))
))
 
(defint2remainderentry GCDN WGCDN FDUMMY BIGGCDN)
 
(defint2entry LAND WAND FDUMMY BIGLAND)
 
(defint2entry LOR WOR FDUMMY BIGLOR)
 
(defint1entry LNOT WNOT FDUMMY BIGLNOT)
 
(defint2entry LXOR WXOR FDUMMY BIGLXOR)
 
% Cant DO Lshift in terms of BETA sized shifts
% Will be redefined in BIG package
(defint2entry LSHIFT WSHIFT FDUMMY BIGLSHIFT)
 
(de LSH(A B) (LSHIFT A B))    % better: copyd
 
(defarith1entry ADD1 IADD1 (LAMBDA (X)
                              (FloatPLUS2 X '1.0)) BIGADD1)
 
(defarith1entry SUB1 ISUB1 (LAMBDA (X)
                              (FloatDIFFERENCE X '1.0)) BIGSUB1)
 
(defarith1entry MINUS IMINUS (LAMBDA (X)
                              (FloatDIFFERENCE '0.0 X)) BIGMINUS)
 
(defarith1entry FIX (LAMBDA (X) X) Floatfix (lambda (x) x) )
 
(de FloatFIX (X)
  (Sys2Int (*WFIX X)))
 
(de Float (X)
  (case (TAG X)
        ((posint-tag NegInt-tag) (INTFloat X))
        ((fixnum-tag)            (INTFloat (FIXVAL X)))
        ((Floatnum-tag)           X)
        ((bignum-tag)            (FloatFROMBIGNUM X))
        (NIL                     (nonnumber1error X 'Float))))
 
(de IntFloat (X)
        (fastMkfltn (*WFloat X)))
 
(defarithpred1entry MMINUSP IMINUSP
                     (Lambda (x) (Floatlessp X '0.0)) BIGMINUSP)
 
(compiletime (put 'crayFloatone 'opencode '((sconst10 (reg 1)))))
 
(de MINUSP (u) (and (numberp u) (MMINUSP u)))

(de Onep (x)
         (or (weq x 1)
             (and (Floatp x)
                  (weq (getmem (Floatbase x)) (crayFloatone)))))
 
(de Zerop (x)
         (or (weq x 0)
             (and (Floatp x) % Float 0 = Int 0 with Cray machines
                  (weq (getmem (Floatbase x)) 0))))
 
(de wgcdn (u v)
   (when (wlessp u 0) (setq u (wminus u)))
   (when (wlessp v 0) (setq v (wminus v)))
   (if (wlessp u v) (wgcdn-aux v u) (wgcdn-aux u v)))
 
(de wgcdn-aux (u v)
 
% all is positive
 
   (cond ((eq v 0) u)
         (t (wgcdn-aux  v (longdivide-remainder (longdivide u v))))
)  )
 
(de wgcdn1 (b1 b2)  % special routine called from big when gcd is in
   (setq b1 (big2sys b1)) % w-range already
   (setq b2 (big2sys b2))
   (setq b1 (wgcdn-aux b1 b2))
   (if (isinum b1) b1 (Sys2Int b1)))
