%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:ERROR-ERRORSET.SL
% Title:        The most basic ERROR and ERRORSET
% Author:       Eric Benson
% Created:      20 August 1981
% Modified:     22-Mar-84 11:23:05 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 26 Jan 1984 0959-PST (Brian Beach)
%  Corrected file name in header.
% 01-Dec-83 14:38:15 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(global '(emsg*))
 
% gets current error message
(fluid '(*backtrace % controls backtrace printing (actual)
       *inner*backtrace % controls backtrace printing (formal)
       *emsgp % controls message printing
       *break % controls breaking
       breaklevel* % nesting level of breaks
       maxbreaklevel* % maximum permitted ...
       stkupbnd stackupperbound
       *continuableerror))
 
% if T, inside a continuable error
(loadtime
  (progn (setq *emsgp t)
         (setq *backtrace nil)
         (setq *break t)))
 
(de error (number message)
  %. Throw to ErrorSet
  (prog (*continuableerror)
        (setq emsg* message)
        (when *emsgp
          (errorprintf "***** %l" message)
          % Error number is not printed
          (when (and *break (lessp breaklevel* maxbreaklevel*))
            (return (break))))
        (return (progn (when *inner*backtrace
                         (backtrace))
                       (throw '$error$ number)))))
 
% More useful version of ERRORSET
(dm errset (u)
  ((lambda (form flag)
           (list (list 'lambda '(*emsgp)
                       (list 'catch ''$error$ (list 'ncons form)))
                 flag))
   (cadr u)
   (if (null (cddr u))
     t
     (caddr u))))
 
(de errorset (form *emsgp *inner*backtrace)
  %. Protected Eval
  (setq stkupbnd (wdifference stackupperbound 200))
    % reset stackupperbound for stack overflow processing
  (catch '$error$ (list (eval form))))
 
% eval form
 
