%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Sun Unix 4.2.
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de start-trap() nil)

(fluid '(*arith387*))    % indicates that 387 processor is present

(compiletime
    %%
    %%  This code generates trap handlers for OS386 DOS extender
    %%
 (progn
   (flag '(build-trap-message smalldump sig__set *freset1 locktraphandler0)
	 'internalfunction)
   (setf *sigcalls* nil)
   %
   % Create a list of initialization code needed to set up the traps. The
   % return value is the entry points for each signal handler.
   % This is later flaged as a
   %  cmacro so the returned code can be inserted directly into the lap coded
   %  procedure 'initializeinterrupts each time it is called. The list
   %  *sigentries is then inserted to define the entry points.

   (de *sigsetup (signumber function handler errorstring imin ch)
      (let ((primus (gensym)) (pares (gensym))(cc (id2int ch)))
       (setf *sigcalls* (append
	     % Sigset takes a function pointer.
	     `((*move (wconst ,signumber) (reg 1))
	       (*move ,handler (reg 2))
	       (*link sig__set expr 2))
	     *sigcalls*))
       % Return the the function definition for the signal handler.
       `((*entry ,function expr 0)
     ,handler

	    % restore DS pointer
	 (*move (reg cs) (reg 1))
	 (*wxor (reg 1) 8)
	 (*move (reg 1)(reg ds))


	   % restore processor status flags
     ,(if (eq imin 0) '(pop (reg 2)) (gensym))
	 (*move 16#3246 (reg 1)) % bits for the future eflags register
	 (push (reg 1))
	 (popfd)                 % set eflags register

	,(if (eq imin 100) '(*link *freset1 expr 0) (gensym))

	   % prepare error analysis
	 (pop (reg 3))           % error address
	 (pop (reg 4))           % old CS
	 (pop (reg 5))           % dummy: previous eflags word
	 (*move (reg cs)(reg 5)) % current CS
     (*move (reg esp) (reg 1))
    ,(if (eq signumber 12) `(*wplus2 (reg st) 2000) (gensym))
     (*move (quote ,errorstring) (reg 2))
     (*move (wconst ,signumber) (reg 1))
     (*link smalldump expr 3)
	 (push (reg 1))
     (*link initializeinterrupts expr 0)
     (pop (reg 2))
     (*move (wconst ,signumber) (reg 1))
     (*wplus2 (reg 1)(wconst 10000))
     (*link error expr 0))
      ) )
 
   % Return the entry point list. Defined as a cmacro.
   (de *sigcall () *sigcalls*)
 
   (defcmacro *sigcall)
   (defcmacro *sigsetup)
 
   )
) % End of compiletime for OS386.
 

(compiletime
    %%
    %%  This code generates trap handlers for DPMI extender (WINDOWS)
    %%
 (progn
   (setf *sigcallsw* nil)
   %
   % Create a list of initialization code needed to set up the traps. The
   % return value is the entry points for each signal handler.
   % This is later flaged as a
   %  cmacro so the returned code can be inserted directly into the lap coded
   %  procedure 'initializeinterrupts each time it is called. The list
   %  *sigentries is then inserted to define the entry points.

   (de *sigsetupw (signumber function handler errorstring imin ch)
      (let ((primus (gensym)) (pares (gensym))(cc (id2int ch)))
       (setf *sigcallsw* (append
	     % Sigset takes a function pointer.
	     *sigcallsw*
	     `((*move (wconst ,signumber) (reg 1))
	       (*move ,handler (reg 2))
	       (*link sig__set expr 2))
	     ))
       % Return the the function definition for the signal handler.
       `((*entry ,function expr 0)
     ,handler

	,(if (eq imin 100) '(*link *freset1 expr 0) (gensym))

	   % prepare error analysis
     (*move (reg 2)(reg 4))       % old stack pointer
   %           (reg 3)            % error address

    ,(if (eq signumber 12) `(*wplus2 (reg st) 2000) (gensym))
     (*move (quote ,errorstring) (reg 2))
     (*move (wconst ,signumber) (reg 1))
     %%%%(*link smalldump expr 3)
     %%%%(push (reg 1))
     %%%%(*link initializeinterrupts expr 0)
     %%%%(pop (reg 1))
     (*link Winterrupthandler expr 0))
      ) )
 
   % Return the entry point list. Defined as a cmacro.
   (de *sigcallw () *sigcallsw*)
 
   (defcmacro *sigcallw)
   (defcmacro *sigsetupw)
 
   )
) % End of compiletime for DPMI.
 
(lap '(
     %  OS386 handlers:

       (*sigsetup 13 SegHandler  SegHandlerInstruction
				    "General Protection Fault" 0 G)
%      (*sigsetup 16#1B BreakHandler  
%                 BreakHandlerInstruction  "Terminal Interrupt" 0 T)
       (*sigsetup 0 SysHandler  SysHandlerInstruction  "Zero Divide" -1 Z)
       (*sigsetup 6 AlrmHandler AlrmHandlerInstruction "Bad Instruction" -1 B)
%      (*sigsetup 8 Dblhandler DoubleInstruction "Double fault" -1 D)
       (*sigsetup 12 Stkhandler StkhandlerInstruction "Stack Overflow" 0 S)
       (*sigsetup 16#75 FLThand FLThandinstr "Floating point exception" 100 F)
       (*sigsetup 16#23 CntrChd CntrChainstr "Interrupt" -1 C)
     
     % DPMI handlers:
       
       (*sigsetupw 0 SysHandler  WSysHandlerInstruction  "Zero Divide" -1 Z) 
       (*sigsetupw 13 SegHandler  WSegHandlerInstruction
				    "General Protection Fault" 0 G)
       (*sigsetupw 16#1B BreakHandler  
		  WBreakHandlerInstruction  "Keyboard Interrupt" 0 T)
       (*sigsetupw 6 AlrmHandler WAlrmHandlerInstruction "Bad Instruction" -1 B)
%      (*sigsetupw 8 Dblhandler WDoubleInstruction "Double fault" -1 D)
       (*sigsetupw 12 Stkhandler WStkhandlerInstruction "Stack Overflow" 0 S)
       (*sigsetupw 16#75 FLThand WFLThandinstr "Floating point exception" 100 F)
       (*sigsetupw 16#23 CntrChd WCntrChainstr "Interrupt" -1 C)
       

     %  init sequence  
       (*entry initializeinterrupts expr 0)
       (*move ($fluid **windows)(reg 1))
       (cmp 1 (reg 1))
       (jne nowin)
       (*sigcallw)
       (jmp common)
  nowin
       (*sigcall)
       (*call locktraphandler)
  common     
       (*call *freset)
       (*exit 0)))
 

(de *freset()(when *arith387* (*freset1)))  % only when 387 present

(lap '((*entry *freset1 expr 0)

       %%% (*move (wconst (char ?)) (reg 4))
       %%% (*move 16#0200 (reg 1))
       %%% (int 16#21)

	   % clear 387 processor
       (*move 0 (reg 1))
       (halfword 16#f0e6) % out f0,al
       (*move 16#20 (reg 1))
       (halfword 16#a0e6)

       (jmp (label primus))
    primus
       (jmp (label pares))
    pares

       (halfword 16#20e6)


       (*move 100000 (reg 5)) % we have to wait far a while ?!
     lab
       (*move ($fluid onewordbuffer) (reg 4))
       (*wplus2 (reg 5) -1)
       (*jumpwgreaterp (label lab) (reg 5) 0)

       (halfword 16#e3db)
       (wait)

       (fclex)
       (wait)

       (*move ($fluid onewordbuffer) (reg 5))
       (*move 0 (reg 2))
       (*move (reg 2) (displacement (reg 5) 0))
       (fstcw (displacement (reg 5) 0))
       (wait)

       (*move (displacement (reg 5) 0) (reg 2))
       (*wor (reg 2)  2#110000000000)
       (*wand (reg 2) 2#11111111110010)
       (*move (reg 2) (displacement (reg 5) 0))
       (fldcw (displacement (reg 5) 0))
       (wait)

       %%% (*move (wconst (char ?)) (reg 4))
       %%% (*move 16#0200 (reg 1))
       %%% (int 16#21)

       (*exit 0)))

   % lock this code in memory
(compiletime
   (put 'locktraphandler1 'opencode % (locktraphandler base pages)
     '((*move (reg 1) (reg EDX))
       (*move (reg 2) (reg ECX))
       (*move 16#eb06 (reg 1))
	   (int 16#21))))

(de lockpages(a e)
   (prog (n)
     (setq a (wtimes2 (wquotient (inf a) 4096) 4096))
     (setq e (wtimes2 (iadd1 (wquotient (inf e) 4096)) 4096))
     (setq n (wquotient (wdifference e a) 4096))
     (locktraphandler1 a n)
  ))

(de locktraphandler()
    % prevent blowup in case of non-paging environment
    (errorset '(locktraphandler0) nil nil))

(de locktraphandler0 ()
   (prog (x)
    (*lock387)
    (lockpages (cdr (getd 'start-trap)) (cdr (getd 'end-trap)) )
  ))

(de *lock387() nil)   % hook for overloading by arith387

(lap '((*entry sig__set expr 2)
       (*move ($fluid **windows)(reg 4))
       (cmp 1 (reg 4))
       (jne nowin)
       (*move (reg 2)(reg 3))   % address    
       (*move (reg cs)(reg 2))  % segment selector
       (*move (reg 1)(reg 1))   % interrupt number
       (*move 14 (reg 4))
       (*call psll-call)
       (*exit 0)
   nowin
       (*wor (reg 1) 16#2500)    % function number for interrupt
       (*move (reg 2)(reg 4))
       (int 16#21)
       (*exit 0) ))

(fluid '(*dump))

(compiletime
   (ds sgetv(a p)
       (*get-stack (wplus2 a (wtimes2 p addressingunitsperitem)))))

(commentoutcode
(de smalldump1 (array estri eip)
    (when *dump
	(console-newline)
	(console-print-string "*** Dump of Registers ***")
	(console-newline)          (console-newline)  
	(console-print-string "*** Reg 1 (EAX) : " )(unixputn (sgetv array 6))
	(console-newline) 
	(console-print-string "*** Reg 2 (EBX) : " )(unixputn (sgetv array 5))
	(console-newline) 
	(console-print-string "*** Reg 3 (ECX) : " )(unixputn (sgetv array 4))
	(console-newline) 
	(console-print-string "*** Reg 4 (EDX) : " )(unixputn (sgetv array 3))
	(console-newline) 
	(console-print-string "*** Reg 5 (EBP) : " )(unixputn (sgetv array 2))
	(console-newline)
	(console-print-string "*** Reg T1 (EDI): " )(unixputn (sgetv array 1))
	(console-newline) 
	(console-print-string "*** Reg T2 (ESI): " )(unixputn (sgetv array 0))
	(console-newline) 
	(console-print-string "*** Reg ST (ESP): " )(unixputn (sgetv array 7))
	(console-newline) 
	(console-print-string "*** Reg CS:       " )
	(unixputn (sgetv array (wplus2 10 imin)))
	(console-newline)
	(console-print-string "*** Reg EIP:      " )
	(unixputn (sgetv array (wplus2 imin 9)))
	(console-newline))
))

(de smalldump (signbr estri eip oldcs cs)
	(setq eip (build-trap-message estri (when (eq oldcs cs) eip)))
	(when (equal estri "Stack Overflow") 
	      (console-print-string "***** ")
	      (console-print-string eip)
			  (unixputn eip)
	      (console-newline)
	      (Throw '$error$ nil))
	eip)

(de build-trap-message (trap-type trap-addr)
    (let (extra-info sy)
	 (if trap-addr
	  (progn
	(setq sy (code-address-to-symbol (inf trap-addr)))
	(setf extra-info (bldmsg "%w%w%w%xx" " in " sy " +"
			    (wdifference trap-addr (inf (getfcodepointer sy)))))
      )
	  (setq extra-info " (unknown place)"))
    (bldmsg "%w%w" trap-type extra-info)))
 
(compiletime (fluid '(code-address* closest-address* closest-symbol*)))
 
(de code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0) numb)
	(mapobl #'(lambda (symbol)
		 (when (fcodep symbol)
		       (let ((address (inf (getfcodepointer symbol))))
			    (when (and (ileq address code-address*)
				       (igreaterp address closest-address*))
				  (setq closest-address* address)
				  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de end-trap() nil)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  special code for Windows:
%
%  when an interrupthandler is called under Windows, the stack position
%  is lost because of the 16-bit architecture of Windows. Therefore
%  the interrupt handler is entered an artificial (high!) stack pointer. 
%  Here we try to find a reasonable stack position for the error
%  recovery.
 

(compiletime 
   (progn
      (ds catchtopsp()(wgetv catchstackptr 2))
      (put 'actsp 'opencode '((*move (reg st)(reg 1))) )
      (put 'setsp 'opencode '((*move (reg 2)(reg st))) )  %preserves (reg 1)
   ))

(fluid '(werror*))

(de Winterrupthandler(errnbr txt adr stack)
   (let ((asp (actsp))        % actual stack
	 (csp (wdifference (catchtopsp) 400)) )
       (when (wleq csp asp)   % stack is growing downwards
	     (console-print-string 
		"fatal stack overflow during trap handler - restarting")
	     (Throw '$error$ nil))
     %     stack and address are known only modulo 2**16
     %     try to reconstruct the old stack pointer
       (when (and (wgreaterp stack 1000) (wlessp stack asp))
	     (while (wlessp (wplus2 stack 16#10000) asp)
		    (setq stack (wplus2 stack 16#10000)))
	     (setq csp stack))
       (when (weq errnbr 16#1B)(setq errnbr 2))  % the unix number
       (setq werror* (wplus2 errnbr 10000))
       (werror* (setsp (wbuild-trap-message txt adr) csp)) ))

(de werror*(u)(error werror* u))

(de wbuild-trap-message (trap-type a)
 (if (not !*dump)
  trap-type
  (let (l sy)
      (setq l "")
      (while (wlessp a nextbps)
	(setq sy (code-address-to-symbol a))
	(when (not (eq sy 'INITCODE))
	    (setq l (bldmsg "%w %w+%w" l sy 
		 (wdifference a (inf (getfcodepointer sy)))) ))
	(setq a (wplus2 a 16#10000))
      )
    (bldmsg "%w in one of (%w)" trap-type l))))
 

(lap '((*entry stacko expr 0)
       (*move (reg st) (reg 1))
       (*exit 0) ))

% End of file.
