%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for Mips
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 10-Mar-89 (Winfried Neun)
%  Handling for Mips as for Sparc (SUN4). Complete redesign.
%  Also installed fluid *bruch* to allow gc (etc) to turn off
%  cntrl-c interrupt detection, which destroy system. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(sigsetup build-trap-message) 'iinternalfunction))

(fluid '(*bruch* Mips-sigcp Mips-sigaddr Mips-sigcode Mips-signumber
         *dump Mips-sigpc Mips-dumparray*))

% this is the interface for toploop

(de initializeinterrupts ()
    (ieee_handler (strbase (strinf "set"))
                  (strbase (strinf "common"))
                  (symfnc (id2int 'signalhandler))) 

          % all signals are initiated expect for those iin the followin list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 18 19 20 21 22 25 28))) (sigsetup i)))))

% this installs the handler

(de sigsetup (signumber)
    (sun3_sigset signumber (wtimes2 4 (symfnc (id2int 'signalhandler)))))
 
% this is the 'handler'

(setq Mips-dumparray* (gtwarray 127))
 
(de signal-done ()
     (prog (handel)
         (initializeinterrupts)
         (when (eq 2#10101111101 (wshift (wgetv mips-sigpc 0) -21))
                (console-print-string "***** Stack overflow *****")
                (console-newline)
                (wputv Mips-sigcp  29
                       (wplus2 (wgetv Mips-sigcp 29) 10000)))
         (setq handel (wshift (symfnc (id2int 'lisprpv)) 2))
         (wputv Mips-sigcp 37 handel)
)    )            %        'Brutal' version like Cray X-MP for UNICOS
 

(setq *bruch* 0)

(de lisprpv()
   (prog (mess)
    (setq *bruch* 0)
    (smalldump)
    (setq mess (atsoc (wshift Mips-signumber 2)
 			'((1 . "Hangup")
                          (2 . "Interrupt")
                          (3 . "Quit") 
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap") 
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (when (eq Mips-signumber 8)
          (cond ((eq Mips-sigcode 16#c8)
                         (setq mess '( 0 . "Floating Point divide by zero")))
                ((eq Mips-sigcode 16#d4)
                         (setq mess '( 0 . "Floating Point overflow")))
                ((eq Mips-sigcode 16#cc)
                         (setq mess '( 0 . "Floating Point underflow")))))
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  Mips-signumber)))
    (build-trap-message mess Mips-sigpc)
    (setq Mips-sigaddr 0 Mips-sigcp 0) % for gc
    (return (stderror))))

(de smalldump()
  (when *dump 
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn Mips-signumber)
    (console-print-string " , signal code: 16#")
    (unixputn Mips-sigcode)
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn Mips-sigpc)
    (console-print-string " , signal Address: 16#")
    (unixputn Mips-sigaddr)
    (console-print-string " , context: 16#")
    (unixputn Mips-sigcp)
    (console-newline)
    (setq Mips-sigaddr 0 Mips-sigcp 0) % for gc
    (console-newline) (console-newline)

    (console-print-string " Reg 0-3 ") (unixputn (wgetv Mips-dumparray* 0))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 1))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 2))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 3)) 
    (console-newline)
    (console-print-string " Reg 4-7 ") (unixputn (wgetv Mips-dumparray* 4))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 5))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 6))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 7)) 
    (console-newline)
    (console-print-string "Reg 8-11 ") (unixputn (wgetv Mips-dumparray* 8))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 9))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 10))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 11)) 
    (console-newline)
    (console-print-string "Reg12-15 ") (unixputn (wgetv Mips-dumparray* 12))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 13))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 14))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 15)) 
    (console-newline)
    (console-print-string "Reg16-19 ") (unixputn (wgetv Mips-dumparray* 16))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 17))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 18))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 19)) 
    (console-newline)
    (console-print-string "Reg20-23 ") (unixputn (wgetv Mips-dumparray* 20))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 21))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 22))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 23))
    (console-newline)
    (console-print-string "Reg24-27 ")
    (unixputn (wgetv Mips-dumparray* 24))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 25))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 26))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 27))
    (console-newline)
    (console-print-string "Reg28-31 ") 
    (unixputn (wgetv Mips-dumparray* 28))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 29))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 30))
    (console-print-string " ") (unixputn (wgetv Mips-dumparray* 31))
    (console-newline) (console-newline)
    (console-print-string " cp0 stat ") (unixputn (wgetv Mips-dumparray* 32))
    (console-newline)
    (console-print-string " M low    ") (unixputn (wgetv Mips-dumparray* 33))
    (console-newline)
    (console-print-string " M high   ") (unixputn (wgetv Mips-dumparray* 34))
    (console-newline)
    (console-print-string " sc_cause ") (unixputn (wgetv Mips-dumparray* 35))
    (console-newline)
    (console-print-string " badvaddr ") (unixputn (wgetv Mips-dumparray* 36))
    (console-newline)
    (console-print-string " sc_pc    ") (unixputn (wgetv Mips-dumparray* 37))
    (console-newline)
    (console-print-string " semaddr  ") (unixputn (wgetv Mips-dumparray* 38))
    (console-newline)
    (console-print-string " sc_signo ") (unixputn (wgetv Mips-dumparray* 39))
    (console-newline)
    (console-print-string " sc_smask ") (unixputn (wgetv Mips-dumparray* 40))
    (console-newline)
    (console-print-string " onstack ") (unixputn (wgetv Mips-dumparray* 41))
    (console-newline) (console-newline)
    (console-print-string " ++++ Dump of Stack ++++") (console-newline)
    (console-newline) 
    (for (from i 77 126 5)
      (do (unixputn (wgetv Mips-dumparray* i)) (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 1 i)))
	  (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 2 i)))
	  (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 3 i)))
	  (console-print-string " ")
          (unixputn (wgetv Mips-dumparray* (wplus2 4 i))) 
          (console-newline)))
))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in " 
                                 (code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(compiletime (fluid '(code-address* closest-address* closest-symbol*)))

(de code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0))
    (mapobl (function hugolein))
%      (mapobl #'(lambda (symbol)
%          (when (fcodep symbol)
%	       (let ((address (codeinf (getfcodepointer symbol))))
%			    (when (and (ileq address
%					     code-address*)
%				       (igreaterp address
%						  closest-address*))
%				  (setq closest-address*
%					address)
%				  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de hugolein (symbol)
                    (when (fcodep symbol)
                            (let ((address (codeinf (getfcodepointer symbol))))
                                    (when (and (ileq address
                                                     code-address*)
                                               (igreaterp address
                                                          closest-address*))
                                          (setq closest-address*
                                                address)
                                          (setq closest-symbol* symbol))))))
