%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXU:MULS.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for MIPS processor
% Author:       H. Melenk, W. Neun
% Created:      25 January 1989
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DEC ALpha version by W. Neun, ZIB Berlin
%
%  this file is needed compiletime only
%
% Double length multiply and divide for the 68020 processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
% They map the 2*60 bit operations needed by BIG to the 2*32 bit operations
% of the processor by shifting etc.
(fluid '(*second-value*))
 
(compiletime (load inum))
(bothtimes
(de wdivide (x y)
   (let ((xx (wquotient x y)))
     (setq *second-value* (Wdifference x (wtimes2 xx y)))
     xx))
)

(put 'wtimesdouble 'opencode
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The high order bits are placed in fluid variable.
     '(% double length unsigned mutiply;
	(MULq (reg 1) (reg 2) (reg t1))
	(umulh (reg 1) (reg 2) (reg 3))
	(*WSHIFT (reg 3) 2)
	(srl (reg t1) 62 (reg 2))
	(*WOR (reg 3)(reg 2))
	(*MOVE (reg 3)($FLUID *second-value*))
	(sll (reg t1) 2 (reg 1))
	(*WSHIFT (reg 1) -2) ))
 
(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '( (MULq (reg 1) (reg 2) (reg t1))
	(umulh (reg 1) (reg 2) (reg 4))
	(*move (reg t1) (reg 1))
	(*move (reg 4) ($FLUID *second-value*))))


(commentoutcode
(de wquotientdouble (X Y Z)
      (PROG (A B C D)
               (SETQ A (oldWQUOTIENTDOUBLE X Y Z))
               (SETQ B *SECOND-VALUE*)
               (SETQ C (WQUOTIENTDOUBLELAP X Y Z))
               (SETQ D *SECOND-VALUE*)
               (unless (AND (EQ A C) (EQ B D)) (print (list "falsch"
                         x y z "," a b c d)) (quit))
               (return c)))


)
(fluid '(*second-value*))

(compiletime (progn
 
 (defcmacro *wrshift3)

 (de *wrshift3 (a1 a2 a3)
   (prog (resultingcode*)
     (setq a1 (resolveoperand '(reg t1) a1))
     (setq a2 (resolveoperand '(reg t1) a2))
     (setq a3 (resolveoperand '(reg t1) a3))
     (return `((srl ,a2 ,a3 ,a1)))))
))

(commentoutcode % this is the algorithm for the lap hack behind:
(de oldwquotientdouble (arg1 arg2 arg3)

   (prog (quot rem carrybit i)

             % first divide the uper 30 bits by divisor and put the
             % quotient into the result and continue with the remainder
 
     (setq quot (wdivide arg1 arg3))
     (setq rem *second-value*)
 
     (setq arg2 (wshift arg2 2))  %initial shift
 
     (ifor (from i 0 61 1) (do
 
            % now shift the remainder left by 1 and add the most significant
            % bit of arg2. shift arg2 left one bit
            % if sum is greater or equal divisor, add a bit '1' to remainder
            % else add a '0'.
            % do it until arg2 is eaten up. quot is the result (quotient),
            %          rem is the remainder (in *second-value*)
 
            (progn
             (setq quot (wshift quot 1))
             (setq carrybit (wshift arg2 -63))
             (setq arg2 (wshift arg2 1))
             (setq rem (wplus2 (wshift rem 1) carrybit))
             (when (wgeq rem arg3) (setq rem (wdifference rem arg3))
                                   (setq quot (wplus2 quot 1)))
      )     )              )
     (setq *second-value* rem)
     (return quot)))
)
 
(lap '((*entry WQUOTIENTDOUBLE expr 3)
       (*alloc 0)
       (*move (reg 1) (reg t1)) % arg1
       (*move (reg 2) (reg t2)) % arg2
       (*move (reg 3) (reg t3)) % arg3 
%      (*move (reg 3) (reg 2))
%      (*link wdivide expr 2)
       (*move (quote 0) (reg t4))  % quot
       (*move (reg 1) (reg t5)) % rem
       (*wshift (reg t2) 2)   
                                   62 times!
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab1) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab1
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab2) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab2
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab3) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab3
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab4) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab4
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab5) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab5
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab6) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab6
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab7) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab7
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab8) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab8
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab9) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab9
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab10) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab10
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab11) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab11
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab12) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab12
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab13) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab13
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab14) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab14
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab15) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab15
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab16) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab16
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab17) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab17
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab18) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab18
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab19) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab19
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab20) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab20
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab21) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab21
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab22) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab22
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab23) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab23
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab24) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab24
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab25) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab25
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab26) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab26
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab27) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab27
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab28) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab28
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab29) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab29
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab30) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab30
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab31) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab31
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab32) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab32
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab33) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab33
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab34) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab34
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab35) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab35
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab36) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab36
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab37) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab37
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab38) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab38
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab39) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab39
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab40) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab40
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab41) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab41
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab42) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab42
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab43) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab43
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab44) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab44
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab45) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab45
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab46) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab46
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab47) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab47
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab48) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab48
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab49) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab49
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab50) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab50
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab51) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab51
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab52) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab52
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab53) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab53
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab54) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab54
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab55) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab55
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab56) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab56
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab57) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab57
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab58) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab58
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab59) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab59
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab60) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab60
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab61) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab61
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpwlessp (label lab62) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab62                           %done
       (*move (reg t5) ($fluid *second-value*))
       (*move (reg t4) (reg 1))
       (*exit 0)))


(de wquotientdoublesmall (arg1 arg2 arg3 norm)

   (prog (quot summa)

             % first divide the uper 30 bits by divisor and put the
             % quotient into the result and continue with the remainder

     (when (eq 0 arg1) (return (wdivide arg2 arg3)))

     (setq arg2 (wshift arg2 2))  %initial shift

     (setq summa 0 quot 0)
   lupe
     (when (wgeq (iplus2 norm summa) 62) (setq norm  (wdifference 62 summa)))
     (setq summa (iplus2 summa norm))

     (setq arg1 (wor (wshift arg1  norm) (wshift arg2 (wdifference norm 64))))
     (setq arg2 (wshift arg2 norm))
     (setq quot (wor (wshift quot norm) (wdivide arg1 arg3)))
     (setq arg1 *second-value*)
     (when (neq summa 62) (go lupe))

     (return quot)))

(put 'addcarry 'opencode
       '((*MOVE ($FLUID carry*) (reg 2))
         (*WPLUS2 (reg 1)(reg 2))
         (srl (reg 1) 62 (reg 3))
         (*MOVE (reg 3) ($FLUID carry*))
         (*WSHIFT (reg 1) 2)
         (*WSHIFT (reg 1) -2)))

(put 'subcarry 'opencode
       '((*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         (srl (reg 1) 63 (reg 2))
         (*MOVE (reg 2) ($FLUID carry*))
         (*WSHIFT (reg 2) 62)
         (*WPLUS2 (reg 1) (reg 2))
       ))
         

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   functions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.
 
%   get the value for bbase* into a register
(put 'bbase** 'opencode
     '((*MOVE 1 (reg 1))
       (*WSHIFT (reg 1) 62)))
 
%   get the value for logicalbits* into a register     
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
        (*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*WSHIFT (reg 1) 2)
        (*WSHIFT (reg 1) -2)))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -62)))

(put 'bbase** 'destroys '((reg 1))) 
(put 'logicalbits** 'destroys '((reg 1)))  
(put 'remainder-bbase 'destroys '((reg 1)))  
(put 'quotient-bbase 'destroys '((reg 1)))  


