%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS32.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for DEC Alpha processor
% Author:       H. Melenk
% Created:      11 Sept  1989
% Modified:
% Mode:         Lisp
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the  processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
%
% Full 32 bit version : 28-June 1993 (H. Melenk)

(fluid '(*second-value* carry*))

(compiletime (load inum))

(bothtimes
(de wdivide (x y)
   (let ((xx (wquotient x y)))
     (setq *second-value* (Wdifference x (wtimes2 xx y)))
     xx))
)

%-------------------- plus, difference with carry -------------


% add and set carry
%
%   (a + b) -> (carry*,result)

(bothtimes

(put 'addAndSetCarry 'opencode
       '(
         (ADDQ (reg 1)(reg 2)(reg 3))
           % move cf to carry*
         (CMPULT (reg 3)(reg 1)(reg 4))
         (CMPULT (reg 3)(reg 2)(reg 5))
         (*move (reg 3)(reg 1))
         (*wor (reg 4) (reg 5))
         (*move (reg 4)($fluid carry*))
       ))

)

% add with carry
%
%   (a + b + carry*) -> (carry*,result)

(bothtimes

(put 'addwithcarry 'opencode
       '(
           % add with carry
         (*MOVE ($fluid carry*)(reg 3))
         (ADDQ (reg 2)(reg 3)(reg 3)) 
         (CMPULT (reg 3)(reg 2)(reg 4)) % new carry
         (ADDQ (reg 1)(reg 3)(reg 2))
         (CMPULT (reg 2)(reg 1)(reg 5))
         (*wor (reg 4)(reg 5))
         (CMPULT (reg 2)(reg 3)(reg 5))
         (*wor (reg 4)(reg 5))
         (*move (reg 4)($fluid carry*))
         (*move (reg 2)(reg 1))
       ))

)

% add and add carry
%
%   (a + b) -> result,  (*second-value* + carry) -> *second-value*

(bothtimes

(put 'addAndAddCarry 'opencode
       '(
           % add (reg 2) to (reg 1)
         (ADDQ (reg 1)(reg 2)(reg 3))
           % compute carry 
         (CMPULT (reg 3)(reg 2)(reg 5))
         (*move ($FLUID *second-value*) (reg 2))
         (CMPULT (reg 3)(reg 1)(reg 4))
         (*move (reg 3)(reg 1))
         (*wor (reg 4)(reg 5))
         (ADDQ (reg 2)(reg 4)(reg 4))
         (*move (reg 4) ($FLUID *second-value*))
       ))

)

% subtract with borrow
%
%  (a - (b + carry!*)) -> result, carry*=1 if borrow

(bothtimes

(put 'subtractwithborrow 'opencode
       '(
           % move carry* to cf
         (*move ($fluid carry*)(reg 4))    
           % add carry to b
         (ADDQ (reg 4)(reg 2)(reg 3))
         (CMPULT (reg 3)(reg 2)(reg 4)) % 1st borrow
           % is (reg 1) smaller than (reg 3)?
         (CMPULT (reg 1)(reg 3)(reg 5))
         (ADDQ (reg 5)(reg 4)(reg 4))
         (*move (reg 4) ($FLUID carry*))
           % subtract 
         (SUBQ (reg 1)(reg 3)(reg 1))
       ))

)

%------------------- unsigned greaterp ---------------------

% code based on integer carry; 
 
(bothtimes

(put 'ugreaterp* 'opencode 
   % returns 1 if arg1 > arg2 unsigned.
'( (CMPULT (reg 2)(reg 1)(reg 1)))) 

)

(bothtimes

(ds ugreaterp(a b)(eq 1 (ugreaterp* a b)))

)

%-------------------- times, quotient -------------------------

(put 'wtimesdouble 'opencode
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The high order bits are placed in fluid variable.
     '(% double length unsigned mutiply;
        (MULq (reg 1) (reg 2) (reg t1))
        (umulh (reg 1) (reg 2) (reg 3))
        (*MOVE (reg 3)($FLUID *second-value*))
        (*move (reg t1) (reg 1))
      ))

(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '( (MULq (reg 1) (reg 2) (reg t1))
        (umulh (reg 1) (reg 2) (reg 4))
        (*move (reg t1) (reg 1))
        (*move (reg 4) ($FLUID *second-value*))))

(de wquotientdouble(a0 a1 b)
   (prog(q rem)
       % unsigned double quotient using the routine
       % for signed quotient.
       % simple case: neither a nor b formally negative
     (when (and (wgreaterp a0 0) (wgreaterp b 0))
           (return (wquotientdouble* a0 a1 b)))
       % approximate quotient
     (setq q (wquotientdouble*
              (wshift a0 -1)
              (wor (wshift a0 (plus bitsperword -1)) (wshift a1 -1))
              (wshift b -1)))
     (setq rem
            (wor (wshift *second-value* 1)
                 (wand a1 1)))
     (when (eq 0 (wand b 1)) (go ret))
   % last bit of b is 1: adjust remainder
     (setq carry* 0)
     (setq rem (subtractwithborrow rem q))
     (when (weq carry* 0) (go ret))
   % if carry, q has been to big
   % decrease q until rem swaps again
     (setq carry* 0)
     (while (weq carry* 0)
       (setq q (isub1 q))
       (setq rem (addAndSetCarry b rem)))
   ret
     (setq *second-value* rem)
     (return q) ))

(compiletime (progn

 (defcmacro *wrshift3)

 (de *wrshift3 (a1 a2 a3)
   (prog (resultingcode*)
     (setq a1 (resolveoperand '(reg t1) a1))
     (setq a2 (resolveoperand '(reg t1) a2))
     (setq a3 (resolveoperand '(reg t1) a3))
     (return `((srl ,a2 ,a3 ,a1)))))
))

(commentoutcode % this is the algorithm for the lap hack behind:
(de wquotientdouble* (arg1 arg2 arg3)

   (prog (quot rem carrybit i)

             % first divide the uper 30 bits by divisor and put the
             % quotient into the result and continue with the remainder

     (setq quot 0)
     (setq rem arg1)

     (ifor (from i 0 63 1) (do

            % now shift the remainder left by 1 and add the most significant            % bit of arg2. shift arg2 left one bit
            % if sum is greater or equal divisor, add a bit '1' to remainder            % else add a '0'.
            % do it until arg2 is eaten up. quot is the result (quotient),
            %          rem is the remainder (in *second-value*)

           (progn
             (setq quot (wshift quot 1))
             (setq carrybit (wshift arg2 -63))
             (setq arg2 (wshift arg2 1))
             (setq rem (wplus2 (wshift rem 1) carrybit))
             (when (not (ugreaterp arg3 rem)) 
                   (setq rem (wdifference rem arg3))
                   (setq quot (wplus2 quot 1)))
      )     )              )
     (setq *second-value* rem)
     (return quot)))
)

(compiletime (progn

(defcmacro *jumpunsignedlessp)

(de *jumpunsignedlessp (Lbl ArgOne ArgTwo)
        (*jumpif ArgOne ArgTwo lbl 'CMPULT 'BNE))

))

(lap '((*entry WQUOTIENTDOUBLE* expr 3)
       (*alloc 0)
       (*move (reg 1) (reg t1)) % arg1
       (*move (reg 2) (reg t2)) % arg2
       (*move (reg 3) (reg t3)) % arg3 
%      (*move (reg 3) (reg 2))
%      (*link wdivide expr 2)
       (*move (quote 0) (reg t4))  % quot
       (*move (reg 1) (reg t5)) % rem
%                                  64 times!
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab1) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab1
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab2) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab2
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab3) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab3
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab4) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab4
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab5) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab5
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab6) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab6
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab7) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab7
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab8) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab8
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab9) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab9
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab10) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab10
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab11) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab11
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab12) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab12
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab13) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab13
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab14) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab14
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab15) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab15
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab16) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab16
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab17) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab17
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab18) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab18
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab19) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab19
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab20) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab20
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab21) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab21
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab22) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab22
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab23) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab23
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab24) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab24
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab25) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab25
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab26) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab26
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab27) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab27
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab28) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab28
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab29) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab29
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t5) 1) 
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab30) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab30
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab31) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab31
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab32) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab32
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab33) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab33
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab34) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab34
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab35) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab35
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab36) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab36
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab37) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab37
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab38) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab38
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab39) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab39
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab40) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab40
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab41) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab41
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab42) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab42
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab43) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab43
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab44) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab44
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab45) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab45
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab46) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab46
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab47) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab47
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab48) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab48
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab49) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab49
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab50) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab50
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab51) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab51
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab52) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab52
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab53) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab53
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab54) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab54
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab55) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab55
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab56) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab56
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab57) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab57
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab58) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab58
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab59) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab59
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab60) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab60
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab61) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab61
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab62) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab62                           %done
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab63) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab63
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 63) % carrybit
       (*wshift (reg t2) 1)
       (*wshift (reg t5) 1)
       (*wplus2 (reg t5) (reg 2))
       (*jumpunsignedlessp (label lab64) (reg t5) (reg t3))
       (*wdifference (reg t5) (reg t3))
       (*wplus2 (reg t4) 1)
    lab64
       (*move (reg t5) ($fluid *second-value*))
       (*move (reg t4) (reg 1))
       (*exit 0)))

