%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS32.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for MIPS processor
% Author:       H. Melenk
% Created:      11 Sept  1989
% Modified:
% Mode:         Lisp
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the  processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
%
% Full 32 bit version : 28-June 1993 (H. Melenk)

(fluid '(*second-value* carry*))

(compiletime (load inum))
 

% --------------------- plus, minus ---------------------------

% add and set carry
%
%   (a + b) -> (carry*,result)

(bothtimes

(put 'addAndSetCarry 'opencode
       '(
         (ADDU  (reg 3) (reg 1)(reg 2))
           % move cf to carry*
         (SLTU (reg 4)(reg 3) (reg 1))
         (SLTU (reg 5)(reg 3) (reg 2))
         (*wor (reg 4)(reg 5))
         (*move(reg 4) ($FLUID carry*))
         (*move(reg 3)(reg 1))
       ))

)

% add with carry
%
%   (a + b + carry*) -> (carry*,result)

(bothtimes

(put 'addwithcarry 'opencode
       '(
         (*MOVE ($fluid carry*)(reg 3))
         (ADDU (reg 3)(reg 2)(reg 3))
         (SLTU (reg 4)(reg 3)(reg 2)) % new carry
         (ADDU (reg 2)(reg 1)(reg 3))
         (SLTU (reg 5)(reg 2)(reg 1))
         (*wor (reg 4)(reg 5))
         (SLTU (reg 5)(reg 2)(reg 3))
         (*wor (reg 4)(reg 5))
         (*move (reg 4)($fluid carry*))
         (*move (reg 2)(reg 1))

       ))

)

% add and add carry
%
%   (a + b) -> result,  (*second-value* + carry) -> *second-value*

(bothtimes

(put 'addAndAddCarry 'opencode
       '(
           % add (reg 2) to (reg 1)
         (ADDU (reg 3)(reg 1)(reg 2))
           % compute carry
         (SLTU (reg 5)(reg 3)(reg 2))
         (SLTU (reg 4)(reg 3)(reg 1))
         (*move (reg 3)(reg 1))
         (*wor (reg 4)(reg 5))
         (*move ($FLUID *second-value*) (reg 2))
         (ADDU (reg 4)(reg 2)(reg 4))
         (*move (reg 4) ($FLUID *second-value*))
       ))

)

% subtract with borrow
%
%  (a - (b + carry!*)) -> result, carry*=1 if borrow

(bothtimes

(put 'subtractwithborrow 'opencode
       '(
           % move carry* to cf
         (*move ($fluid carry*)(reg 4))
           % add carry to b
         (ADDU (reg 3)(reg 4)(reg 2))
         (SLTU (reg 4)(reg 3)(reg 2)) % 1st borrow
           % is (reg 1) smaller than (reg 3)?
         (SLTU (reg 5)(reg 1)(reg 3))
         (ADDU (reg 4)(reg 5)(reg 4))
         (*move (reg 4) ($FLUID carry*))
           % subtract
         (SUBU (reg 1)(reg 1)(reg 3))
       ))

)

%------------------- unsigned greaterp ---------------------

% code based on integer carry; 
 
(bothtimes

(put 'ugreaterp* 'opencode 
   % returns 1 if arg1 > arg2 unsigned.
'( (SLTU (reg 1)(reg 2)(reg 1))))

)


(ds ugreaterp(a b)(eq 1 (ugreaterp* a b)))


%------------------------- times, quotient -------------------

(bothtimes
(put 'wdivide 'opencode
                '((div (reg 1) (reg 2))
                  (Mfhi (reg 3))
                  (mflo (reg 1))
                  (*move (reg 3) ($fluid *second-value*))))
)

(compiletime
(put 'udivide 'opencode
                '((divu (reg 1) (reg 2))
                  (Mfhi (reg 3))
                  (mflo (reg 1))
                  (*move (reg 3) ($fluid *second-value*))))
)



(put 'wtimesdouble 'opencode
      '((MULTU(reg 1) (reg 2))
      (mfhi  (reg 4))
      (mflo  (reg 1))
      (*move (reg 4) ($FLUID *second-value*))))

(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '((MULT (reg 1) (reg 2))
      (mfhi  (reg 4))
      (mflo  (reg 1))
      (*move (reg 4) ($FLUID *second-value*))))

%--------------------------------------------------------------------
%  Unsigned divide 64:32 -> (32,32) in software
%--------------------------------------------------------------------

(commentoutcode

    % pattern for the lap code below

(de wquotientdouble (arg1 arg2 arg3)

   (prog (quot rem locarry hicarry i)

     (setq quot 0)
     (setq rem arg1)

     (ifor (from i 0 31 1) (do

            % now shift the remainder left by 1 and add the most significant
            % bit of arg2. shift arg2 left one bit
            % if sum is greater or equal divisor, add a bit '1' to remainder
            % else add a '0'.
            % do it until arg2 is eaten up. quot is the result (quotient),
            %          rem is the remainder (in *second-value*)

            (progn
             (setq quot (wshift quot 1))
             (setq locarry (wshift arg2 -31))
             (setq arg2 (wshift arg2 1))
             (setq hicarry (wshift rem -31))
             (setq rem (wplus2 (wshift rem 1) locarry))
             (when (or (not (ugreaterp arg3 rem)) 
                       (eq hicarry 1))
                   (setq rem (wdifference rem arg3))
                   (setq quot (wplus2 quot 1)))
      )     )              )
     (setq *second-value* rem)
     (return quot)))

)

(compiletime

(defcmacro *jumpunsignedlessp)

(de *jumpunsignedlessp (Lbl ArgOne ArgTwo)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg Error) Lbl)
                 )

        '(( (SLTU (reg t6) ArgOne ArgTwo)
            (BNE (reg t6)(reg 0) ArgThree)
            (NOP)
         ))))))
)


 (lap '((*entry wquotientdouble expr 3)
%  (prog (quot rem locarry hicarry i)
%          t1   t2   t3      t4    5
         %    (setq quot 0)
     (*move (quote 0) (reg t1))
         %    (setq rem arg1)
     (*move (reg 1)(reg t2))
         %    (ifor (from i 0 31 1) (do
     (*move (quote -33) (reg 5))
loop
     (*wplus2 (reg 5)(quote 1))
     (*jumpeq (label finish) (reg 5)(quote 0))
         %  (progn
         %   (setq quot (wshift quot 1))
     (*wshift (reg t1) 1)
         %   (setq locarry (wshift arg2 -31))
     (*move (reg 2)(reg t3))
     (*wshift (reg t3) -31)
         %   (setq arg2 (wshift arg2 1))
     (*wshift (reg 2) 1)
         %   (setq hicarry (wshift rem -31))
     (*move (reg t2) (reg t4))
     (*wshift (reg t4) -31)
         %   (setq rem (wplus2 (wshift rem 1) locarry))
     (*wshift (reg t2) 1)
     (*wplus2 (reg t2) (reg t3))
         %   (when (or (not (ugreaterp arg3 rem))
         %             (eq hicarry 1))
     (*wdifference (reg t4)(quote 1))
     (*jumpeq (label yes)(reg t4)(quote 0))
     (*jumpunsignedlessp (label loop) (reg t2)(reg 3))
  yes
         %         (setq rem (wdifference rem arg3))
     (*wdifference (reg t2)(reg 3))
         %         (setq quot (wplus2 quot 1)))
     (*wplus2 (reg t1) (quote 1))
     (*jump (label loop))
  finish
        %  )     )              )
        %  (setq *second-value* rem)
     (*move (reg t2)($fluid *second-value*))
        % (return quot)))
     (*move (reg t1)(reg 1))
     (*exit 0)
))

