%
% PAD_SYSCALLS.RED - Pad related routines
%
% Author:      G. Q. Maguire Jr.
%              Computer Science Dept.
%              University of Utah
% Date:        2 November 1982
% Copyright (c) 1982 G. Q. Maguire Jr.
%

% 14-MAR-84, JWpeterson, Fixed handling of returned status
% 28-aug-84, JW-peterson, Fixed bugs in pad_inq_windows

% ---------------------------------------------------------------------
% PAD related routines
% ---------------------------------------------------------------------

(de pad_clear_frame (streamid seek_key) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 seek_key '(integer32))      % Stream_$SK_T
  (cleararg 3) 

  (syscall 'pad_$clear_frame 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_close_frame (streamid) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 

  (syscall 'pad_$close_frame 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_cooked (streamid) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 

  (syscall 'pad_$cooked 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_cpr_enable (streamid report_type) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 report_type '(integer16))   % Integer*2
  (cleararg 3) 

  (syscall 'pad_$cpr_enable 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_create (padname 
                              panetype 
                              relatedstreamid 
                              side 
                              paneoptions 
                              panesize) 
(prog (padnamelength panestreamid status) 
  (setq padnamelength (stringlength padname)) 


  (putarg 1 padname '(string))          % ARRAY of CHAR
  (putarg 2 padnamelength '(integer16)) % 16 bit integer name length
  (putarg 3 panetype '(integer16))      % PAD_$TYPE_T
  (putarg 4 relatedstreamid '(integer16)) 	% Stream_$ID_T
  (putarg 5 side '(integer16))          % Pad_$Side_T
  (putarg 6 paneoptions '(integer16))   % Integer*2
  (putarg 7 panesize '(integer16))      % Integer*2
  (cleararg 8) 
  (cleararg 9) 

  (syscall 'pad_$create 9) 

  (setq status (getarg 9 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (setq panestreamid (getarg 8 '(integer16) '(inum))) 
                                                % Stream_$ID_T
            (apolloresetstatus) 
            (return panestreamid) nil)) (t 
         
    (return (apolloerror status))))))


(de pad_create_window (pathname windowtype window) 
(prog (windowarray status padnamelength streamid) 
  (setq padnamelength (stringlength pathname)) 
  (putarg 1 pathname '(string)) 
  (putarg 2 padnamelength '(integer16)) 
  (putarg 3 windowtype '(integer16))    % PAD_$Type_T
  (putarg 4 1 '(integer16))             % "Unit #", hardcoded to 1
  (setq windowarray (gthalfwords 4)) 
  (puthalfword windowarray 0 (first window)) 	% Convert list of window params
  (puthalfword windowarray 1 (second window)) 	% into consequtive int16's.
  (puthalfword windowarray 2 (third window)) 
  (puthalfword windowarray 3 (fourth window)) 
  (putobjpointer 5 windowarray) 
  (cleararg 6) 
  (cleararg 7) 

  (syscall 'pad_$create_window 7) 
  
  (setq streamid (getarg 6 '(integer16) '(inum))) 
  (setq status (getarg 7 '(integer32) '(fixn))) 

  (cond ((equal status 0) (progn (apolloresetstatus) (return streamid))) (t 
     (return (apolloerror status))))))


(de pad_create_frame (streamid width height) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 width '(integer16))         % Integer*2
  (putarg 3 height '(integer16))        % Integer*2
  (cleararg 4) 

  (syscall 'pad_$create_frame 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_def_pfk (streamid 
                               keyname 
                               definition) 
(prog (definitionlength status) 

  (setq definitionlength (stringlength definition)) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 keyname '(string))          % array of 4 chars
  (putarg 3 definition '(string))       % array of chars
  (putarg 4 definitionlength '(integer16)) 	% Integer*2
  (cleararg 5) 

  (syscall 'pad_$def_pfk 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_delete_frame (streamid) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 

  (syscall 'pad_$delete_frame 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_dm_cmd (streamid 
                                                command) 
(prog (commandlength status) 

  (setq commandlength (stringlength command)) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 command '(string))          % array of chars
  (putarg 3 commandlength '(integer16)) % Integer*2
  (cleararg 4) 

  (syscall 'pad_$dm_cmd 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_edit_wait (streamid) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 

  (syscall 'pad_$edit_wait 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_inq_font (streamid) 
(prog (fontwidth 
             fontheight 
             fontname 
             fontnamesize 
             fontnamelength 
             blankstring 
             status) 

  (setq fontnamesize 256)               % Maximum font name length
  (setq blankstring (mkstring 256 32)) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 
  (cleararg 3) 
  (putarg 4 blankstring '(string))      % ARRAY [1..256] of CHAR
  (putarg 5 fontnamesize '(integer16))  % Integer*2
  (cleararg 6) 
  (cleararg 7) 

  (syscall 'pad_$inq_font 7) 

  (setq status (getarg 7 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
   (progn (apolloresetstatus) 
           (setq fontwidth (getarg 2 '(integer16) '(inum))) 	% 16 bit integer
           (setq fontheight (getarg 3 '(integer16) '(inum))) 	% 16 bit integer

                                                  % 16 bit integer name length
           (setq fontnamelength (getarg 6 '(integer16) '(inum))) 

           % convert to LISP string
            (setq fontname (getarg 4 (cons 'string fontnamelength) '(string))) 
            (return (list fontwidth fontheight fontname)))) (t 
         
    (return (apolloerror status))))))


(de pad_inq_position (streamid) 
(prog (xposition yposition status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 
  (cleararg 3) 
  (cleararg 4) 

  (syscall 'pad_$inq_position 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
   (progn (apolloresetstatus) 
           (setq xposition (getarg 2 '(integer16) '(inum))) 	% 16 bit integer
           (setq yposition (getarg 3 '(integer16) '(inum))) 	% 16 bit integer

            (return (list xposition yposition)) nil)) (t 
         
    (return (apolloerror status))))))


(fluid '(pad_window_list *))
(setq pad_window_list* (gtbps 40))

(de pad_inq_windows (streamid) 
(prog (windowlistsize windowlist windowcount status) 

  (setq windowlistsize 1)               % only one window (makes data cleaner)

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putobjpointer 2 pad_window_list*) 
  (putarg 3 windowlistsize '(integer16)) 	% INTEGER*2
  (cleararg 4) 
  (cleararg 5) 

  (syscall 'pad_$inq_windows 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
   (progn (apolloresetstatus) 
           (setq windowlist (list 
              (gethalfword pad_window_list* 0) 	% top
              (gethalfword pad_window_list* 1) 	% left
              (gethalfword pad_window_list* 2) 	% width
              (gethalfword pad_window_list* 3))) 	% height
           
           (setq windowcount (getarg 4 '(integer16) '(inum))) 	% 16 bit integer

            (return (list windowlist windowcount)))) (t 
         
    (return (apolloerror status))))))


(de pad_load_font (streamid pathname) 
(prog (pathnamelength fontid status) 
  (setq pathnamelength (stringlength pathname)) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 pathname '(string))         % ARRAY of CHAR
  (putarg 3 pathnamelength '(integer16)) 	% 16 bit integer name length
  (cleararg 4) 
  (cleararg 5) 

  (syscall 'pad_$load_font 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn 
            (apolloresetstatus) 
            (setq fontid (getarg 4 '(integer32) '(fixn))) 
            (return fontid))) (t 
         
    (return (apolloerror status))))))


(de pad_locate (streamid) 
(prog (xposition yposition character status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 
  (cleararg 3) 
  (cleararg 4) 
  (cleararg 5) 

  (syscall 'pad_$locate 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
   (progn (apolloresetstatus) 
           (setq xposition (getarg 2 '(integer16) '(inum))) 	% 16 bit integer
           (setq yposition (getarg 3 '(integer16) '(inum))) 	% 16 bit integer
           (setq character (getarg 4 '(integer16) '(inum))) 	% 16 bit integer
            (return (list xposition yposition character)) nil)) (t 
         
    (return (apolloerror status))))))


(de pad_move (streamid relorabs xposition yposition) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 relorabs '(integer16))      % ???
  (putarg 3 xposition '(integer16))     % INTEGER*2
  (putarg 4 yposition '(integer16))     % INTEGER*2
  (cleararg 5) 

  (syscall 'pad_$move 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_raw (streamid) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (cleararg 2) 

  (syscall 'pad_$raw 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_set_scale (streamid xscale yscale) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 xscale '(integer16))        % INTEGER*2
  (putarg 3 yscale '(integer16))        % INTEGER*2
  (cleararg 4) 

  (syscall 'pad_$set_scale 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_set_tabs (streamid tabstopvector) 
(prog (numberoftabstops status) 

  (setq numberoftabstops (plus (size tabstopvector) 1)) 

% needs better error handling
  (cond ((greaterp numberoftabstops 100) (printf "Too many tab stops%n"))) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 tabstopvector '(integer16 . numberoftabstops)) 
  (putarg 3 numberoftabstops '(integer16)) 	% INTEGER*2
  (cleararg 4) 

  (syscall 'pad_$set_tabs 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de pad_use_font (streamid fontid) 
(prog (status) 

  (putarg 1 streamid '(integer16))      % Stream_$ID_T
  (putarg 2 fontid '(integer16))        % INTEGER*2
  (cleararg 3) 

  (syscall 'pad_$use_font 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((eq status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))

